#ifndef PYTHONSYNTAXHIGHLIGHTER_GLOBAL_H
#define PYTHONSYNTAXHIGHLIGHTER_GLOBAL_H

#include <QtCore/qglobal.h>

#if defined(PYTHONSYNTAXHIGHLIGHTER_LIBRARY)
#  define PYTHONSYNTAXHIGHLIGHTERSHARED_EXPORT Q_DECL_EXPORT
#else
#  define PYTHONSYNTAXHIGHLIGHTERSHARED_EXPORT Q_DECL_IMPORT
#endif

#endif // PYTHONSYNTAXHIGHLIGHTER_GLOBAL_H
