#ifndef MULTIEDITORJSON_GLOBAL_H
#define MULTIEDITORJSON_GLOBAL_H

#include <QtCore/qglobal.h>

#if defined(MULTIEDITORJSON_LIBRARY)
#  define MULTIEDITORJSON_EXPORT Q_DECL_EXPORT
#else
#  define MULTIEDITORJSON_EXPORT Q_DECL_IMPORT
#endif

#endif // MULTIEDITORJSON_GLOBAL_H
