# Just my custom script to make copying fresh jar into DbAndroid faster
# Pawel Salawa

rm SQLiteStudioRemote.jar
cp /home/googie/projects/AndroidStudioProjects/SQLiteStudioRemoteProject/sqlitestudioremote/build/intermediates/bundles/release/classes.jar SQLiteStudioRemote.jar
