#ifndef SCRIPTINGPYTHON_GLOBAL_H
#define SCRIPTINGPYTHON_GLOBAL_H

#include <QtCore/qglobal.h>

#if defined(SCRIPTINGPYTHON_LIBRARY)
#  define SCRIPTINGPYTHONSHARED_EXPORT Q_DECL_EXPORT
#else
#  define SCRIPTINGPYTHONSHARED_EXPORT Q_DECL_IMPORT
#endif

#endif // SCRIPTINGPYTHON_GLOBAL_H
