#ifndef TCLSYNTAXHIGHLIGHTER_GLOBAL_H
#define TCLSYNTAXHIGHLIGHTER_GLOBAL_H

#include <QtCore/qglobal.h>

#if defined(TCLSYNTAXHIGHLIGHTER_LIBRARY)
#  define TCLSYNTAXHIGHLIGHTER_EXPORT Q_DECL_EXPORT
#else
#  define TCLSYNTAXHIGHLIGHTER_EXPORT Q_DECL_IMPORT
#endif

#endif // TCLSYNTAXHIGHLIGHTER_GLOBAL_H
