#ifndef CODEFORMATTER_H
#define CODEFORMATTER_H

#include "core_global.h"
#include "letos.h"
#include <QHash>

class CodeFormatterPlugin;
class Db;

class API_EXPORT CodeFormatter
{
    public:
        QString format(const QString& lang, const QString& code, Db* contextDb);

        void setFormatter(const QString& lang, CodeFormatterPlugin* formatterPlugin);
        CodeFormatterPlugin* getFormatter(const QString& lang);
        bool hasFormatter(const QString& lang);
        void fullUpdate();
        void updateCurrent();
        void storeCurrentSettings();

    private:
        QHash<QString,QHash<QString,CodeFormatterPlugin*>> availableFormatters;
        QHash<QString,CodeFormatterPlugin*> currentFormatter;
        bool modifyingConfig = false;
};

#define FORMATTER LETOS->getCodeFormatter()

#endif // CODEFORMATTER_H
