/* Driver template for the LEMON parser generator.
** The author disclaims copyright to this source code.
**
** This version of "lempar.c" is modified, slightly, for use by SQLite.
** The only modifications are the addition of a couple of NEVER()
** macros to disable tests that are needed in the case of a general
** LALR(1) grammar but which are always false in the
** specific grammar used by SQLite.
*/
/* First off, code is included that follows the "include" declaration
** in the input grammar file. */
#include <stdio.h>

#include "token.h"
#include "parsercontext.h"
#include "parser_helper_stubs.h"
#include "common/utils_sql.h"
#include "common/global.h"
#include "parser/ast/sqlitealtertable.h"
#include "parser/ast/sqliteanalyze.h"
#include "parser/ast/sqliteattach.h"
#include "parser/ast/sqlitebegintrans.h"
#include "parser/ast/sqlitecommittrans.h"
#include "parser/ast/sqlitecreateindex.h"
#include "parser/ast/sqlitecreatetable.h"
#include "parser/ast/sqlitecreatetrigger.h"
#include "parser/ast/sqlitecreateview.h"
#include "parser/ast/sqlitecreatevirtualtable.h"
#include "parser/ast/sqlitedelete.h"
#include "parser/ast/sqlitedetach.h"
#include "parser/ast/sqlitedropindex.h"
#include "parser/ast/sqlitedroptable.h"
#include "parser/ast/sqlitedroptrigger.h"
#include "parser/ast/sqlitedropview.h"
#include "parser/ast/sqliteemptyquery.h"
#include "parser/ast/sqliteinsert.h"
#include "parser/ast/sqlitepragma.h"
#include "parser/ast/sqlitereindex.h"
#include "parser/ast/sqliterelease.h"
#include "parser/ast/sqliterollback.h"
#include "parser/ast/sqlitesavepoint.h"
#include "parser/ast/sqliteselect.h"
#include "parser/ast/sqliteupdate.h"
#include "parser/ast/sqlitevacuum.h"
#include "parser/ast/sqliteexpr.h"
#include "parser/ast/sqlitecolumntype.h"
#include "parser/ast/sqliteconflictalgo.h"
#include "parser/ast/sqlitesortorder.h"
#include "parser/ast/sqliteindexedcolumn.h"
#include "parser/ast/sqliteforeignkey.h"
#include "parser/ast/sqlitewith.h"
#include "parser/ast/sqliteupsert.h"
#include "parser/ast/sqlitewindowdefinition.h"
#include "parser/ast/sqlitefilterover.h"
#include "parser/ast/sqlitenulls.h"
#include <QObject>
#include <QDebug>
#include <limits.h>

#ifdef assert
#   undef assert
#endif
#define assert(X) Q_ASSERT(X)
#define UNUSED_PARAMETER(X) (void)(X)
#define DONT_INHERIT_TOKENS(X) noTokenInheritanceFields << X

/* Next is all token values, in a form suitable for use by makeheaders.
** This section will be null unless lemon is run with the -m switch.
*/
/*
** These constants (all generated automatically by the parser generator)
** specify the various kinds of tokens (terminals) that the parser
** understands.
**
** Each symbol here is a terminal symbol in the grammar.
*/
/* Make sure the INTERFACE macro is defined.
*/
#ifndef INTERFACE
# define INTERFACE 1
#endif
/* The next thing included is series of defines which control
** various aspects of the generated parser.
**    YYCODETYPE         is the data type used for storing terminal
**                       and nonterminal numbers.  "unsigned char" is
**                       used if there are fewer than 250 terminals
**                       and nonterminals.  "int" is used otherwise.
**    YYNOCODE           is a number of type YYCODETYPE which corresponds
**                       to no legal terminal or nonterminal number.  This
**                       number is used to fill in empty slots of the hash
**                       table.
**    YYFALLBACK         If defined, this indicates that one or more tokens
**                       have fall-back values which should be used if the
**                       original value of the token will not parse.
**    YYACTIONTYPE       is the data type used for storing terminal
**                       and nonterminal numbers.  "unsigned char" is
**                       used if there are fewer than 250 rules and
**                       states combined.  "int" is used otherwise.
**    sqlite3_parseTOKENTYPE     is the data type used for minor tokens given
**                       directly to the parser from the tokenizer.
**    YYMINORTYPE        is the data type used for all minor tokens.
**                       This is typically a union of many types, one of
**                       which is sqlite3_parseTOKENTYPE.  The entry in the union
**                       for base tokens is called "yy0".
**    YYSTACKDEPTH       is the maximum depth of the parser's stack.  If
**                       zero the stack is dynamically sized using realloc()
**    sqlite3_parseARG_SDECL     A static variable declaration for the %extra_argument
**    sqlite3_parseARG_PDECL     A parameter declaration for the %extra_argument
**    sqlite3_parseARG_STORE     Code to store %extra_argument into yypParser
**    sqlite3_parseARG_FETCH     Code to extract %extra_argument from yypParser
**    YYNSTATE           the combined number of states.
**    YYNRULE            the number of rules in the grammar
**    YYERRORSYMBOL      is the code number of the error symbol.  If not
**                       defined, then do no error processing.
*/
#define YYCODETYPE unsigned short int
#define YYNOCODE 333
#define YYACTIONTYPE unsigned short int
#define YYWILDCARD 79
#define sqlite3_parseTOKENTYPE Token*
typedef union {
  int yyinit;
  sqlite3_parseTOKENTYPE yy0;
  SqliteColumnType* yy1;
  ParserCreateTableColumnConstraintList* yy11;
  QVariant* yy13;
  ParserTermOrLiteral* yy14;
  SqliteCreateTable::Constraint* yy15;
  SqliteUpsert* yy16;
  SqliteLimit* yy44;
  SqliteSelect::Core::JoinOp* yy85;
  SqliteCreateTrigger::Event* yy87;
  SqliteWindowDefinition::Window::Frame* yy89;
  SqliteSelect* yy113;
  SqliteCreateTrigger::Scope* yy115;
  SqliteCreateTable::Column* yy123;
  SqliteInitially* yy136;
  SqliteWith* yy137;
  ParserExprList* yy143;
  SqliteQuery* yy147;
  ParserXFullName* yy148;
  SqliteWith::CommonTableExpression* yy158;
  QStringList* yy159;
  SqliteForeignKey::Condition::Reaction* yy160;
  SqliteWindowDefinition::Window::Frame::Bound* yy169;
  ParserSetValueList* yy177;
  SqliteSelect::Core* yy182;
  ParserFkConditionList* yy184;
  SqliteSelect::Core::JoinConstraint* yy191;
  SqliteConflictAlgo* yy194;
  ParserQueryList* yy198;
  SqliteWindowDefinition::Window* yy199;
  SqliteExpr::LikeOp* yy232;
  QString* yy239;
  ParserIndexedColumnList* yy240;
  SqliteSelect::Core::SingleSource* yy257;
  ParserDeferSubClause* yy265;
  SqliteWindowDefinition::Window::Frame::Exclude* yy271;
  SqliteWindowDefinition::Window::Frame::RangeOrRows* yy282;
  ParserStubTransDetails* yy284;
  ParserExprNestedList* yy294;
  ParserStubCreateTableOption* yy303;
  SqliteNulls* yy331;
  ParserWindowDefList* yy339;
  SqliteFilterOver::Filter* yy341;
  ParserStubInsertOrReplace* yy368;
  ParserResultColumnList* yy373;
  SqliteWith::CommonTableExpression::AsMode* yy407;
  ParserCreateTableColumnList* yy426;
  SqliteSelect::Core::JoinSource* yy455;
  SqliteExpr* yy456;
  ParserStubAlias* yy457;
  ParserOrderByList* yy459;
  SqliteIndexedColumn* yy460;
  SqliteForeignKey::Condition* yy475;
  ParserCreateTableOptionList* yy497;
  SqliteCreateTrigger::Time* yy500;
  ParserIndexedBy* yy504;
  SqliteSortOrder* yy509;
  ParserCteList* yy512;
  bool* yy515;
  SqliteWindowDefinition* yy522;
  SqliteCreateTable::Column::Constraint* yy552;
  SqliteFilterOver* yy567;
  int* yy642;
  ParserStubExplain* yy643;
  ParserCreateTableConstraintList* yy645;
  SqliteSelect::CompoundOperator* yy646;
  ParserFullName* yy648;
  ParserOtherSourceList* yy659;
  SqliteFilterOver::Over* yy663;
} YYMINORTYPE;
#ifndef YYSTACKDEPTH
#define YYSTACKDEPTH 100
#endif
#define sqlite3_parseARG_SDECL ParserContext* parserContext;
#define sqlite3_parseARG_PDECL ,ParserContext* parserContext
#define sqlite3_parseARG_FETCH ParserContext* parserContext = yypParser->parserContext
#define sqlite3_parseARG_STORE yypParser->parserContext = parserContext
#define YYNSTATE 929
#define YYNRULE 515
#define YYFALLBACK 1
#define YY_NO_ACTION      (YYNSTATE+YYNRULE+2)
#define YY_ACCEPT_ACTION  (YYNSTATE+YYNRULE+1)
#define YY_ERROR_ACTION   (YYNSTATE+YYNRULE)

#define GET_CONTEXT yyParser* yypParser = pParser; sqlite3_parseARG_FETCH

/* The yyzerominor constant is used to initialize instances of
** YYMINORTYPE objects to zero. */
static const YYMINORTYPE yyzerominor = { 0 };

/* Define the yytestcase() macro to be a no-op if is not already defined
** otherwise.
**
** Applications can choose to define yytestcase() in the %include section
** to a macro that can assist in verifying code coverage.  For production
** code the yytestcase() macro should be turned off.  But it is useful
** for testing.
*/
#ifndef yytestcase
# define yytestcase(X)
#endif


/* Next are the tables used to determine what action to take based on the
** current state and lookahead token.  These tables are used to implement
** functions that take a state number and lookahead value and return an
** action integer.
**
** Suppose the action integer is N.  Then the action is determined as
** follows
**
**   0 <= N < YYNSTATE                  Shift N.  That is, push the lookahead
**                                      token onto the stack and goto state N.
**
**   YYNSTATE <= N < YYNSTATE+YYNRULE   Reduce by rule N-YYNSTATE.
**
**   N == YYNSTATE+YYNRULE              A syntax error has occurred.
**
**   N == YYNSTATE+YYNRULE+1            The parser accepts its input.
**
**   N == YYNSTATE+YYNRULE+2            No such action.  Denotes unused
**                                      slots in the yy_action[] table.
**
** The action table is constructed as a single large table named yy_action[].
** Given state S and lookahead X, the action is computed as
**
**      yy_action[ yy_shift_ofst[S] + X ]
**
** If the index value yy_shift_ofst[S]+X is out of range or if the value
** yy_lookahead[yy_shift_ofst[S]+X] is not equal to X or if yy_shift_ofst[S]
** is equal to YY_SHIFT_USE_DFLT, it means that the action is not in the table
** and that yy_default[S] should be used instead.
**
** The formula above is for computing the action when the lookahead is
** a terminal symbol.  If the lookahead is a non-terminal (as occurs after
** a reduce action) then the yy_reduce_ofst[] array is used in place of
** the yy_shift_ofst[] array and YY_REDUCE_USE_DFLT is used in place of
** YY_SHIFT_USE_DFLT.
**
** The following are the tables generated in this section:
**
**  yy_action[]        A single table containing all actions.
**  yy_lookahead[]     A table containing the lookahead for each entry in
**                     yy_action.  Used to detect hash collisions.
**  yy_shift_ofst[]    For each state, the offset into yy_action for
**                     shifting terminals.
**  yy_reduce_ofst[]   For each state, the offset into yy_action for
**                     shifting non-terminals after a reduce.
**  yy_default[]       Default action for each state.
*/
#define YY_ACTTAB_COUNT (3573)
static const YYACTIONTYPE yy_action[] = {
 /*     0 */   921,  396,  876,  558,   32,  857,  712,  282,   74,   75,
 /*    10 */   563,  305,  562,  124,  872,  872,   72,   72,   73,   73,
 /*    20 */    73,   73,  926,   71,   71,   71,   71,   70,   70,   69,
 /*    30 */    69,   69,   68,   66,  284,  869,   74,   75,  563,  305,
 /*    40 */    65,  128,  872,  872,   72,   72,   73,   73,   73,   73,
 /*    50 */   565,   71,   71,   71,   71,   70,   70,   69,   69,   69,
 /*    60 */    68,   66,  284,   83, 1356,  211,  725,  360,  502,  921,
 /*    70 */   275,  763,  870,  512,   73,   73,   73,   73,   63,   71,
 /*    80 */    71,   71,   71,   70,   70,   69,   69,   69,   68,   66,
 /*    90 */   284,  635,  921,  867,  497,   60,   61,  770,   95,  428,
 /*   100 */   865,  864,  382,   62, 1103,  513,  828,  765,  764,  921,
 /*   110 */   724,    2, 1103,   58,  827,  176,  880, 1103,  921,  464,
 /*   120 */   559,  562,  921,  921,  889,  918,  875,  387,  637, 1178,
 /*   130 */   892,   69,   69,   69,   68,   66,  284,  880,  891,  890,
 /*   140 */   889,  746, 1103,   71,   71,   71,   71,   70,   70,   69,
 /*   150 */    69,   69,   68,   66,  284, 1103,  549, 1103, 1103,  565,
 /*   160 */   501, 1103,  790,  343,  266,  265,  345, 1103, 1103, 1103,
 /*   170 */  1103, 1103, 1103,  880,  887,  610,  880,  860,  868,   12,
 /*   180 */    76,  870,  921,  357,  664,   65,  415,   63,  571,  920,
 /*   190 */  1103,  921,  921,  889,  929,  611,  612,  888,  241,  892,
 /*   200 */   920,  356,  418,  118,   60,   61,  567,  891,  890,  889,
 /*   210 */    14,  380,   62, 1178,  921,  921,  919,  918,  921,  893,
 /*   220 */     2, 1178,  894,  432,  687,  880, 1178,  726,  727,  559,
 /*   230 */   562,  921,  921,  889,  918,  543,  540,  547,  889,  892,
 /*   240 */   921,  921,  919,  918,  892,  738,  880,  891,  890,  889,
 /*   250 */   920, 1178,  891,  890,  889,  783,  420,  509,  206,  837,
 /*   260 */   737,  658,   88,  504, 1178,  676, 1178, 1178,  565,   86,
 /*   270 */   782,  613,  860,  325,  477,  335, 1178, 1178, 1178, 1178,
 /*   280 */   280, 1178,  880,  300,  515,  880,  193,  868,   12,  647,
 /*   290 */   870, 1025,  659,  852,  661,  452,   63,  660,  120, 1178,
 /*   300 */   920,  494,  927,  610,  921,  921,  180,  132,  358,  499,
 /*   310 */   353,  498,  171,   60,   61,  798,  112,  920,   22,  461,
 /*   320 */   177,   62, 1163,  611,  443,  687,  351,  921,  767,    2,
 /*   330 */  1163,  921,  808,   65,  880, 1163,  711,  876,  559,  791,
 /*   340 */   921,  921,  889,  918,   85,  876,  899,   97,  892,  843,
 /*   350 */   920,  541,  825,  808,  832,  880,  891,  890,  889,  808,
 /*   360 */  1163,  642,  658,   88,   70,   70,   69,   69,   69,   68,
 /*   370 */    66,  284,  293, 1163,  698, 1163, 1163,   85,  887,  832,
 /*   380 */   172, 1271, 1271,  285,  742, 1163,  771, 1163, 1163,  556,
 /*   390 */  1163,  880,  571,  659,  880,  661,  868,   12,  660,  444,
 /*   400 */   860,  888,  231,  538,  803,    4,  463,  741, 1163,  276,
 /*   410 */   567,  832, 1291,  921,  656,  760,  760, 1276,  921,   49,
 /*   420 */   462,   74,   75,  563,  305, 1271, 1271,  872,  872,   72,
 /*   430 */    72,   73,   73,   73,   73,  921,   71,   71,   71,   71,
 /*   440 */    70,   70,   69,   69,   69,   68,   66,  284,   96,  921,
 /*   450 */   921,  919,  918,  921,  921,  919,  918,  569,  176,   45,
 /*   460 */   145,  875,  436,  837,  879,  434, 1271, 1271,  884,  875,
 /*   470 */   127,   74,   75,  563,  305,   49,   51,  872,  872,   72,
 /*   480 */    72,   73,   73,   73,   73,    9,   71,   71,   71,   71,
 /*   490 */    70,   70,   69,   69,   69,   68,   66,  284,  534,  169,
 /*   500 */   879,  810,  493,  490,  489,  790,   74,   75,  563,  305,
 /*   510 */  1271, 1271,  872,  872,   72,   72,   73,   73,   73,   73,
 /*   520 */   488,   71,   71,   71,   71,   70,   70,   69,   69,   69,
 /*   530 */    68,   66,  284,  850,  299,  921,  921,  919,  918,  917,
 /*   540 */   921,  921,  919,  918,   23,  849,  915,   50,  896,  316,
 /*   550 */   893,  598,  318,  894,  916,  720,  711,  921,  921,  919,
 /*   560 */   918,  379,  374, 1271, 1271,  144,  697,   97,  349,  650,
 /*   570 */   920,  599,  598,  120,  651, 1276,  920, 1276,   73,   73,
 /*   580 */    73,   73,   46,   71,   71,   71,   71,   70,   70,   69,
 /*   590 */    69,   69,   68,   66,  284,  638,  881,  319,  304,  303,
 /*   600 */   456,  598, 1277,   74,   75,  563,  305, 1271, 1271,  872,
 /*   610 */   872,   72,   72,   73,   73,   73,   73,   79,   71,   71,
 /*   620 */    71,   71,   70,   70,   69,   69,   69,   68,   66,  284,
 /*   630 */   898,  388, 1394,   59,  921,   57,  288, 1394,  719,  496,
 /*   640 */   711,  862,  861,   78, 1271, 1271,  561,  865,  864,  921,
 /*   650 */   587,   97,  876,  876,  920,  851,   74,   75,  563,  305,
 /*   660 */   161,  163,  872,  872,   72,   72,   73,   73,   73,   73,
 /*   670 */   317,   71,   71,   71,   71,   70,   70,   69,   69,   69,
 /*   680 */    68,   66,  284,  178,   74,   75,  563,  305, 1271, 1271,
 /*   690 */   872,  872,   72,   72,   73,   73,   73,   73,  588,   71,
 /*   700 */    71,   71,   71,   70,   70,   69,   69,   69,   68,   66,
 /*   710 */   284,  921,  141,  876,  925,  287,   68,   66,  284,  575,
 /*   720 */   271,  270,  269,  562,  860,  523,   74,   75,  563,  305,
 /*   730 */   687,  921,  872,  872,   72,   72,   73,   73,   73,   73,
 /*   740 */   922,   71,   71,   71,   71,   70,   70,   69,   69,   69,
 /*   750 */    68,   66,  284,  307,  277,  170,  921,  921,  919,  918,
 /*   760 */  1277,  565, 1277,  125,  842,  595,   24,  620,   88,  524,
 /*   770 */  1285,  921,  921,  919,  918,  876,  875,  875,   85,  856,
 /*   780 */   588,  275,  763,  870,  512,  887,  595,  887,  874,   63,
 /*   790 */   465, 1285,  701,  274,  921,  731,  485, 1285,  619,  571,
 /*   800 */   618,  571,  702,  617,  161,  160,   60,   61,  888,  250,
 /*   810 */   888,  129,  921,  188,   62,  595,  469,  567,  437,  567,
 /*   820 */   669,  876,    2, 1283,    8,  315,  176,  880,  921,  578,
 /*   830 */   882,  559,  570,  921,  921,  889,  918,  875,  921,  920,
 /*   840 */   562,  892,  808,  532, 1285,  544,  568,  680,  880,  891,
 /*   850 */   890,  889, 1285,  921,  921,  919,  918, 1285,  921,  537,
 /*   860 */   402,  542,  887,  808, 1402, 1402,  351,   28,  680,  808,
 /*   870 */   397,  850,  837,  790,  921,  149,  571,  635,  565,  652,
 /*   880 */   653,  458, 1285,  849,  880,  888,  231,  880,  375,  868,
 /*   890 */    12,  531,  921,  777,  567, 1285,  921, 1285, 1285,  875,
 /*   900 */   870,  174,  920,  403,  562,  120,   63, 1285,    6, 1285,
 /*   910 */  1285,  846, 1285,  387,  636,  710,  921,  921,  919,  918,
 /*   920 */   616,  170,  169,   60,   61,  493,  490,  489,  921,  735,
 /*   930 */  1285,   62,  921,  871,  921,  921,  919,  918,  752,    2,
 /*   940 */   776,  772,  565,  488,  880,  875,  475,  837,  559,   77,
 /*   950 */   921,  921,  889,  918,  778,  193,   23,  717,  892,   85,
 /*   960 */   921,  921,  919,  918,  870,  880,  891,  890,  889,  920,
 /*   970 */    63,  769,  359,  754,  627,  779,  857,  560,  282,  528,
 /*   980 */   921,  921,  919,  918,  362,  813,  276,   60,   61,  486,
 /*   990 */   778,  304,  303,  455,  120,   62,  921,  921,  919,  918,
 /*  1000 */   553,  880,  840,    2,  880,  603,  868,   12,  880,  795,
 /*  1010 */   627,  879,  559,  921,  921,  921,  889,  918,  921,  921,
 /*  1020 */   919,  918,  892,  920,  638,  607,  402,  375,  821,  880,
 /*  1030 */   891,  890,  889,   64,  862,  861, 1342,  808,  609,    1,
 /*  1040 */   608,  920,  161,  159,  886, 1402, 1402,  879,  604,  821,
 /*  1050 */   921,  921,  919,  918,  921,  921,  919,  918,  808,  866,
 /*  1060 */   388, 1393,  921,  663,  808,  880, 1393,  540,  880,  554,
 /*  1070 */   868,   12,  505,  803,    4,  855,   74,   75,  563,  305,
 /*  1080 */   424,  920,  872,  872,   72,   72,   73,   73,   73,   73,
 /*  1090 */   213,   71,   71,   71,   71,   70,   70,   69,   69,   69,
 /*  1100 */    68,   66,  284,   74,   75,  563,  305,  128,   85,  872,
 /*  1110 */   872,   72,   72,   73,   73,   73,   73,  604,   71,   71,
 /*  1120 */    71,   71,   70,   70,   69,   69,   69,   68,   66,  284,
 /*  1130 */   921, 1342,  862,  861,  472,  921,  921,  919,  918,  177,
 /*  1140 */   854,  853,  562,  406,  267, 1327,   74,   75,  563,  305,
 /*  1150 */   791,  921,  872,  872,   72,   72,   73,   73,   73,   73,
 /*  1160 */   122,   71,   71,   71,   71,   70,   70,   69,   69,   69,
 /*  1170 */    68,   66,  284,  792,  832,  134,  802,  313,  689,  286,
 /*  1180 */   565,  375,  835,  824,  921,  921,  919,  918,  808,  312,
 /*  1190 */   535,  470,  384,  374,  921,  920,  177,  784,  887,  832,
 /*  1200 */   300,  515,  870,  212,  545,  530,  510,  920,   63,  808,
 /*  1210 */   902,  921,  571,  689,  536,  808,  841,  572,  138,  900,
 /*  1220 */    84,  888,  250,  381,  920,   60,   61,  921,  383,  183,
 /*  1230 */   567,  832,  427,   62,  114,  484,  815,   85,  327,  430,
 /*  1240 */   133,    2, 1283,   49,  128,   27,  880,  902,  141,  920,
 /*  1250 */   559,  921,  921,  921,  889,  918,  905,  887,  544,  511,
 /*  1260 */   892,  920,   81,  562,  450,  525,  532,  880,  891,  890,
 /*  1270 */   889,  571,  838,  921,  921,  919,  918,  385,  920,  199,
 /*  1280 */   888,  129,  795,  397,  836,  588,  844,  789,  795,  567,
 /*  1290 */   788,  361,   84,  557,  696,  840,  920, 1283,  887,  521,
 /*  1300 */  1280,  565,  920,  880,  751,  920,  880,  691,  868,   12,
 /*  1310 */   648,  368,  571,  367,  531,  921,  921,  921,  919,  918,
 /*  1320 */   399,  888,  250,  870,  921,  529,  634,  634,  548,   63,
 /*  1330 */   567,  198,  128,  921,  921,  919,  918,  128,  359,  754,
 /*  1340 */   921, 1278,  837,  816,  759,  128,   60,   61,   54,  921,
 /*  1350 */   921,  919,  918,  413,   62,  520,  392,  921,  442,  412,
 /*  1360 */   322,  640,    2,  795,  601,   49,  389,  880,  361,  822,
 /*  1370 */     4,  559,  921,  921,  921,  889,  918,  920,  359,  754,
 /*  1380 */   648,  892,  920,  551,  562,   53,  372,  532,  880,  891,
 /*  1390 */   890,  889,  834,  601, 1046,   74,   75,  563,  305,  699,
 /*  1400 */   186,  872,  872,   72,   72,   73,   73,   73,   73,  833,
 /*  1410 */    71,   71,   71,   71,   70,   70,   69,   69,   69,   68,
 /*  1420 */    66,  284,  565,  468,  880,   21,  181,  880,  177,  868,
 /*  1430 */    12,  643,  801,  392,  290,  533,  432,  921,  921,  919,
 /*  1440 */   918,  539,  339,  795,  870,  788,  921,  921,  919,  918,
 /*  1450 */    63, 1445,  203,  576,    3,  693,  693,  920,  744,  374,
 /*  1460 */   743,  597,  921,  921,  919,  918,  921,   60,   61,  928,
 /*  1470 */     3,  474,  414,  920,  788,   62,  901,  411,  747,  921,
 /*  1480 */   921,  919,  918,    2,   65,  543,  920,  895,  880,  480,
 /*  1490 */   597,  920,  559,  344,  921,  921,  889,  918,  292,  643,
 /*  1500 */    49,  921,  892,  920,  679,  329,  743,  471,   94,  880,
 /*  1510 */   891,  890,  889,  121,  457,  927,   74,   75,  563,  305,
 /*  1520 */    82,  920,  872,  872,   72,   72,   73,   73,   73,   73,
 /*  1530 */   823,   71,   71,   71,   71,   70,   70,   69,   69,   69,
 /*  1540 */    68,   66,  284,  128,  205,  880,  716,  715,  880,  330,
 /*  1550 */   868,   12,  844,  807,    4,   20,   74,   75,  563,  305,
 /*  1560 */   806,    4,  872,  872,   72,   72,   73,   73,   73,   73,
 /*  1570 */   365,   71,   71,   71,   71,   70,   70,   69,   69,   69,
 /*  1580 */    68,   66,  284,  449,  903,  818,  291,  633,  921,  921,
 /*  1590 */   919,  918,  758,   74,   75,  563,  305,  920,  921,  872,
 /*  1600 */   872,   72,   72,   73,   73,   73,   73,  910,   71,   71,
 /*  1610 */    71,   71,   70,   70,   69,   69,   69,   68,   66,  284,
 /*  1620 */   817,   55,  812,  921,  921,  919,  918,  804,    4,  753,
 /*  1630 */   276,   74,   75,  563,  305,  805,    4,  872,  872,   72,
 /*  1640 */    72,   73,   73,   73,   73,  791,   71,   71,   71,   71,
 /*  1650 */    70,   70,   69,   69,   69,   68,   66,  284,    5,  811,
 /*  1660 */   791,  809,   74,   75,  563,  305,   19,  750,  872,  872,
 /*  1670 */    72,   72,   73,   73,   73,   73,  439,   71,   71,   71,
 /*  1680 */    71,   70,   70,   69,   69,   69,   68,   66,  284,  119,
 /*  1690 */   920,  791,  781,   74,   75,  563,  305,  504,  733,  872,
 /*  1700 */   872,   72,   72,   73,   73,   73,   73,  780,   71,   71,
 /*  1710 */    71,   71,   70,   70,   69,   69,   69,   68,   66,  284,
 /*  1720 */   921,  921,  919,  918,   74,   75,  563,  305,   18,  200,
 /*  1730 */   872,  872,   72,   72,   73,   73,   73,   73,  207,   71,
 /*  1740 */    71,   71,   71,   70,   70,   69,   69,   69,   68,   66,
 /*  1750 */   284,  364,   17,   11,  175,   74,   75,  563,  305,  785,
 /*  1760 */   348,  872,  872,   72,   72,   73,   73,   73,   73,  196,
 /*  1770 */    71,   71,   71,   71,   70,   70,   69,   69,   69,   68,
 /*  1780 */    66,  284,  646, 1306,   74,   75,  563,  305,  195,  194,
 /*  1790 */   872,  872,   72,   72,   73,   73,   73,   73,  584,   71,
 /*  1800 */    71,   71,   71,   70,   70,   69,   69,   69,   68,   66,
 /*  1810 */   284,  276,  920,  921,  172,  773,   74,   75,  563,  305,
 /*  1820 */   135,   33,  872,  872,   72,   72,   73,   73,   73,   73,
 /*  1830 */   516,   71,   71,   71,   71,   70,   70,   69,   69,   69,
 /*  1840 */    68,   66,  284,  768, 1357,   74,   75,  563,  305,   16,
 /*  1850 */   921,  872,  872,   72,   72,   73,   73,   73,   73,  766,
 /*  1860 */    71,   71,   71,   71,   70,   70,   69,   69,   69,   68,
 /*  1870 */    66,  284,  762, 1355,   74,   75,  563,  305,  761,  508,
 /*  1880 */   872,  872,   72,   72,   73,   73,   73,   73,  583,   71,
 /*  1890 */    71,   71,   71,   70,   70,   69,   69,   69,   68,   66,
 /*  1900 */   284,  507,  920,  506,   43,   74,   75,  563,  305,  756,
 /*  1910 */   309,  872,  872,   72,   72,   73,   73,   73,   73,  908,
 /*  1920 */    71,   71,   71,   71,   70,   70,   69,   69,   69,   68,
 /*  1930 */    66,  284,  921,  503,   42,  921,  921,  919,  918,   15,
 /*  1940 */   126,  308,  191,  295,  562,   74,   75,  563,  305,   26,
 /*  1950 */   723,  872,  872,   72,   72,   73,   73,   73,   73,  495,
 /*  1960 */    71,   71,   71,   71,   70,   70,   69,   69,   69,   68,
 /*  1970 */    66,  284,  921,  921,  919,  918,   41,   74,   44,  563,
 /*  1980 */   305,  714,  565,  872,  872,   72,   72,   73,   73,   73,
 /*  1990 */    73,  492,   71,   71,   71,   71,   70,   70,   69,   69,
 /*  2000 */    69,   68,   66,  284,  870,  921,  923,  911,   91,   90,
 /*  2010 */    63,  374,  447,  487,  350,  887,  629,   25,  909,  907,
 /*  2020 */   920,  920,  374,  878,   89,  920,  920,   60,   61,  571,
 /*  2030 */   168,  356,  920,  920,  167,   62,  920,  471,  888,  253,
 /*  2040 */   482,  692,  164,    2,  904,  446,  897,  567,  880,  625,
 /*  2050 */   417,  920,  559,  921,  921,  921,  889,  918,  920,  920,
 /*  2060 */   920,  416,  892,  445,  441,  562,   10,  622,  398,  880,
 /*  2070 */   891,  890,  889,  210,   75,  563,  305,  920,  920,  872,
 /*  2080 */   872,   72,   72,   73,   73,   73,   73,  921,   71,   71,
 /*  2090 */    71,   71,   70,   70,   69,   69,   69,   68,   66,  284,
 /*  2100 */   837,  166,  110,  565,  844,  880,  162,  802,  880,  546,
 /*  2110 */   868,   12,  887,  108,  921,  844,  289,  107,  885,  887,
 /*  2120 */   550,  624,   40,  887,   39,  870,  571,  921,  921,  919,
 /*  2130 */   918,   63,  920,  571,   38,  888,  250,  571,  887,   92,
 /*  2140 */   921,   37,  888,  250,  567, 1274,  888,  251,   60,   61,
 /*  2150 */   410,  567,  571,  361,  877,  567,   62,  408,  128,  668,
 /*  2160 */   106,  888,  250,  326,    2,  662,  657,  920,  920,  880,
 /*  2170 */   567,  624,  429,  559,  921,  921,  921,  889,  918,  544,
 /*  2180 */   522,  797,  655,  892,  921,  459,  562,  115,  177,  793,
 /*  2190 */   880,  891,  890,  889,  649,  920,  157,  397,  544,  422,
 /*  2200 */   787,  173,  839,  920,  397,  519,  324,  755,  837,  921,
 /*  2210 */   921,  919,  918,  352,  920,  920,  799,   13,  296,  920,
 /*  2220 */   321,  920,  189,  397,  565,   77,  880,  920,  730,  880,
 /*  2230 */   778,  868,   12,  708,  346,  921,  921,  921,  919,  918,
 /*  2240 */   214,  694,  920,  341,  476,  686,  870,  920,  920,  155,
 /*  2250 */   630,  779,   63,  683,  101,  920,  621,  920,  920,  920,
 /*  2260 */   887,  154,  921,  921,  919,  918,  778,  920,  453,   60,
 /*  2270 */    61,  153,  152,  593,  571,  921,  150,   62, 1120, 1120,
 /*  2280 */   606,  147,  920,  888,  231,    2,  732,  920,  605,  438,
 /*  2290 */   880,  921,  567,   30,  559,  921,  921,  921,  919,  918,
 /*  2300 */   435,   29,  306,  562,  892,  433,  921,  921,  919,  918,
 /*  2310 */   431,  880,  891,  890,  889,  563,  305,  581,  921,  872,
 /*  2320 */   872,   72,   72,   73,   73,   73,   73,  621,   71,   71,
 /*  2330 */    71,   71,   70,   70,   69,   69,   69,   68,   66,  284,
 /*  2340 */   577,  565,  440,  272,  268,  837,   99,  880,  775,  294,
 /*  2350 */   880,  674,  868,   12,  591,  921,  920,  921,  921,  919,
 /*  2360 */   918,  404,  378,  870,  921,  276,  924,  562,  920,   63,
 /*  2370 */   566,  185,  674,  123,  912,  920,  920,  401,  400,  342,
 /*  2380 */   363,  473,  323,  810,  920,  914,   60,   61,  921,  483,
 /*  2390 */   320,  666,  921,  920,   62,  920,  920,  921,  921,  919,
 /*  2400 */   918,  913,    2,  347,  920,  565,  117,  880,  184,  314,
 /*  2410 */   310,  559,  666,  921,  921,  889,  918,  921,  921,  919,
 /*  2420 */   918,  892,  887,  920,  920,  311,  921,  870,  880,  891,
 /*  2430 */   890,  889,  574,   63,  631,  182,  571,  573,  562,  920,
 /*  2440 */   921,  921,  919,  918,  113,  888,  250,   80,  179,  883,
 /*  2450 */    60,   61,  284,   93,  567,  631,  281,  373,   62,  555,
 /*  2460 */   302,  426,   56,   24,  880,  301,    2,  880,  675,  868,
 /*  2470 */    12,  880,  921,  370,  527,  559,  565,  921,  921,  889,
 /*  2480 */   918,  369,  298,  208,   52,  892,  921,  921,  919,  918,
 /*  2490 */   921,   48,  880,  891,  890,  889,  543,  831,  870,  366,
 /*  2500 */   425,  830,  562,  273,   63,  829,  192,  397,  820,  143,
 /*  2510 */   921,  921,  919,  918,  921,  921,  919,  918,  814,  826,
 /*  2520 */   514,   60,   61,  515,  757,  745,  921,  707,  880,   62,
 /*  2530 */   786,  880,  172,  868,   12, 1281,  921,    2, 1283,   85,
 /*  2540 */   565,  500,  880,  740,  739,  419,  559,  354,  921,  921,
 /*  2550 */   889,  918,  774,  736,  735,  391,  892,  887,  734,  713,
 /*  2560 */   390,  921,  870,  880,  891,  890,  889,  111,   63,  706,
 /*  2570 */   705,  571,  704,  562,  481,  703,  142,  279,  165,  340,
 /*  2580 */   888,  250,  460,   77, 1113,   60,   61,  478,  778,  567,
 /*  2590 */    36,  338,  681,   62,  921,  921,  919,  918,  337,  880,
 /*  2600 */   921,    2,  880,  215,  868,   12,  880,  921,  887,  779,
 /*  2610 */   559,  565,  921,  921,  889,  918,  336,  297,  109,  921,
 /*  2620 */   892,  887,  571,  887,  778, 1402, 1402,  880,  891,  890,
 /*  2630 */   889,  888,  129,  870,  190,  571,  333,  571,  921,   63,
 /*  2640 */   567,  695,  397,  332,  888,  129,  888,  129,  921,  921,
 /*  2650 */   919,  918,  579,  567,  331,  567,   60,   61,  921,  921,
 /*  2660 */   919,  918,  921,  880,   62,  677,  880,  204,  868,   12,
 /*  2670 */   116,  328,    2,  579,  921,  467,  671,  880,  921,  466,
 /*  2680 */   667,  559,  197,  921,  921,  919,  918,  921,  887,  665,
 /*  2690 */   140,  892,  158,  837,  645,  421,  139,  409,  880,  891,
 /*  2700 */   890,  889,  571,  921,  887,  454,  837,  921,  837,  156,
 /*  2710 */   451,  888,  250,  906,  105,  448,  921,  615,  571,  104,
 /*  2720 */   567,  921,  921,  921,  919,  918,  921,  888,  129,  921,
 /*  2730 */   921,  919,  918,  103,  880,  887,  567,  880,  102,  868,
 /*  2740 */    12,  921,  921,  919,  918,   35,  729,  921,  423,  571,
 /*  2750 */   151,  921,  405,   34,  602,   31,  921,  137,  888,  250,
 /*  2760 */   921,  921,  919,  918,  921,  600,  596,  567,  136,  586,
 /*  2770 */   585,  588,  921,  397,  146,  700,  582,  921,  407,  863,
 /*  2780 */   552,   67,  921,  859,  921,  921,  919,  918,  202,  837,
 /*  2790 */   921,  848,  796,  216,  371,  526,  921,  921,  919,  918,
 /*  2800 */   921,  921,  919,  918,  201,  921,  800,  355,  731,  921,
 /*  2810 */   921,  919,  918,  690,  921,  654,  491,  264,  921,  334,
 /*  2820 */   397,  670,  148,  921,  684,  921,  921,  919,  918,  921,
 /*  2830 */   921,  919,  918,  921,  672,  187,  386,    7,  921,  921,
 /*  2840 */   919,  918,  873,  921,  921,  919,  918,  858,  921,  921,
 /*  2850 */   919,  918,  847,  678,  921,  278,  718,  479,  673,  921,
 /*  2860 */   688,   87,  209,  685,  921,   47,  641,  682,  564,  921,
 /*  2870 */   921,  919,  918,  921,  921,  919,  918,  921,  921,  921,
 /*  2880 */   919,  918,  100,  728,  644,  819,  921,  921,  919,  918,
 /*  2890 */   722,  721,  626,  632,  921,  921,  919,  918,  639,  921,
 /*  2900 */   921,  919,  918,  749,  921,  921,  919,  918,  748,  709,
 /*  2910 */    85, 1446,  921,  921,  919,  918,  628, 1446, 1446,  594,
 /*  2920 */   592, 1446, 1446, 1446, 1446, 1446, 1446,  921,  921,  919,
 /*  2930 */   918, 1446, 1446, 1446, 1446,  590,  921,  921,  919,  918,
 /*  2940 */   921,  921,  919,  918, 1446,  921,  921,  919,  918, 1446,
 /*  2950 */  1446, 1446, 1446, 1446, 1446,  921,  921,  919,  918,  623,
 /*  2960 */  1446, 1446,  614,  589, 1446, 1446, 1446, 1446, 1446, 1446,
 /*  2970 */  1446, 1446, 1446, 1446,  839, 1446,  921,  921,  919,  918,
 /*  2980 */  1446,  921,  921,  919,  918,  710,  921,  921,  919,  918,
 /*  2990 */   887,  580, 1446,  887, 1446,  887,  845, 1446, 1446,  921,
 /*  3000 */   921,  919,  918, 1446,  571, 1446,  794,  571, 1446,  571,
 /*  3010 */  1446, 1446, 1446,  888,  261, 1446,  888,  260,  888,  262,
 /*  3020 */   887, 1446,  567, 1446, 1446,  567,  887,  567,   35, 1446,
 /*  3030 */  1446,  887, 1446,  887,  571, 1446, 1446,  887, 1446, 1446,
 /*  3040 */   571, 1446, 1446,  888,  263,  571, 1446,  571, 1446,  888,
 /*  3050 */   283,  571,  567, 1446,  888,  377,  888,  259,  567,  887,
 /*  3060 */   888,  376, 1446,  567, 1446,  567, 1446, 1446, 1446,  567,
 /*  3070 */   887, 1446,  887,  571, 1446,  837,  887, 1446,  837, 1446,
 /*  3080 */   837, 1446,  888,  227,  571, 1446,  571, 1446, 1446, 1446,
 /*  3090 */   571,  567, 1446,  888,  247,  888,  252, 1446, 1446,  888,
 /*  3100 */   395,  887,  567,  887,  567,  837,  887, 1446,  567, 1446,
 /*  3110 */  1446,  837, 1446, 1446, 1446,  571,  837,  571,  837, 1446,
 /*  3120 */   571, 1446,  837, 1446,  888,  394,  888,  393, 1446,  888,
 /*  3130 */   246, 1446, 1446,  567, 1446,  567,  887, 1446,  567, 1446,
 /*  3140 */  1446, 1446,  887, 1446,  837, 1446, 1446, 1446,  887, 1446,
 /*  3150 */   571, 1446, 1446,  887, 1446,  837,  571,  837, 1446,  888,
 /*  3160 */   229,  837,  571, 1446, 1446,  888,  245,  571,  567,  887,
 /*  3170 */  1446,  888,  244, 1446,  567, 1446,  888,  228, 1446, 1446,
 /*  3180 */   567, 1446,  887,  571,  887,  567,  837, 1446,  837, 1446,
 /*  3190 */  1446,  837,  888,  258, 1446, 1446,  571, 1446,  571,  887,
 /*  3200 */  1446,  567, 1446, 1446, 1446,  888,  257,  888,  217, 1446,
 /*  3210 */  1446, 1446,  887,  571,  567, 1446,  567, 1446, 1446, 1446,
 /*  3220 */  1446,  837,  888,  226, 1446,  887,  571,  837,  887, 1446,
 /*  3230 */  1446,  567,  887,  837, 1446,  888,  225, 1446,  837,  571,
 /*  3240 */  1446, 1446,  571, 1446,  567, 1446,  571, 1446,  888,  130,
 /*  3250 */  1446,  888,  224, 1446,  837,  888,  254,  567,  887, 1446,
 /*  3260 */   567, 1446,  887, 1446,  567, 1446, 1446,  837, 1446,  837,
 /*  3270 */  1446, 1446,  571, 1446, 1446,  887,  571, 1446, 1446,  887,
 /*  3280 */  1446,  888,  256, 1446,  837,  888,  249, 1446, 1446,  571,
 /*  3290 */   567, 1446,  887,  571,  567, 1446, 1446,  837,  888,  255,
 /*  3300 */  1446, 1446,  888,  243, 1446,  887,  571,  567,  887, 1446,
 /*  3310 */   837,  567,  887,  837, 1446,  888,  242,  837, 1446,  571,
 /*  3320 */  1446, 1446,  571, 1446,  567, 1446,  571, 1446,  888,  239,
 /*  3330 */  1446,  888,  221, 1446, 1446,  888,  220,  567, 1446,  887,
 /*  3340 */   567, 1446,  887,  837,  567, 1446, 1446,  837, 1446, 1446,
 /*  3350 */  1446, 1446, 1446,  571, 1446, 1446,  571, 1446, 1446,  887,
 /*  3360 */   837, 1446,  888,  219,  837,  888,  218, 1446, 1446, 1446,
 /*  3370 */  1446,  567,  887,  571,  567,  887, 1446,  837,  887, 1446,
 /*  3380 */  1446, 1446,  888,  230, 1446, 1446,  571, 1446, 1446,  571,
 /*  3390 */   837,  567,  571,  837,  887,  888,  237,  837,  888,  236,
 /*  3400 */  1446,  888,  131, 1446,  567,  887, 1446,  567,  571, 1446,
 /*  3410 */   567,  887, 1446, 1446, 1446, 1446, 1446,  888,  235,  571,
 /*  3420 */  1446, 1446,  887, 1446,  837,  571,  567,  837,  888,  234,
 /*  3430 */  1446, 1446, 1446, 1446,  888,  233,  571,  567,  887, 1446,
 /*  3440 */  1446,  887, 1446,  567,  837,  888,  232, 1446, 1446, 1446,
 /*  3450 */  1446, 1446,  571, 1446,  567,  571, 1446,  837, 1446, 1446,
 /*  3460 */   837,  888,  248,  837,  888,  222, 1446, 1446,  887, 1446,
 /*  3470 */   567, 1446, 1446,  567, 1446, 1446, 1446, 1446, 1446,  837,
 /*  3480 */  1446,  887,  571, 1446, 1446,  887, 1446,  887, 1446, 1446,
 /*  3490 */   837,  888,  240, 1446, 1446,  571,  837, 1446, 1446,  571,
 /*  3500 */   567,  571, 1446, 1446,  888,  238, 1446,  837,  888,  223,
 /*  3510 */   888,   98, 1446,  567, 1446, 1446, 1446,  567, 1446,  518,
 /*  3520 */  1446, 1446, 1446,  837, 1446, 1446,  837, 1446, 1446, 1446,
 /*  3530 */  1446, 1446, 1446, 1446, 1446, 1446, 1446, 1446, 1446, 1446,
 /*  3540 */  1446, 1446, 1446, 1446, 1446, 1446, 1446, 1446, 1446, 1446,
 /*  3550 */  1446, 1446, 1446,  837, 1446, 1446, 1446, 1446, 1446, 1446,
 /*  3560 */  1446, 1446, 1446, 1446, 1446, 1446,  837, 1446, 1446, 1446,
 /*  3570 */   837, 1446,  837,
};
static const YYCODETYPE yy_lookahead[] = {
 /*     0 */     4,  217,    4,  213,   48,  221,  222,  223,   80,   81,
 /*    10 */    82,   83,   16,  116,   86,   87,   88,   89,   90,   91,
 /*    20 */    92,   93,  108,   95,   96,   97,   98,   99,  100,  101,
 /*    30 */   102,  103,  104,  105,  106,  116,   80,   81,   82,   83,
 /*    40 */   121,  251,   86,   87,   88,   89,   90,   91,   92,   93,
 /*    50 */    54,   95,   96,   97,   98,   99,  100,  101,  102,  103,
 /*    60 */   104,  105,  106,  115,  108,  117,   15,  121,  171,    4,
 /*    70 */   252,  253,   76,  255,   90,   91,   92,   93,   82,   95,
 /*    80 */    96,   97,   98,   99,  100,  101,  102,  103,  104,  105,
 /*    90 */   106,   79,    4,  116,   43,   99,  100,  101,  121,  224,
 /*   100 */   225,  226,  112,  107,  108,  159,   31,  161,  162,    4,
 /*   110 */    59,  115,  116,  185,   39,  117,  120,  121,    4,   24,
 /*   120 */   124,   16,  126,  127,  128,  129,  128,  115,  116,   24,
 /*   130 */   134,  101,  102,  103,  104,  105,  106,  141,  142,  143,
 /*   140 */   144,   76,  146,   95,   96,   97,   98,   99,  100,  101,
 /*   150 */   102,  103,  104,  105,  106,  159,   82,  161,  162,   54,
 /*   160 */   101,  165,  164,   66,   99,  100,   69,  171,  172,  173,
 /*   170 */   174,  175,  176,  177,  205,  205,  180,  302,  182,  183,
 /*   180 */   115,   76,    4,  134,   58,  121,  205,   82,  219,  219,
 /*   190 */   194,  126,  127,  128,    0,  225,  226,  228,  229,  134,
 /*   200 */   219,  152,  151,  115,   99,  100,  237,  142,  143,  144,
 /*   210 */   115,  114,  107,  108,  126,  127,  128,  129,    4,  160,
 /*   220 */   115,  116,  163,   29,  213,  120,  121,  133,  134,  124,
 /*   230 */    16,  126,  127,  128,  129,  171,  205,  163,  128,  134,
 /*   240 */   126,  127,  128,  129,  134,   21,  141,  142,  143,  144,
 /*   250 */   219,  146,  142,  143,  144,  141,   82,   72,  277,  290,
 /*   260 */    36,  250,  251,   78,  159,  177,  161,  162,   54,  116,
 /*   270 */   156,  301,  302,  147,  148,  149,  171,  172,  173,  174,
 /*   280 */   106,  176,  177,  157,  158,  180,  205,  182,  183,   64,
 /*   290 */    76,  117,  281,  116,  283,   70,   82,  286,  121,  194,
 /*   300 */   219,  127,  108,  205,  126,  127,  132,  133,  134,  135,
 /*   310 */   136,  137,  138,   99,  100,  116,  121,  219,  307,  308,
 /*   320 */   121,  107,  108,  225,  226,  213,  152,    4,  116,  115,
 /*   330 */   116,    4,   34,  121,  120,  121,  205,    4,  124,  221,
 /*   340 */   126,  127,  128,  129,   71,    4,  215,  216,  134,   42,
 /*   350 */   219,  320,  321,   55,    5,  141,  142,  143,  144,   61,
 /*   360 */   146,  136,  250,  251,   99,  100,  101,  102,  103,  104,
 /*   370 */   105,  106,  241,  159,  243,  161,  162,   71,  205,   30,
 /*   380 */   195,   40,   41,  115,   35,  171,  268,  173,  174,   82,
 /*   390 */   176,  177,  219,  281,  180,  283,  182,  183,  286,  301,
 /*   400 */   302,  228,  229,  322,  323,  324,  133,   58,  194,  291,
 /*   410 */   237,   62,  187,    4,  108,   40,   41,   26,    4,  121,
 /*   420 */   308,   80,   81,   82,   83,   84,   85,   86,   87,   88,
 /*   430 */    89,   90,   91,   92,   93,    4,   95,   96,   97,   98,
 /*   440 */    99,  100,  101,  102,  103,  104,  105,  106,  115,  126,
 /*   450 */   127,  128,  129,  126,  127,  128,  129,  121,  117,   84,
 /*   460 */    85,  128,  151,  290,  141,  154,   40,   41,  141,  128,
 /*   470 */   116,   80,   81,   82,   83,  121,   50,   86,   87,   88,
 /*   480 */    89,   90,   91,   92,   93,  115,   95,   96,   97,   98,
 /*   490 */    99,  100,  101,  102,  103,  104,  105,  106,  325,  132,
 /*   500 */   177,  328,  135,  136,  137,  164,   80,   81,   82,   83,
 /*   510 */    84,   85,   86,   87,   88,   89,   90,   91,   92,   93,
 /*   520 */   153,   95,   96,   97,   98,   99,  100,  101,  102,  103,
 /*   530 */   104,  105,  106,   10,  257,  126,  127,  128,  129,   21,
 /*   540 */   126,  127,  128,  129,  267,   22,   28,  121,  139,   66,
 /*   550 */   160,  120,   69,  163,   36,  141,  205,  126,  127,  128,
 /*   560 */   129,  210,  205,   40,   41,  118,  215,  216,  116,    7,
 /*   570 */   219,  140,  141,  121,   12,  184,  219,  186,   90,   91,
 /*   580 */    92,   93,   94,   95,   96,   97,   98,   99,  100,  101,
 /*   590 */   102,  103,  104,  105,  106,   79,  116,  114,   99,  100,
 /*   600 */    38,  170,   26,   80,   81,   82,   83,   84,   85,   86,
 /*   610 */    87,   88,   89,   90,   91,   92,   93,   89,   95,   96,
 /*   620 */    97,   98,   99,  100,  101,  102,  103,  104,  105,  106,
 /*   630 */   227,  115,  116,  184,    4,  186,   75,  121,  235,  236,
 /*   640 */   205,  142,  143,  115,   40,   41,  224,  225,  226,    4,
 /*   650 */   215,  216,    4,    4,  219,  298,   80,   81,   82,   83,
 /*   660 */   117,  118,   86,   87,   88,   89,   90,   91,   92,   93,
 /*   670 */   187,   95,   96,   97,   98,   99,  100,  101,  102,  103,
 /*   680 */   104,  105,  106,  118,   80,   81,   82,   83,   84,   85,
 /*   690 */    86,   87,   88,   89,   90,   91,   92,   93,   17,   95,
 /*   700 */    96,   97,   98,   99,  100,  101,  102,  103,  104,  105,
 /*   710 */   106,    4,  121,    4,  197,  154,  104,  105,  106,  202,
 /*   720 */   147,  148,  149,   16,  302,   82,   80,   81,   82,   83,
 /*   730 */   213,    4,   86,   87,   88,   89,   90,   91,   92,   93,
 /*   740 */   110,   95,   96,   97,   98,   99,  100,  101,  102,  103,
 /*   750 */   104,  105,  106,  192,  193,   41,  126,  127,  128,  129,
 /*   760 */   184,   54,  186,  117,   42,  120,  175,  250,  251,  126,
 /*   770 */    34,  126,  127,  128,  129,    4,  128,  128,   71,  130,
 /*   780 */    17,  252,  253,   76,  255,  205,  141,  205,  140,   82,
 /*   790 */   149,   55,  235,  115,    4,  238,   82,   61,  281,  219,
 /*   800 */   283,  219,  245,  286,  117,  118,   99,  100,  228,  229,
 /*   810 */   228,  229,    4,  132,  107,  170,  287,  237,  137,  237,
 /*   820 */   179,    4,  115,  116,  115,  205,  117,  120,    4,  312,
 /*   830 */   253,  124,  255,  126,  127,  128,  129,  128,    4,  219,
 /*   840 */    16,  134,   34,   19,  108,  265,  266,  120,  141,  142,
 /*   850 */   143,  144,  116,  126,  127,  128,  129,  121,    4,   51,
 /*   860 */   146,  279,  205,   55,  157,  158,  152,  121,  141,   61,
 /*   870 */   290,   10,  290,  164,    4,  112,  219,   79,   54,  147,
 /*   880 */   148,  149,  146,   22,  177,  228,  229,  180,  205,  182,
 /*   890 */   183,   67,    4,  121,  237,  159,    4,  161,  162,  128,
 /*   900 */    76,  129,  219,  207,   16,  121,   82,  171,   84,  173,
 /*   910 */   174,  140,  176,  115,  116,  125,  126,  127,  128,  129,
 /*   920 */   300,   41,  132,   99,  100,  135,  136,  137,    4,  145,
 /*   930 */   194,  107,    4,  134,  126,  127,  128,  129,  296,  115,
 /*   940 */   168,  169,   54,  153,  120,  128,  257,  290,  124,  115,
 /*   950 */   126,  127,  128,  129,  120,  205,  267,  140,  134,   71,
 /*   960 */   126,  127,  128,  129,   76,  141,  142,  143,  144,  219,
 /*   970 */    82,  101,  330,  331,  120,  141,  221,  222,  223,  171,
 /*   980 */   126,  127,  128,  129,  327,  328,  291,   99,  100,  116,
 /*   990 */   156,   99,  100,  297,  121,  107,  126,  127,  128,  129,
 /*  1000 */   317,  177,  319,  115,  180,  310,  182,  183,  120,  205,
 /*  1010 */   156,  141,  124,    4,  126,  127,  128,  129,  126,  127,
 /*  1020 */   128,  129,  134,  219,   79,  133,  146,  205,   32,  141,
 /*  1030 */   142,  143,  144,  115,  142,  143,   13,   34,  146,  115,
 /*  1040 */   148,  219,  117,  118,  120,  157,  158,  177,  120,   53,
 /*  1050 */   126,  127,  128,  129,  126,  127,  128,  129,   55,  116,
 /*  1060 */   115,  116,    4,   26,   61,  177,  121,  205,  180,  213,
 /*  1070 */   182,  183,  322,  323,  324,  116,   80,   81,   82,   83,
 /*  1080 */   276,  219,   86,   87,   88,   89,   90,   91,   92,   93,
 /*  1090 */   115,   95,   96,   97,   98,   99,  100,  101,  102,  103,
 /*  1100 */   104,  105,  106,   80,   81,   82,   83,  251,   71,   86,
 /*  1110 */    87,   88,   89,   90,   91,   92,   93,  189,   95,   96,
 /*  1120 */    97,   98,   99,  100,  101,  102,  103,  104,  105,  106,
 /*  1130 */     4,  108,  142,  143,  116,  126,  127,  128,  129,  121,
 /*  1140 */   116,  319,   16,  313,  314,  108,   80,   81,   82,   83,
 /*  1150 */   221,    4,   86,   87,   88,   89,   90,   91,   92,   93,
 /*  1160 */   121,   95,   96,   97,   98,   99,  100,  101,  102,  103,
 /*  1170 */   104,  105,  106,  164,    5,    9,  121,   11,  120,   13,
 /*  1180 */    54,  205,  116,  321,  126,  127,  128,  129,   34,   23,
 /*  1190 */    19,  116,   26,  205,    4,  219,  121,  268,  205,   30,
 /*  1200 */   157,  158,   76,  115,   35,   51,  213,  219,   82,   55,
 /*  1210 */   205,    4,  219,  155,   43,   61,  116,  212,   52,  214,
 /*  1220 */   165,  228,  229,   57,  219,   99,  100,    4,   62,   63,
 /*  1230 */   237,   62,  244,  107,   68,  205,   65,   71,  116,  209,
 /*  1240 */    74,  115,  116,  121,  251,  146,  120,  205,  121,  219,
 /*  1250 */   124,    4,  126,  127,  128,  129,  214,  205,  265,  266,
 /*  1260 */   134,  219,  115,   16,  205,  166,   19,  141,  142,  143,
 /*  1270 */   144,  219,   26,  126,  127,  128,  129,  111,  219,  113,
 /*  1280 */   228,  229,  205,  290,  116,   17,  298,  275,  205,  237,
 /*  1290 */   278,  205,  165,  317,  213,  319,  219,  171,  205,  213,
 /*  1300 */   184,   54,  219,  177,  296,  219,  180,  213,  182,  183,
 /*  1310 */   120,  262,  219,  264,   67,    4,  126,  127,  128,  129,
 /*  1320 */   154,  228,  229,   76,    4,  171,  119,  120,  163,   82,
 /*  1330 */   237,  279,  251,  126,  127,  128,  129,  251,  330,  331,
 /*  1340 */     4,   26,  290,  172,  296,  251,   99,  100,  165,  126,
 /*  1350 */   127,  128,  129,  276,  107,  269,  270,    4,  192,  276,
 /*  1360 */   116,  116,  115,  205,  141,  121,  121,  120,  205,  323,
 /*  1370 */   324,  124,    4,  126,  127,  128,  129,  219,  330,  331,
 /*  1380 */   190,  134,  219,  290,   16,  165,  293,   19,  141,  142,
 /*  1390 */   143,  144,  116,  170,  116,   80,   81,   82,   83,  121,
 /*  1400 */   132,   86,   87,   88,   89,   90,   91,   92,   93,  116,
 /*  1410 */    95,   96,   97,   98,   99,  100,  101,  102,  103,  104,
 /*  1420 */   105,  106,   54,  116,  177,   14,   73,  180,  121,  182,
 /*  1430 */   183,  120,  269,  270,  276,   67,   29,  126,  127,  128,
 /*  1440 */   129,  117,  275,  205,   76,  278,  126,  127,  128,  129,
 /*  1450 */    82,  198,  199,  200,  201,  119,  120,  219,  246,  205,
 /*  1460 */   248,  141,  126,  127,  128,  129,    4,   99,  100,  200,
 /*  1470 */   201,  275,  205,  219,  278,  107,  123,  205,  219,  126,
 /*  1480 */   127,  128,  129,  115,  121,  171,  219,  228,  120,  205,
 /*  1490 */   170,  219,  124,  209,  126,  127,  128,  129,  244,  188,
 /*  1500 */   121,    4,  134,  219,  246,  213,  248,  205,  194,  141,
 /*  1510 */   142,  143,  144,  121,  276,  108,   80,   81,   82,   83,
 /*  1520 */   115,  219,   86,   87,   88,   89,   90,   91,   92,   93,
 /*  1530 */   116,   95,   96,   97,   98,   99,  100,  101,  102,  103,
 /*  1540 */   104,  105,  106,  251,  277,  177,  219,  220,  180,  277,
 /*  1550 */   182,  183,  298,  323,  324,   14,   80,   81,   82,   83,
 /*  1560 */   323,  324,   86,   87,   88,   89,   90,   91,   92,   93,
 /*  1570 */   231,   95,   96,   97,   98,   99,  100,  101,  102,  103,
 /*  1580 */   104,  105,  106,  205,  122,   45,  284,  209,  126,  127,
 /*  1590 */   128,  129,  116,   80,   81,   82,   83,  219,    4,   86,
 /*  1600 */    87,   88,   89,   90,   91,   92,   93,  110,   95,   96,
 /*  1610 */    97,   98,   99,  100,  101,  102,  103,  104,  105,  106,
 /*  1620 */    60,  185,   32,  126,  127,  128,  129,  323,  324,  116,
 /*  1630 */   291,   80,   81,   82,   83,  323,  324,   86,   87,   88,
 /*  1640 */    89,   90,   91,   92,   93,  221,   95,   96,   97,   98,
 /*  1650 */    99,  100,  101,  102,  103,  104,  105,  106,   81,   60,
 /*  1660 */   221,   53,   80,   81,   82,   83,   14,  116,   86,   87,
 /*  1670 */    88,   89,   90,   91,   92,   93,  205,   95,   96,   97,
 /*  1680 */    98,   99,  100,  101,  102,  103,  104,  105,  106,  115,
 /*  1690 */   219,  221,  268,   80,   81,   82,   83,   78,  116,   86,
 /*  1700 */    87,   88,   89,   90,   91,   92,   93,  268,   95,   96,
 /*  1710 */    97,   98,   99,  100,  101,  102,  103,  104,  105,  106,
 /*  1720 */   126,  127,  128,  129,   80,   81,   82,   83,   14,  116,
 /*  1730 */    86,   87,   88,   89,   90,   91,   92,   93,  268,   95,
 /*  1740 */    96,   97,   98,   99,  100,  101,  102,  103,  104,  105,
 /*  1750 */   106,  231,   14,   14,   14,   80,   81,   82,   83,  126,
 /*  1760 */   116,   86,   87,   88,   89,   90,   91,   92,   93,  116,
 /*  1770 */    95,   96,   97,   98,   99,  100,  101,  102,  103,  104,
 /*  1780 */   105,  106,  188,  108,   80,   81,   82,   83,  116,  116,
 /*  1790 */    86,   87,   88,   89,   90,   91,   92,   93,  205,   95,
 /*  1800 */    96,   97,   98,   99,  100,  101,  102,  103,  104,  105,
 /*  1810 */   106,  291,  219,    4,  195,  168,   80,   81,   82,   83,
 /*  1820 */   118,  117,   86,   87,   88,   89,   90,   91,   92,   93,
 /*  1830 */   118,   95,   96,   97,   98,   99,  100,  101,  102,  103,
 /*  1840 */   104,  105,  106,  101,  108,   80,   81,   82,   83,  115,
 /*  1850 */     4,   86,   87,   88,   89,   90,   91,   92,   93,  160,
 /*  1860 */    95,   96,   97,   98,   99,  100,  101,  102,  103,  104,
 /*  1870 */   105,  106,  116,  108,   80,   81,   82,   83,  116,  172,
 /*  1880 */    86,   87,   88,   89,   90,   91,   92,   93,  205,   95,
 /*  1890 */    96,   97,   98,   99,  100,  101,  102,  103,  104,  105,
 /*  1900 */   106,  115,  219,  171,   14,   80,   81,   82,   83,  116,
 /*  1910 */   116,   86,   87,   88,   89,   90,   91,   92,   93,  110,
 /*  1920 */    95,   96,   97,   98,   99,  100,  101,  102,  103,  104,
 /*  1930 */   105,  106,    4,  115,  175,  126,  127,  128,  129,   14,
 /*  1940 */   116,  116,   18,   48,   16,   80,   81,   82,   83,  115,
 /*  1950 */     6,   86,   87,   88,   89,   90,   91,   92,   93,  117,
 /*  1960 */    95,   96,   97,   98,   99,  100,  101,  102,  103,  104,
 /*  1970 */   105,  106,  126,  127,  128,  129,  115,   80,   81,   82,
 /*  1980 */    83,    8,   54,   86,   87,   88,   89,   90,   91,   92,
 /*  1990 */    93,   48,   95,   96,   97,   98,   99,  100,  101,  102,
 /*  2000 */   103,  104,  105,  106,   76,    4,  205,  205,  115,  115,
 /*  2010 */    82,  205,  205,   48,  116,  205,  209,  115,  205,  205,
 /*  2020 */   219,  219,  205,  177,  115,  219,  219,   99,  100,  219,
 /*  2030 */   138,  152,  219,  219,  118,  107,  219,  205,  228,  229,
 /*  2040 */    82,  124,  146,  115,  205,  205,  205,  237,  120,  209,
 /*  2050 */   244,  219,  124,    4,  126,  127,  128,  129,  219,  219,
 /*  2060 */   219,  244,  134,  205,  205,   16,   13,  209,  209,  141,
 /*  2070 */   142,  143,  144,  117,   81,   82,   83,  219,  219,   86,
 /*  2080 */    87,   88,   89,   90,   91,   92,   93,    4,   95,   96,
 /*  2090 */    97,   98,   99,  100,  101,  102,  103,  104,  105,  106,
 /*  2100 */   290,  118,  165,   54,  298,  177,  117,  121,  180,  181,
 /*  2110 */   182,  183,  205,  151,    4,  298,  284,  178,  205,  205,
 /*  2120 */   213,  120,   89,  205,   89,   76,  219,  126,  127,  128,
 /*  2130 */   129,   82,  219,  219,   89,  228,  229,  219,  205,   56,
 /*  2140 */     4,   89,  228,  229,  237,  108,  228,  229,   99,  100,
 /*  2150 */    18,  237,  219,  205,  205,  237,  107,   24,  251,  179,
 /*  2160 */   151,  228,  229,  158,  115,  108,  108,  219,  219,  120,
 /*  2170 */   237,  170,  265,  124,    4,  126,  127,  128,  129,  265,
 /*  2180 */   266,  205,   60,  134,    4,   25,   16,   49,  121,  205,
 /*  2190 */   141,  142,  143,  144,   49,  219,  146,  290,  265,  266,
 /*  2200 */   205,  205,  119,  219,  290,  205,  187,  205,  290,  126,
 /*  2210 */   127,  128,  129,  205,  219,  219,  167,  115,  270,  219,
 /*  2220 */   114,  219,  166,  290,   54,  115,  177,  219,  205,  180,
 /*  2230 */   120,  182,  183,  205,  205,    4,  126,  127,  128,  129,
 /*  2240 */   115,  205,  219,  205,  205,  205,   76,  219,  219,  118,
 /*  2250 */   124,  141,   82,  205,  114,  219,  120,  219,  219,  219,
 /*  2260 */   205,  118,  126,  127,  128,  129,  156,  219,  205,   99,
 /*  2270 */   100,  118,  118,  205,  219,    4,  118,  107,  168,  169,
 /*  2280 */   116,  112,  219,  228,  229,  115,  116,  219,  116,  137,
 /*  2290 */   120,    4,  237,  115,  124,    4,  126,  127,  128,  129,
 /*  2300 */    82,  115,  134,   16,  134,   82,  126,  127,  128,  129,
 /*  2310 */    47,  141,  142,  143,  144,   82,   83,  134,    4,   86,
 /*  2320 */    87,   88,   89,   90,   91,   92,   93,  191,   95,   96,
 /*  2330 */    97,   98,   99,  100,  101,  102,  103,  104,  105,  106,
 /*  2340 */    46,   54,  205,  271,  314,  290,  218,  177,  168,  234,
 /*  2350 */   180,  120,  182,  183,  205,    4,  219,  126,  127,  128,
 /*  2360 */   129,  205,  205,   76,    4,  291,  204,   16,  219,   82,
 /*  2370 */   205,  109,  141,  234,   63,  219,  219,  316,  316,  205,
 /*  2380 */   325,  205,  205,  328,  219,  204,   99,  100,    4,   77,
 /*  2390 */   205,  120,    4,  219,  107,  219,  219,  126,  127,  128,
 /*  2400 */   129,  204,  115,  116,  219,   54,  115,  120,  206,  205,
 /*  2410 */   205,  124,  141,  126,  127,  128,  129,  126,  127,  128,
 /*  2420 */   129,  134,  205,  219,  219,  205,    4,   76,  141,  142,
 /*  2430 */   143,  144,  204,   82,  120,  206,  219,  211,   16,  219,
 /*  2440 */   126,  127,  128,  129,  208,  228,  229,  121,  118,  274,
 /*  2450 */    99,  100,  106,  115,  237,  141,  106,  233,  107,  318,
 /*  2460 */   299,  117,  184,  175,  177,  294,  115,  180,  177,  182,
 /*  2470 */   183,  120,    4,  259,  172,  124,   54,  126,  127,  128,
 /*  2480 */   129,  260,  265,  261,  174,  134,  126,  127,  128,  129,
 /*  2490 */     4,  173,  141,  142,  143,  144,  171,  263,   76,  262,
 /*  2500 */    44,  263,   16,   27,   82,  280,  274,  290,  326,  118,
 /*  2510 */   126,  127,  128,  129,  126,  127,  128,  129,  326,  280,
 /*  2520 */   255,   99,  100,  158,  331,  230,    4,  139,  177,  107,
 /*  2530 */   170,  180,  195,  182,  183,  184,    4,  115,  116,   71,
 /*  2540 */    54,  146,  120,  230,  240,   37,  124,  230,  126,  127,
 /*  2550 */   128,  129,  168,  232,  145,  233,  134,  205,  230,  240,
 /*  2560 */   233,    4,   76,  141,  142,  143,  144,  208,   82,  230,
 /*  2570 */   230,  219,  230,   16,  233,  230,  208,  274,  118,  305,
 /*  2580 */   228,  229,   33,  115,  116,   99,  100,  306,  120,  237,
 /*  2590 */   184,  259,  263,  107,  126,  127,  128,  129,  282,  177,
 /*  2600 */     4,  115,  180,  176,  182,  183,  120,    4,  205,  141,
 /*  2610 */   124,   54,  126,  127,  128,  129,  262,  265,  247,    4,
 /*  2620 */   134,  205,  219,  205,  156,  157,  158,  141,  142,  143,
 /*  2630 */   144,  228,  229,   76,   80,  219,  259,  219,    4,   82,
 /*  2640 */   237,  119,  290,  282,  228,  229,  228,  229,  126,  127,
 /*  2650 */   128,  129,  120,  237,  262,  237,   99,  100,  126,  127,
 /*  2660 */   128,  129,    4,  177,  107,  263,  180,  285,  182,  183,
 /*  2670 */   115,  288,  115,  141,    4,  146,  282,  120,    4,  259,
 /*  2680 */   282,  124,  279,  126,  127,  128,  129,    4,  205,  247,
 /*  2690 */   208,  134,  118,  290,  259,  279,  208,  279,  141,  142,
 /*  2700 */   143,  144,  219,    4,  205,  274,  290,    4,  290,  118,
 /*  2710 */   274,  228,  229,  110,  249,   77,    4,  300,  219,  249,
 /*  2720 */   237,    4,  126,  127,  128,  129,    4,  228,  229,  126,
 /*  2730 */   127,  128,  129,  249,  177,  205,  237,  180,  249,  182,
 /*  2740 */   183,  126,  127,  128,  129,  178,  150,    4,  265,  219,
 /*  2750 */   118,    4,  274,  309,   20,  309,    4,  118,  228,  229,
 /*  2760 */   126,  127,  128,  129,    4,  274,  274,  237,  118,  230,
 /*  2770 */   230,   17,    4,  290,  311,  141,  230,    4,  279,  302,
 /*  2780 */   295,  291,    4,  302,  126,  127,  128,  129,  262,  290,
 /*  2790 */     4,  231,  177,  256,  258,  265,  126,  127,  128,  129,
 /*  2800 */   126,  127,  128,  129,  262,    4,  273,  231,  238,  126,
 /*  2810 */   127,  128,  129,  155,    4,  141,  232,  242,    4,  258,
 /*  2820 */   290,  259,  311,    4,  141,  126,  127,  128,  129,  126,
 /*  2830 */   127,  128,  129,    4,  289,  311,  203,  256,  126,  127,
 /*  2840 */   128,  129,  221,  126,  127,  128,  129,  221,  126,  127,
 /*  2850 */   128,  129,  221,  141,    4,  303,  221,  304,  141,    4,
 /*  2860 */   190,  311,  254,  164,    4,  292,  119,  164,  248,  126,
 /*  2870 */   127,  128,  129,  126,  127,  128,  129,    4,  126,  127,
 /*  2880 */   128,  129,  272,  239,  141,  329,  126,  127,  128,  129,
 /*  2890 */   239,  239,  170,  141,  126,  127,  128,  129,  315,  126,
 /*  2900 */   127,  128,  129,  228,  126,  127,  128,  129,  228,  243,
 /*  2910 */    71,  332,  126,  127,  128,  129,  156,  332,  332,  141,
 /*  2920 */   119,  332,  332,  332,  332,  332,  332,  126,  127,  128,
 /*  2930 */   129,  332,  332,  332,  332,  125,  126,  127,  128,  129,
 /*  2940 */   126,  127,  128,  129,  332,  126,  127,  128,  129,  332,
 /*  2950 */   332,  332,  332,  332,  332,  126,  127,  128,  129,  191,
 /*  2960 */   332,  332,  189,  177,  332,  332,  332,  332,  332,  332,
 /*  2970 */   332,  332,  332,  332,  119,  332,  126,  127,  128,  129,
 /*  2980 */   332,  126,  127,  128,  129,  125,  126,  127,  128,  129,
 /*  2990 */   205,  177,  332,  205,  332,  205,  177,  332,  332,  126,
 /*  3000 */   127,  128,  129,  332,  219,  332,  177,  219,  332,  219,
 /*  3010 */   332,  332,  332,  228,  229,  332,  228,  229,  228,  229,
 /*  3020 */   205,  332,  237,  332,  332,  237,  205,  237,  178,  332,
 /*  3030 */   332,  205,  332,  205,  219,  332,  332,  205,  332,  332,
 /*  3040 */   219,  332,  332,  228,  229,  219,  332,  219,  332,  228,
 /*  3050 */   229,  219,  237,  332,  228,  229,  228,  229,  237,  205,
 /*  3060 */   228,  229,  332,  237,  332,  237,  332,  332,  332,  237,
 /*  3070 */   205,  332,  205,  219,  332,  290,  205,  332,  290,  332,
 /*  3080 */   290,  332,  228,  229,  219,  332,  219,  332,  332,  332,
 /*  3090 */   219,  237,  332,  228,  229,  228,  229,  332,  332,  228,
 /*  3100 */   229,  205,  237,  205,  237,  290,  205,  332,  237,  332,
 /*  3110 */   332,  290,  332,  332,  332,  219,  290,  219,  290,  332,
 /*  3120 */   219,  332,  290,  332,  228,  229,  228,  229,  332,  228,
 /*  3130 */   229,  332,  332,  237,  332,  237,  205,  332,  237,  332,
 /*  3140 */   332,  332,  205,  332,  290,  332,  332,  332,  205,  332,
 /*  3150 */   219,  332,  332,  205,  332,  290,  219,  290,  332,  228,
 /*  3160 */   229,  290,  219,  332,  332,  228,  229,  219,  237,  205,
 /*  3170 */   332,  228,  229,  332,  237,  332,  228,  229,  332,  332,
 /*  3180 */   237,  332,  205,  219,  205,  237,  290,  332,  290,  332,
 /*  3190 */   332,  290,  228,  229,  332,  332,  219,  332,  219,  205,
 /*  3200 */   332,  237,  332,  332,  332,  228,  229,  228,  229,  332,
 /*  3210 */   332,  332,  205,  219,  237,  332,  237,  332,  332,  332,
 /*  3220 */   332,  290,  228,  229,  332,  205,  219,  290,  205,  332,
 /*  3230 */   332,  237,  205,  290,  332,  228,  229,  332,  290,  219,
 /*  3240 */   332,  332,  219,  332,  237,  332,  219,  332,  228,  229,
 /*  3250 */   332,  228,  229,  332,  290,  228,  229,  237,  205,  332,
 /*  3260 */   237,  332,  205,  332,  237,  332,  332,  290,  332,  290,
 /*  3270 */   332,  332,  219,  332,  332,  205,  219,  332,  332,  205,
 /*  3280 */   332,  228,  229,  332,  290,  228,  229,  332,  332,  219,
 /*  3290 */   237,  332,  205,  219,  237,  332,  332,  290,  228,  229,
 /*  3300 */   332,  332,  228,  229,  332,  205,  219,  237,  205,  332,
 /*  3310 */   290,  237,  205,  290,  332,  228,  229,  290,  332,  219,
 /*  3320 */   332,  332,  219,  332,  237,  332,  219,  332,  228,  229,
 /*  3330 */   332,  228,  229,  332,  332,  228,  229,  237,  332,  205,
 /*  3340 */   237,  332,  205,  290,  237,  332,  332,  290,  332,  332,
 /*  3350 */   332,  332,  332,  219,  332,  332,  219,  332,  332,  205,
 /*  3360 */   290,  332,  228,  229,  290,  228,  229,  332,  332,  332,
 /*  3370 */   332,  237,  205,  219,  237,  205,  332,  290,  205,  332,
 /*  3380 */   332,  332,  228,  229,  332,  332,  219,  332,  332,  219,
 /*  3390 */   290,  237,  219,  290,  205,  228,  229,  290,  228,  229,
 /*  3400 */   332,  228,  229,  332,  237,  205,  332,  237,  219,  332,
 /*  3410 */   237,  205,  332,  332,  332,  332,  332,  228,  229,  219,
 /*  3420 */   332,  332,  205,  332,  290,  219,  237,  290,  228,  229,
 /*  3430 */   332,  332,  332,  332,  228,  229,  219,  237,  205,  332,
 /*  3440 */   332,  205,  332,  237,  290,  228,  229,  332,  332,  332,
 /*  3450 */   332,  332,  219,  332,  237,  219,  332,  290,  332,  332,
 /*  3460 */   290,  228,  229,  290,  228,  229,  332,  332,  205,  332,
 /*  3470 */   237,  332,  332,  237,  332,  332,  332,  332,  332,  290,
 /*  3480 */   332,  205,  219,  332,  332,  205,  332,  205,  332,  332,
 /*  3490 */   290,  228,  229,  332,  332,  219,  290,  332,  332,  219,
 /*  3500 */   237,  219,  332,  332,  228,  229,  332,  290,  228,  229,
 /*  3510 */   228,  229,  332,  237,  332,  332,  332,  237,  332,  237,
 /*  3520 */   332,  332,  332,  290,  332,  332,  290,  332,  332,  332,
 /*  3530 */   332,  332,  332,  332,  332,  332,  332,  332,  332,  332,
 /*  3540 */   332,  332,  332,  332,  332,  332,  332,  332,  332,  332,
 /*  3550 */   332,  332,  332,  290,  332,  332,  332,  332,  332,  332,
 /*  3560 */   332,  332,  332,  332,  332,  332,  290,  332,  332,  332,
 /*  3570 */   290,  332,  290,
};
#define YY_SHIFT_USE_DFLT (-104)
#define YY_SHIFT_COUNT (575)
#define YY_SHIFT_MIN   (-103)
#define YY_SHIFT_MAX   (2873)
static const short yy_shift_ofst[] = {
 /*     0 */  1407,  707,  888, 1166,  824, 1368, 1247, 1126, 2422, 2422,
 /*    10 */   306,  214, 2351, 2486, 2486, 2486, 2486, 2486, 2486, 2557,
 /*    20 */  2486, 2486, 1037,   -4,  105, 2287, 2170, 2049, 1928, 2486,
 /*    30 */  2486, 2486, 2486, 2486, 2486, 2486, 2486, 2486, 2486, 2486,
 /*    40 */  2486, 2557, 2486, 2486, 2486, 2486, 2486, 2486, 2486, 2486,
 /*    50 */  2486, 2486, 2486, 2486, 2486, 2486, 2486, 2486, 2486, 2486,
 /*    60 */  2486, 2486, 2486, 2486, 2486, 2486, 2486, 2486, 2486, 2486,
 /*    70 */  2486, 2486, 2486, 2486, 2486, 2486, 2486, 2468,  892,  892,
 /*    80 */   790,  808,  808, 2860, 2110, 2083, 1353, 2860,  126, 2819,
 /*    90 */  2819, 2819, 2855, 2819, 2873,  499,  499,  649,  341,  174,
 /*   100 */   834, 2532, 2136, 2001,  854, 2314, 2291, 2271, 2291, 2231,
 /*   110 */   727, 1058, 1353, 1336, 2850, 2829, 2829, 2829, 2829, 2829,
 /*   120 */  2819, 2873, 2873,  714,  185,  649, 1619, 1619, 1043,  523,
 /*   130 */   523,  -44,   65,  431,  645,  870, 1320, 1223,  928, 1207,
 /*   140 */  1311,   88, 1190,  114,  323,  924, 2814, 2810, 2786, 2801,
 /*   150 */  2778, 2773, 2768, 2722, 2760, 2752, 2747, 2743, 1594, 2717,
 /*   160 */  2712, 2703, 2699, 2683, 2674, 2670, 2658, 2522, 2634, 2388,
 /*   170 */  2596,  414, 1147, 2384, 2180, 2360, 1009, 2615, 1846,  327,
 /*   180 */   409, 1462, 2603, 1809, 1497,  630, 2873, 2873, 2873, 2873,
 /*   190 */   349,  349,  709, 1154,   -2,   -2,   -2,  298,  298,  225,
 /*   200 */   178, 1003, 1003,  194,  273,  643,  643,  643, 1314, 1043,
 /*   210 */  2839, 2839, 2839, 2839, -104, -104, -104,  426,  604,  604,
 /*   220 */   604,  604,  604,  604,  604,  604,  604,  604,  576,  391,
 /*   230 */  1023,  996, 1825, 1794, 1765, 1736, 1704, 1675, 1644, 1613,
 /*   240 */  1582, 1551, 1513, 1476, 1436, 1315,  -72,  646, 1066, 1897,
 /*   250 */  1865, 1993, 2233, 2233,  488,  -16,  -16,  -16,  -16,  -16,
 /*   260 */   -16,   48,  265,   30,  367,  110,  110,  945,  516,   51,
 /*   270 */    51,   51,  772, 1171, 1169,  -54,  375,  681,  732,  562,
 /*   280 */   817,  771,  333,  612,  648,   59,  518, 1268,  763,  591,
 /*   290 */  1307, 1127,  784, 1278,  880,  861, 1099,   64,   64, 1055,
 /*   300 */   390,  449,  861,  990,  990,   74, 2395, 2754, 2395, 2395,
 /*   310 */  2650, 2639, 2734, 2734, 2632, 2567, 2638, 2638, 2638, 2638,
 /*   320 */  2591, 2312, 2288, 2574, 2312, 2554, 2427, 2288, 2427, 2529,
 /*   330 */  2555, 2310, 2325, 2427, 2288, 2554, 2310, 2325, 2427, 2288,
 /*   340 */  2406, 2549, 2460, 2312, 2338, 2312, 2338, 2395, 2395, 2395,
 /*   350 */  2395, 2508, 2338, 2395, 2409, 2395, 2508, 2395, 2395, 2337,
 /*   360 */  2365, 2391, 2476, 2476, 2456, 2456, 2310, 2325, 2310, 2318,
 /*   370 */  2302, 2288, 2278, 2344, 2350, 2338, 2346, 2346, 2330, 2326,
 /*   380 */  2312, 2311, 2311, 2262, 2262, 2262, 2262, -104, -104, -104,
 /*   390 */  -104, -104, -104, -104, -104, -104, -104,  736,  561,  483,
 /*   400 */   798,   12,  573,   97,  311,  528, 1245, 1244,  641, 1122,
 /*   410 */    95,  925, 1075, 1018,  687,  543,  873,  452,   94,  224,
 /*   420 */    49,  354, -103,  212,  199,   75,  307,  177,  -23,  -81,
 /*   430 */   -52, 2294, 2263, 2183, 2223, 2168, 2218, 2186, 2178, 2152,
 /*   440 */  2169, 2158, 2140, 2172, 2164, 2154, 2153, 2143, 2126, 2131,
 /*   450 */  2125, 2056, 2106, 2102, 2050, 2019, 2145, 2067, 2138, 2122,
 /*   460 */  2160, 2058, 2057, 2005, 1980, 2009, 2133, 2132, 2037, 1939,
 /*   470 */  2052, 2045, 2035, 2033, 1962, 1986, 1989, 1937, 2053, 1896,
 /*   480 */  1983, 1956, 1917, 1958, 1916, 1879, 1892, 1909, 1965, 1902,
 /*   490 */  1894, 1898, 1893, 1943, 1973, 1861, 1842, 1944, 1834, 1895,
 /*   500 */  1924, 1824, 1925, 1759, 1818, 1793, 1890, 1732, 1786, 1707,
 /*   510 */  1762, 1756,  336, 1699,  336, 1734, 1742, 1712, 1702, 1647,
 /*   520 */  1673, 1672, 1653, 1633, 1740, 1574, 1363, 1739, 1738, 1714,
 /*   530 */  1652, 1608, 1599, 1590, 1577, 1560, 1540, 1541, 1414, 1405,
 /*   540 */  1324, 1392, 1379, 1411, 1363, 1293, 1276, 1220, 1183, 1165,
 /*   550 */  1168, 1116, 1246, 1039, 1100, 1088,  722, 1039, 1024,  975,
 /*   560 */   959,  943,  918,  799,  746,  678,  565,  447,  480,  370,
 /*   570 */   336,  268,  195,  153,  -10,  -86,
};
#define YY_REDUCE_USE_DFLT (-217)
#define YY_REDUCE_COUNT (396)
#define YY_REDUCE_MIN   (-216)
#define YY_REDUCE_MAX   (3282)
static const short yy_reduce_ofst[] = {
 /*     0 */  1253,  993, 1907,  517, 2055,  657,  173, 1933, 1914,  580,
 /*    10 */    11, 2530, 1093, 2499, 2418, 2416, 2483, 2403, 1052, 2352,
 /*    20 */  2217,  582,  112, 3282, 3280, 3276, 3263, 3236, 3233, 3217,
 /*    30 */  3206, 3200, 3189, 3173, 3170, 3167, 3154, 3137, 3134, 3107,
 /*    40 */  3103, 3100, 3087, 3074, 3070, 3057, 3053, 3027, 3023, 3020,
 /*    50 */  3007, 2994, 2979, 2977, 2964, 2948, 2943, 2937, 2931, 2901,
 /*    60 */  2898, 2896, 2871, 2867, 2865, 2854, 2832, 2828, 2826, 2821,
 /*    70 */  2815, 2790, 2788, 2785, 1918, 1810,  -31, 1086,   98,  -30,
 /*    80 */   131,  750,   81,  351, 1163,  976, 1005,  435,  529, 1817,
 /*    90 */  1806, 1254,  683,  988,   31,  422, -125, -216,  118,  403,
 /*   100 */  1948, 1859, 1858, 1840, 1807, 1378, 1832, 1272, 1302, 1267,
 /*   110 */   -19, 1284, 1042, 1030,  620, 1238, 1158, 1083, 1077,  804,
 /*   120 */   357,  862,  822,  557, 1048,  755, 1008,  642, -182, 1520,
 /*   130 */  1339,  695, 1259, 2220, 2205, 2165, 1913, 1913, 2204, 2185,
 /*   140 */  2177, 2176, 2174, 1913, 2165, 2157, 2156, 2149, 2137, 2068,
 /*   150 */  2036, 1913, 2036, 2036, 2036, 2036, 1913, 2063, 1913, 2039,
 /*   160 */  2039, 2048, 2040, 2039, 2038, 1913, 2036, 2036, 2029, 2028,
 /*   170 */  2023, 2008, 2002, 2000, 1996, 1995, 1984, 1976, 1949, 1913,
 /*   180 */  1841, 1839, 1814, 1813, 1802, 1801, 1683, 1593, 1471, 1059,
 /*   190 */  1258, 1212, 1470, 1312, 1439, 1424,  929, 1304, 1237,  696,
 /*   200 */  1327, 1230, 1046, 1269, 1292, 1196, 1167, 1012, 1049,  577,
 /*   210 */  1094, 1081,  856, -210,  830,  689,  277, 2074, 2074, 2074,
 /*   220 */  2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 /*   230 */  2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 /*   240 */  2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 /*   250 */  2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 /*   260 */  2074, 2074, 2074, 2074, 2666, 2680, 2675, 2583, 2583, 2652,
 /*   270 */  2651, 2644, 2610, 2556, 2620, 2608, 2573, 2550, 2553, 2552,
 /*   280 */  2635, 2631, 2626, 2074, 2621, 2581, 2633, 2524, 2511, 2562,
 /*   290 */  2545, 2561, 2584, 2575, 2570, 2576, 2533, 2542, 2526, 2536,
 /*   300 */  2537, 2485, 2560, 2481, 2477, 2490, 2546, 2463, 2540, 2539,
 /*   310 */  2492, 2491, 2446, 2444, 2478, 2417, 2489, 2484, 2470, 2465,
 /*   320 */  2436, 2488, 2435, 2431, 2482, 2442, 2398, 2420, 2394, 2383,
 /*   330 */  2382, 2402, 2392, 2361, 2377, 2371, 2329, 2354, 2316, 2332,
 /*   340 */  2281, 2274, 2303, 2368, 2341, 2359, 2327, 2345, 2342, 2340,
 /*   350 */  2339, 2319, 2322, 2328, 2321, 2317, 2304, 2313, 2295, 2193,
 /*   360 */  2265, 2232, 2192, 2182, 2239, 2225, 2238, 2237, 2234, 2222,
 /*   370 */  2221, 2214, 2171, 2141, 2161, 2224, 2074, 2074, 2175, 2226,
 /*   380 */  2236, 2229, 2202, 2228, 2197, 2181, 2162, 2062, 2061, 2030,
 /*   390 */  2139, 2115, 2072, 2074, 2074, 2074, 2128,
};
static const YYACTIONTYPE yy_default[] = {
 /*     0 */   934, 1269, 1269, 1402, 1269, 1269, 1269, 1269, 1269, 1269,
 /*    10 */  1402, 1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269,
 /*    20 */  1269, 1269, 1402, 1269, 1269, 1269, 1269, 1269, 1269, 1269,
 /*    30 */  1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269,
 /*    40 */  1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269,
 /*    50 */  1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269,
 /*    60 */  1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269, 1269,
 /*    70 */  1269, 1269, 1269, 1269, 1269, 1269, 1269, 1120, 1444, 1444,
 /*    80 */  1444, 1421, 1421, 1444, 1113, 1444,  962, 1444, 1444, 1444,
 /*    90 */  1444, 1444, 1444, 1444, 1444, 1444, 1444,  986, 1109,  975,
 /*   100 */  1120, 1444, 1444, 1444, 1444, 1444, 1191, 1206, 1191, 1183,
 /*   110 */  1172, 1444, 1444, 1444, 1307, 1199, 1199, 1199, 1199, 1199,
 /*   120 */  1444, 1444, 1444, 1059, 1233, 1444, 1235, 1234, 1444, 1157,
 /*   130 */  1157, 1271, 1444, 1360, 1365, 1227, 1444, 1444, 1444, 1444,
 /*   140 */  1444, 1192, 1444, 1121, 1227, 1444, 1381, 1444, 1380, 1444,
 /*   150 */  1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444, 1207,
 /*   160 */  1184, 1444, 1444, 1173, 1444, 1444, 1444, 1444, 1444, 1444,
 /*   170 */  1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444, 1229, 1444,
 /*   180 */  1444, 1444, 1444, 1444, 1444,  940, 1444, 1444, 1444, 1444,
 /*   190 */  1444, 1444, 1109, 1421, 1109, 1109, 1109, 1421, 1421,  972,
 /*   200 */   980, 1421, 1421,  934, 1402, 1146, 1146, 1146, 1151, 1444,
 /*   210 */  1402, 1402, 1402, 1402, 1395, 1099, 1099, 1167, 1190, 1189,
 /*   220 */  1188, 1187, 1126, 1177, 1165, 1169, 1168, 1284, 1271, 1271,
 /*   230 */  1271, 1271, 1271, 1271, 1271, 1271, 1271, 1271, 1271, 1271,
 /*   240 */  1271, 1271, 1271, 1271, 1271, 1271, 1271, 1271, 1271, 1271,
 /*   250 */  1271, 1237, 1251, 1236, 1244, 1256, 1245, 1250, 1249, 1248,
 /*   260 */  1239, 1238, 1240, 1241, 1444, 1444, 1444, 1444, 1444, 1444,
 /*   270 */  1444, 1444, 1112, 1444, 1444, 1081, 1444, 1386, 1444, 1334,
 /*   280 */  1444, 1444,  988, 1242, 1444, 1097,  943, 1386, 1386, 1176,
 /*   290 */  1273, 1110, 1026, 1050, 1010, 1157, 1128, 1151, 1151, 1110,
 /*   300 */  1097, 1279, 1157, 1444, 1444, 1271, 1061, 1386, 1061, 1061,
 /*   310 */  1130, 1130, 1359, 1359, 1130, 1307, 1072, 1072, 1072, 1072,
 /*   320 */  1130,  969, 1176, 1130,  969, 1063, 1179, 1176, 1179, 1212,
 /*   330 */  1195, 1166, 1151, 1179, 1176, 1063, 1166, 1151, 1179, 1176,
 /*   340 */  1341, 1339, 1130,  969, 1292,  969, 1292, 1061, 1061, 1061,
 /*   350 */  1061, 1042, 1292, 1061, 1026, 1061, 1042, 1061, 1061, 1440,
 /*   360 */  1444, 1130, 1431, 1431, 1160, 1160, 1166, 1151, 1166, 1164,
 /*   370 */  1161, 1176, 1444, 1444, 1298, 1292, 1255, 1243, 1130, 1045,
 /*   380 */   969,  951,  951,  939,  939,  939,  939, 1399, 1399, 1395,
 /*   390 */  1028, 1028, 1115, 1254, 1253, 1252,  997, 1270, 1444, 1444,
 /*   400 */  1444, 1444, 1444, 1444, 1444, 1308, 1444, 1444, 1444, 1444,
 /*   410 */  1444, 1134, 1444, 1444, 1134, 1134, 1444, 1444, 1444, 1444,
 /*   420 */  1444, 1444, 1444, 1444, 1444, 1444, 1405, 1444, 1444, 1444,
 /*   430 */  1444, 1444,  935, 1444, 1444, 1444, 1444, 1444, 1444, 1444,
 /*   440 */  1444, 1132, 1444, 1444, 1444, 1132, 1132, 1132, 1444, 1132,
 /*   450 */  1389, 1444, 1444, 1444, 1444, 1444, 1444, 1338, 1337, 1444,
 /*   460 */  1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444, 1196, 1444,
 /*   470 */  1444, 1444, 1444, 1444, 1444, 1180, 1135, 1444, 1326, 1444,
 /*   480 */  1132, 1444, 1444, 1444, 1132, 1444, 1444, 1444, 1444, 1444,
 /*   490 */  1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444,
 /*   500 */  1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444,
 /*   510 */  1444, 1444, 1085, 1091, 1086, 1444, 1444, 1444, 1225, 1444,
 /*   520 */  1444, 1444, 1444, 1444, 1444, 1444, 1162, 1444, 1444, 1444,
 /*   530 */  1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444, 1444,
 /*   540 */  1444, 1437, 1152, 1444, 1282, 1444, 1444, 1444, 1444, 1272,
 /*   550 */  1444, 1270, 1444, 1404, 1444, 1444, 1444, 1403, 1444, 1444,
 /*   560 */  1444, 1444, 1444, 1272, 1444, 1444, 1226, 1225, 1444, 1444,
 /*   570 */  1084,  983,  958, 1444,  949, 1444,  931,  936, 1388, 1385,
 /*   580 */  1384, 1374, 1373, 1377, 1371, 1376, 1375, 1370, 1387, 1379,
 /*   590 */  1382, 1372, 1378, 1369, 1383, 1368, 1366, 1367, 1364, 1362,
 /*   600 */  1361, 1363, 1358, 1354, 1314, 1312, 1310, 1319, 1318, 1317,
 /*   610 */  1316, 1315, 1311, 1309, 1313, 1305, 1304, 1202, 1181, 1170,
 /*   620 */  1079, 1353, 1351, 1352, 1303, 1301, 1302, 1078, 1077, 1076,
 /*   630 */  1071, 1070, 1069, 1068, 1392, 1401, 1400, 1398, 1397, 1396,
 /*   640 */  1390, 1391, 1290, 1289, 1287, 1286, 1288,  971, 1330, 1333,
 /*   650 */  1332, 1331, 1336, 1335, 1328, 1340, 1345, 1344, 1349, 1348,
 /*   660 */  1347, 1346, 1343, 1325, 1211, 1210, 1208, 1204, 1215, 1214,
 /*   670 */  1213, 1203, 1205, 1209, 1186, 1194, 1193, 1182, 1185, 1064,
 /*   680 */  1175, 1171, 1139, 1137, 1174, 1138, 1136, 1080, 1329, 1075,
 /*   690 */  1074, 1073,  970,  961, 1133,  960,  959,  974, 1048, 1049,
 /*   700 */  1057, 1060, 1055, 1058, 1054, 1053, 1052, 1056, 1051, 1047,
 /*   710 */   977,  976,  987, 1041, 1024, 1013,  979, 1015, 1012, 1011,
 /*   720 */  1016, 1033, 1032, 1039, 1038, 1037, 1036, 1035, 1031, 1034,
 /*   730 */  1030, 1029, 1017, 1009, 1008, 1027, 1007, 1044, 1043, 1040,
 /*   740 */  1006, 1067, 1066, 1065, 1062, 1005, 1004, 1003, 1002, 1001,
 /*   750 */  1000, 1267, 1266, 1443, 1439, 1442, 1441, 1438, 1268, 1265,
 /*   760 */  1275, 1259, 1257, 1082, 1094, 1093, 1092, 1089, 1104, 1102,
 /*   770 */  1101, 1100, 1145, 1144, 1143, 1142, 1141, 1140, 1125, 1123,
 /*   780 */  1118, 1117, 1124, 1122, 1119, 1149, 1150, 1148, 1147, 1116,
 /*   790 */  1108, 1106, 1107, 1105, 1201, 1198, 1200, 1197, 1127, 1129,
 /*   800 */  1114, 1111, 1098, 1419, 1417, 1420, 1418, 1416, 1424, 1426,
 /*   810 */  1425, 1430, 1428, 1427, 1423, 1436, 1435, 1434, 1433, 1432,
 /*   820 */  1422, 1429, 1415, 1414, 1413, 1412, 1154, 1159, 1158, 1153,
 /*   830 */  1088, 1087, 1350, 1263, 1222, 1264, 1258, 1270, 1262, 1410,
 /*   840 */  1408, 1411, 1407, 1406, 1295, 1297, 1300, 1299, 1296, 1156,
 /*   850 */  1155, 1294, 1293, 1409, 1261, 1232,  993,  991,  992, 1322,
 /*   860 */  1321, 1324, 1323, 1320,  995,  994,  990,  989, 1230, 1224,
 /*   870 */  1223, 1247, 1246, 1231, 1221,  982,  981, 1228, 1220, 1219,
 /*   880 */  1218, 1090, 1083, 1260, 1217, 1131, 1216, 1023, 1022, 1021,
 /*   890 */  1020, 1019, 1018, 1096, 1095,  999, 1014,  998,  996,  973,
 /*   900 */   963,  968,  966,  967,  965,  964,  956,  953,  955,  952,
 /*   910 */   957,  954,  950,  948,  947,  946,  945,  944,  985,  984,
 /*   920 */   983,  978,  942,  941,  938,  937,  933,  932,  930,
};

/* The next table maps tokens into fallback tokens.  If a construct
** like the following:
**
**      %fallback ID X Y Z.
**
** appears in the grammar, then ID becomes a fallback token for X, Y,
** and Z.  Whenever one of the tokens X, Y, or Z is input to the parser
** but it does not parse, the type of the token is changed to ID and
** the parse is retried before an error is thrown.
*/
#ifdef YYFALLBACK
static const YYCODETYPE yyFallback[] = {
    0,  /*          $ => nothing */
    0,  /*    ILLEGAL => nothing */
    0,  /*    COMMENT => nothing */
    0,  /*      SPACE => nothing */
    0,  /*         ID => nothing */
    4,  /*      ABORT => ID */
    4,  /*     ACTION => ID */
    4,  /*      AFTER => ID */
    4,  /*     ALWAYS => ID */
    4,  /*    ANALYZE => ID */
    4,  /*        ASC => ID */
    4,  /*     ATTACH => ID */
    4,  /*     BEFORE => ID */
    4,  /*      BEGIN => ID */
    4,  /*         BY => ID */
    4,  /*    CASCADE => ID */
    4,  /*       CAST => ID */
    4,  /*   COLUMNKW => ID */
    4,  /*   CONFLICT => ID */
    4,  /*    CURRENT => ID */
    4,  /*   DATABASE => ID */
    4,  /*   DEFERRED => ID */
    4,  /*       DESC => ID */
    4,  /*     DETACH => ID */
    4,  /*         DO => ID */
    4,  /*       EACH => ID */
    4,  /*        END => ID */
    4,  /*    EXCLUDE => ID */
    4,  /*  EXCLUSIVE => ID */
    4,  /*    EXPLAIN => ID */
    4,  /*       FAIL => ID */
    4,  /*      FIRST => ID */
    4,  /*  FOLLOWING => ID */
    4,  /*        FOR => ID */
    4,  /*     GROUPS => ID */
    4,  /*     IGNORE => ID */
    4,  /*  IMMEDIATE => ID */
    4,  /*  INITIALLY => ID */
    4,  /*    INSTEAD => ID */
    4,  /*       LAST => ID */
    4,  /*    LIKE_KW => ID */
    4,  /*      MATCH => ID */
    4,  /* MATERIALIZED => ID */
    4,  /*         NO => ID */
    4,  /*      NULLS => ID */
    4,  /*     OTHERS => ID */
    4,  /*       PLAN => ID */
    4,  /*      QUERY => ID */
    4,  /*        KEY => ID */
    4,  /*         OF => ID */
    4,  /*     OFFSET => ID */
    4,  /*  PARTITION => ID */
    4,  /*     PRAGMA => ID */
    4,  /*  PRECEDING => ID */
    4,  /*      RAISE => ID */
    4,  /*      RANGE => ID */
    4,  /*  RECURSIVE => ID */
    4,  /*    RELEASE => ID */
    4,  /*    REPLACE => ID */
    4,  /*   RESTRICT => ID */
    4,  /*        ROW => ID */
    4,  /*       ROWS => ID */
    4,  /*   ROLLBACK => ID */
    4,  /*  SAVEPOINT => ID */
    4,  /*       TEMP => ID */
    4,  /*       TIES => ID */
    4,  /*    TRIGGER => ID */
    4,  /*  UNBOUNDED => ID */
    4,  /*     VACUUM => ID */
    4,  /*       VIEW => ID */
    4,  /*    VIRTUAL => ID */
    4,  /*       WITH => ID */
    4,  /*     WITHIN => ID */
    4,  /*    WITHOUT => ID */
    4,  /*    REINDEX => ID */
    4,  /*     RENAME => ID */
    4,  /*   CTIME_KW => ID */
    4,  /*         IF => ID */
    4,  /*     FILTER => ID */
};
#endif /* YYFALLBACK */

/* The following structure represents a single element of the
** parser's stack.  Information stored includes:
**
**   +  The state number for the parser at this level of the stack.
**
**   +  The value of the token stored at this level of the stack.
**      (In other words, the "major" token.)
**
**   +  The semantic value stored at this level of the stack.  This is
**      the information used by the action routines in the grammar.
**      It is sometimes called the "minor" token.
*/
struct yyStackEntry {
  YYACTIONTYPE stateno;  /* The state-number */
  YYCODETYPE major;      /* The major token value.  This is the code
                         ** number for the token at this stack level */
  YYMINORTYPE minor;     /* The user-supplied minor token value.  This
                         ** is the value of the token  */
  QList<Token*>* tokens = nullptr;
};
typedef struct yyStackEntry yyStackEntry;

/* The state of the parser is completely contained in an instance of
** the following structure */
struct yyParser {
  int yyidx;                    /* Index of top element in stack */
#ifdef YYTRACKMAXSTACKDEPTH
  int yyidxMax;                 /* Maximum value of yyidx */
#endif
  int yyerrcnt;                 /* Shifts left before out of the error */
  sqlite3_parseARG_SDECL                /* A place to hold %extra_argument */
#if YYSTACKDEPTH<=0
  int yystksz;                  /* Current side of the stack */
  yyStackEntry *yystack;        /* The parser's stack */
#else
  yyStackEntry yystack[YYSTACKDEPTH];  /* The parser's stack */
#endif
};
typedef struct yyParser yyParser;

#ifndef NDEBUG
#include <stdio.h>
static FILE *yyTraceFILE = 0;
static char *yyTracePrompt = 0;
#endif /* NDEBUG */

void *sqlite3_parseCopyParserState(void* other)
{
  yyParser *pParser;
  yyParser *otherParser = (yyParser*)other;

  // Copy parser
  pParser = (yyParser*)malloc((size_t)sizeof(yyParser));
  memcpy(pParser, other, (size_t)sizeof(yyParser));

#if YYSTACKDEPTH<=0
  // Copy stack
  int stackSize = sizeof(yyStackEntry) * pParser->yystksz;
  pParser->yystack = malloc((size_t)stackSize);
  memcpy(pParser->yystack, ((yyParser*)other)->yystack, (size_t)stackSize);
#endif

  for (int i = 0; i <= pParser->yyidx; i++)
  {
      pParser->yystack[i].tokens = new QList<Token*>();
      *(pParser->yystack[i].tokens) = *(otherParser->yystack[i].tokens);
  }

  return pParser;
}

void sqlite3_parseAddToken(
  void* other,
  Token* token
  sqlite3_parseARG_PDECL
){
    yyParser *otherParser = (yyParser*)other;
    if (otherParser->yyidx < 0)
    {
        parserContext->addDanglingToken(token);
        return; // Nothing on stack yet. Might happen when parsing just whitespaces, nothing else.
    }

    otherParser->yystack[otherParser->yyidx].tokens->append(token);
}

void sqlite3_parseRestoreParserState(void* saved, void* target)
{
  yyParser *pParser = (yyParser*)target;
  yyParser *savedParser = (yyParser*)saved;

  for (int i = 0; i <= pParser->yyidx; i++)
      delete pParser->yystack[i].tokens;

  memcpy(pParser, saved, (size_t)sizeof(yyParser));

  for (int i = 0; i <= savedParser->yyidx; i++)
  {
      pParser->yystack[i].tokens = new QList<Token*>();
      *(pParser->yystack[i].tokens) = *(savedParser->yystack[i].tokens);
  }

#if YYSTACKDEPTH<=0
  // Copy stack
  int stackSize = sizeof(yyStackEntry) * pParser->yystksz;
  pParser->yystack = relloc(pParser->yystack, (size_t)stackSize);
  memcpy(pParser->yystack, ((yyParser*)saved)->yystack, (size_t)stackSize);
#endif
}

void sqlite3_parseFreeSavedState(void* other)
{
    yyParser *pParser = (yyParser*)other;
    for (int i = 0; i <= pParser->yyidx; i++)
        delete pParser->yystack[i].tokens;

#if YYSTACKDEPTH<=0
    free(pParser->yystack);
#endif
    free(other);
}

#ifndef NDEBUG
/*
** Turn parser tracing on by giving a stream to which to write the trace
** and a prompt to preface each trace message.  Tracing is turned off
** by making either argument NULL
**
** Inputs:
** <ul>
** <li> A FILE* to which trace output should be written.
**      If NULL, then tracing is turned off.
** <li> A prefix string written at the beginning of every
**      line of trace output.  If NULL, then tracing is
**      turned off.
** </ul>
**
** Outputs:
** None.
*/
void sqlite3_parseTrace(FILE *TraceFILE, char *zTracePrompt){
  yyTraceFILE = TraceFILE;
  yyTracePrompt = zTracePrompt;
  if( yyTraceFILE==0 ) yyTracePrompt = 0;
  else if( yyTracePrompt==0 ) yyTraceFILE = 0;
}
#endif /* NDEBUG */

#ifndef NDEBUG
/* For tracing shifts, the names of all terminals and nonterminals
** are required.  The following table supplies these names */
static const char *const yyTokenName[] = {
  "$",             "ILLEGAL",       "COMMENT",       "SPACE",       
  "ID",            "ABORT",         "ACTION",        "AFTER",       
  "ALWAYS",        "ANALYZE",       "ASC",           "ATTACH",      
  "BEFORE",        "BEGIN",         "BY",            "CASCADE",     
  "CAST",          "COLUMNKW",      "CONFLICT",      "CURRENT",     
  "DATABASE",      "DEFERRED",      "DESC",          "DETACH",      
  "DO",            "EACH",          "END",           "EXCLUDE",     
  "EXCLUSIVE",     "EXPLAIN",       "FAIL",          "FIRST",       
  "FOLLOWING",     "FOR",           "GROUPS",        "IGNORE",      
  "IMMEDIATE",     "INITIALLY",     "INSTEAD",       "LAST",        
  "LIKE_KW",       "MATCH",         "MATERIALIZED",  "NO",          
  "NULLS",         "OTHERS",        "PLAN",          "QUERY",       
  "KEY",           "OF",            "OFFSET",        "PARTITION",   
  "PRAGMA",        "PRECEDING",     "RAISE",         "RANGE",       
  "RECURSIVE",     "RELEASE",       "REPLACE",       "RESTRICT",    
  "ROW",           "ROWS",          "ROLLBACK",      "SAVEPOINT",   
  "TEMP",          "TIES",          "TRIGGER",       "UNBOUNDED",   
  "VACUUM",        "VIEW",          "VIRTUAL",       "WITH",        
  "WITHIN",        "WITHOUT",       "REINDEX",       "RENAME",      
  "CTIME_KW",      "IF",            "FILTER",        "ANY",         
  "OR",            "AND",           "NOT",           "IS",          
  "BETWEEN",       "IN",            "ISNULL",        "NOTNULL",     
  "NE",            "EQ",            "GT",            "LE",          
  "LT",            "GE",            "ESCAPE",        "BITAND",      
  "BITOR",         "LSHIFT",        "RSHIFT",        "PLUS",        
  "MINUS",         "STAR",          "SLASH",         "REM",         
  "CONCAT",        "PTR",           "COLLATE",       "BITNOT",      
  "SEMI",          "TRANSACTION",   "ID_TRANS",      "COMMIT",      
  "TO",            "CREATE",        "TABLE",         "LP",          
  "RP",            "AS",            "DOT",           "ID_TAB_NEW",  
  "ID_DB",         "COMMA",         "CTX_ROWID_KW",  "CTX_STRICT_KW",
  "EXISTS",        "ID_COL_NEW",    "INDEXED",       "GENERATED",   
  "STRING",        "JOIN_KW",       "ID_COL_TYPE",   "RIGHT_ASSOC", 
  "CONSTRAINT",    "DEFAULT",       "NULL",          "PRIMARY",     
  "UNIQUE",        "CHECK",         "REFERENCES",    "ID_CONSTR",   
  "ID_COLLATE",    "ID_TAB",        "INTEGER",       "FLOAT",       
  "BLOB",          "AUTOINCR",      "ON",            "INSERT",      
  "DELETE",        "UPDATE",        "ID_FK_MATCH",   "SET",         
  "DEFERRABLE",    "FOREIGN",       "DROP",          "ID_VIEW_NEW", 
  "ID_VIEW",       "SELECT",        "VALUES",        "UNION",       
  "ALL",           "EXCEPT",        "INTERSECT",     "DISTINCT",    
  "ID_ALIAS",      "FROM",          "USING",         "CTX_JOIN_EXPR",
  "JOIN",          "ID_JOIN_OPTS",  "ID_IDX",        "ORDER",       
  "GROUP",         "HAVING",        "LIMIT",         "WHERE",       
  "RETURNING",     "ID_COL",        "INTO",          "NOTHING",     
  "ID_FN",         "ID_ERR_MSG",    "VARIABLE",      "CASE",        
  "WHEN",          "THEN",          "ELSE",          "INDEX",       
  "ID_IDX_NEW",    "ID_PRAGMA",     "ID_TRIG_NEW",   "ID_TRIG",     
  "ALTER",         "ADD",           "WINDOW",        "OVER",        
  "error",         "cmd",           "input",         "cmdlist",     
  "ecmd",          "explain",       "cmdx",          "transtype",   
  "trans_opt",     "nm",            "savepoint_opt",  "temp",        
  "ifnotexists",   "fullname",      "columnlist",    "conslist_opt",
  "table_options",  "select",        "table_option",  "column",      
  "columnid",      "type",          "carglist",      "id",          
  "id_opt",        "ids",           "typetoken",     "typename",    
  "signed",        "plus_num",      "minus_num",     "ccons",       
  "term",          "expr",          "onconf",        "sortorder",   
  "autoinc",       "idxlist_opt",   "refargs",       "defer_subclause",
  "gen_always",    "tnm",           "refarg",        "refact",      
  "init_deferred_pred_opt",  "conslist",      "tconscomma",    "tcons",       
  "idxlist",       "defer_subclause_opt",  "resolvetype",   "orconf",      
  "raisetype",     "ifexists",      "select_stmt",   "with",        
  "selectnowith",  "oneselect",     "multiselect_op",  "values",      
  "distinct",      "selcollist",    "from",          "where_opt",   
  "groupby_opt",   "having_opt",    "orderby_opt",   "limit_opt",   
  "window_clause",  "nexprlist",     "exprlist",      "sclp",        
  "as",            "joinsrc",       "singlesrc",     "seltablist",  
  "joinop",        "joinconstr_opt",  "dbnm",          "indexed_opt", 
  "idlist",        "xfullname",     "indexed_by",    "sortlist",    
  "nulls",         "delete_stmt",   "returning",     "update_stmt", 
  "setlist",       "idlist_opt",    "insert_stmt",   "insert_cmd",  
  "upsert",        "rp_opt",        "exprx",         "not_opt",     
  "likeop",        "case_operand",  "case_exprlist",  "case_else",   
  "filter_over",   "uniqueflag",    "idxlist_single",  "collate",     
  "vinto",         "nmnum",         "number",        "trigger_time",
  "trigger_event",  "foreach_clause",  "when_clause",   "trigger_cmd_list",
  "trigger_cmd",   "database_kw_opt",  "key_opt",       "kwcolumn_opt",
  "create_vtab",   "vtabarglist",   "vtabarg",       "vtabargtoken",
  "anylist",       "wqlist",        "wqas",          "wqcte",       
  "windowdefn_list",  "windowdefn",    "window",        "frame_opt",   
  "range_or_rows",  "frame_bound_s",  "frame_exclude_opt",  "frame_bound_e",
  "frame_bound",   "frame_exclude",  "filter_clause",  "over_clause", 
};
#endif /* NDEBUG */

#ifndef NDEBUG
/* For tracing reduce actions, the names of all rules are required.
*/
static const char *const yyRuleName[] = {
 /*   0 */ "input ::= cmdlist",
 /*   1 */ "cmdlist ::= cmdlist ecmd",
 /*   2 */ "cmdlist ::= ecmd",
 /*   3 */ "ecmd ::= SEMI",
 /*   4 */ "ecmd ::= explain cmdx SEMI",
 /*   5 */ "explain ::=",
 /*   6 */ "explain ::= EXPLAIN",
 /*   7 */ "explain ::= EXPLAIN QUERY PLAN",
 /*   8 */ "cmdx ::= cmd",
 /*   9 */ "cmd ::= BEGIN transtype trans_opt",
 /*  10 */ "trans_opt ::=",
 /*  11 */ "trans_opt ::= TRANSACTION",
 /*  12 */ "trans_opt ::= TRANSACTION nm",
 /*  13 */ "trans_opt ::= TRANSACTION ID_TRANS",
 /*  14 */ "transtype ::=",
 /*  15 */ "transtype ::= DEFERRED",
 /*  16 */ "transtype ::= IMMEDIATE",
 /*  17 */ "transtype ::= EXCLUSIVE",
 /*  18 */ "cmd ::= COMMIT trans_opt",
 /*  19 */ "cmd ::= END trans_opt",
 /*  20 */ "cmd ::= ROLLBACK trans_opt",
 /*  21 */ "savepoint_opt ::= SAVEPOINT",
 /*  22 */ "savepoint_opt ::=",
 /*  23 */ "cmd ::= SAVEPOINT nm",
 /*  24 */ "cmd ::= RELEASE savepoint_opt nm",
 /*  25 */ "cmd ::= ROLLBACK trans_opt TO savepoint_opt nm",
 /*  26 */ "cmd ::= SAVEPOINT ID_TRANS",
 /*  27 */ "cmd ::= RELEASE savepoint_opt ID_TRANS",
 /*  28 */ "cmd ::= ROLLBACK trans_opt TO savepoint_opt ID_TRANS",
 /*  29 */ "cmd ::= CREATE temp TABLE ifnotexists fullname LP columnlist conslist_opt RP table_options",
 /*  30 */ "cmd ::= CREATE temp TABLE ifnotexists fullname AS select",
 /*  31 */ "cmd ::= CREATE temp TABLE ifnotexists nm DOT ID_TAB_NEW",
 /*  32 */ "cmd ::= CREATE temp TABLE ifnotexists ID_DB|ID_TAB_NEW",
 /*  33 */ "table_options ::=",
 /*  34 */ "table_options ::= table_option",
 /*  35 */ "table_options ::= table_options COMMA table_option",
 /*  36 */ "table_option ::= WITHOUT nm",
 /*  37 */ "table_option ::= nm",
 /*  38 */ "table_option ::= WITHOUT CTX_ROWID_KW",
 /*  39 */ "table_option ::= CTX_STRICT_KW",
 /*  40 */ "ifnotexists ::=",
 /*  41 */ "ifnotexists ::= IF NOT EXISTS",
 /*  42 */ "temp ::= TEMP",
 /*  43 */ "temp ::=",
 /*  44 */ "columnlist ::= columnlist COMMA column",
 /*  45 */ "columnlist ::= column",
 /*  46 */ "column ::= columnid type carglist",
 /*  47 */ "columnid ::= nm",
 /*  48 */ "columnid ::= ID_COL_NEW",
 /*  49 */ "id ::= ID|INDEXED|GENERATED",
 /*  50 */ "id_opt ::= id",
 /*  51 */ "id_opt ::=",
 /*  52 */ "ids ::= ID",
 /*  53 */ "ids ::= STRING",
 /*  54 */ "nm ::= id",
 /*  55 */ "nm ::= STRING",
 /*  56 */ "nm ::= JOIN_KW",
 /*  57 */ "type ::=",
 /*  58 */ "type ::= typetoken",
 /*  59 */ "typetoken ::= typename",
 /*  60 */ "typetoken ::= typename LP signed RP",
 /*  61 */ "typetoken ::= typename LP signed COMMA signed RP",
 /*  62 */ "typename ::= ids",
 /*  63 */ "typename ::= typename ids",
 /*  64 */ "typename ::= ID_COL_TYPE",
 /*  65 */ "signed ::= plus_num",
 /*  66 */ "signed ::= minus_num",
 /*  67 */ "carglist ::= carglist ccons",
 /*  68 */ "carglist ::=",
 /*  69 */ "ccons ::= CONSTRAINT nm",
 /*  70 */ "ccons ::= DEFAULT term",
 /*  71 */ "ccons ::= DEFAULT LP expr RP",
 /*  72 */ "ccons ::= DEFAULT PLUS term",
 /*  73 */ "ccons ::= DEFAULT MINUS term",
 /*  74 */ "ccons ::= DEFAULT id",
 /*  75 */ "ccons ::= DEFAULT CTIME_KW",
 /*  76 */ "ccons ::= NULL onconf",
 /*  77 */ "ccons ::= NOT NULL onconf",
 /*  78 */ "ccons ::= PRIMARY KEY sortorder onconf autoinc",
 /*  79 */ "ccons ::= UNIQUE onconf",
 /*  80 */ "ccons ::= CHECK LP expr RP",
 /*  81 */ "ccons ::= REFERENCES nm idxlist_opt refargs",
 /*  82 */ "ccons ::= defer_subclause",
 /*  83 */ "ccons ::= COLLATE ids",
 /*  84 */ "ccons ::= gen_always AS LP expr RP id_opt",
 /*  85 */ "ccons ::= CONSTRAINT ID_CONSTR",
 /*  86 */ "ccons ::= COLLATE ID_COLLATE",
 /*  87 */ "ccons ::= REFERENCES ID_TAB",
 /*  88 */ "ccons ::= CHECK LP RP",
 /*  89 */ "term ::= NULL",
 /*  90 */ "term ::= INTEGER",
 /*  91 */ "term ::= FLOAT",
 /*  92 */ "term ::= STRING|BLOB",
 /*  93 */ "tnm ::= term",
 /*  94 */ "tnm ::= nm",
 /*  95 */ "gen_always ::= GENERATED ALWAYS",
 /*  96 */ "gen_always ::=",
 /*  97 */ "autoinc ::=",
 /*  98 */ "autoinc ::= AUTOINCR",
 /*  99 */ "refargs ::=",
 /* 100 */ "refargs ::= refargs refarg",
 /* 101 */ "refarg ::= MATCH nm",
 /* 102 */ "refarg ::= ON INSERT refact",
 /* 103 */ "refarg ::= ON DELETE refact",
 /* 104 */ "refarg ::= ON UPDATE refact",
 /* 105 */ "refarg ::= MATCH ID_FK_MATCH",
 /* 106 */ "refact ::= SET NULL",
 /* 107 */ "refact ::= SET DEFAULT",
 /* 108 */ "refact ::= CASCADE",
 /* 109 */ "refact ::= RESTRICT",
 /* 110 */ "refact ::= NO ACTION",
 /* 111 */ "defer_subclause ::= NOT DEFERRABLE init_deferred_pred_opt",
 /* 112 */ "defer_subclause ::= DEFERRABLE init_deferred_pred_opt",
 /* 113 */ "init_deferred_pred_opt ::=",
 /* 114 */ "init_deferred_pred_opt ::= INITIALLY DEFERRED",
 /* 115 */ "init_deferred_pred_opt ::= INITIALLY IMMEDIATE",
 /* 116 */ "conslist_opt ::=",
 /* 117 */ "conslist_opt ::= COMMA conslist",
 /* 118 */ "conslist ::= conslist tconscomma tcons",
 /* 119 */ "conslist ::= tcons",
 /* 120 */ "tconscomma ::= COMMA",
 /* 121 */ "tconscomma ::=",
 /* 122 */ "tcons ::= CONSTRAINT nm",
 /* 123 */ "tcons ::= PRIMARY KEY LP idxlist autoinc RP onconf",
 /* 124 */ "tcons ::= UNIQUE LP idxlist RP onconf",
 /* 125 */ "tcons ::= CHECK LP expr RP onconf",
 /* 126 */ "tcons ::= FOREIGN KEY LP idxlist RP REFERENCES nm idxlist_opt refargs defer_subclause_opt",
 /* 127 */ "tcons ::= CONSTRAINT ID_CONSTR",
 /* 128 */ "tcons ::= FOREIGN KEY LP idxlist RP REFERENCES ID_TAB",
 /* 129 */ "tcons ::= CHECK LP RP onconf",
 /* 130 */ "defer_subclause_opt ::=",
 /* 131 */ "defer_subclause_opt ::= defer_subclause",
 /* 132 */ "onconf ::=",
 /* 133 */ "onconf ::= ON CONFLICT resolvetype",
 /* 134 */ "orconf ::=",
 /* 135 */ "orconf ::= OR resolvetype",
 /* 136 */ "resolvetype ::= raisetype",
 /* 137 */ "resolvetype ::= IGNORE",
 /* 138 */ "resolvetype ::= REPLACE",
 /* 139 */ "cmd ::= DROP TABLE ifexists fullname",
 /* 140 */ "cmd ::= DROP TABLE ifexists nm DOT ID_TAB",
 /* 141 */ "cmd ::= DROP TABLE ifexists ID_DB|ID_TAB",
 /* 142 */ "ifexists ::= IF EXISTS",
 /* 143 */ "ifexists ::=",
 /* 144 */ "cmd ::= CREATE temp VIEW ifnotexists fullname idxlist_opt AS select",
 /* 145 */ "cmd ::= CREATE temp VIEW ifnotexists nm DOT ID_VIEW_NEW",
 /* 146 */ "cmd ::= CREATE temp VIEW ifnotexists ID_DB|ID_VIEW_NEW",
 /* 147 */ "cmd ::= DROP VIEW ifexists fullname",
 /* 148 */ "cmd ::= DROP VIEW ifexists nm DOT ID_VIEW",
 /* 149 */ "cmd ::= DROP VIEW ifexists ID_DB|ID_VIEW",
 /* 150 */ "cmd ::= select_stmt",
 /* 151 */ "select_stmt ::= select",
 /* 152 */ "select ::= with selectnowith",
 /* 153 */ "selectnowith ::= oneselect",
 /* 154 */ "selectnowith ::= selectnowith multiselect_op oneselect",
 /* 155 */ "selectnowith ::= selectnowith multiselect_op values",
 /* 156 */ "selectnowith ::= values",
 /* 157 */ "selectnowith ::= selectnowith COMMA values",
 /* 158 */ "oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt orderby_opt limit_opt",
 /* 159 */ "oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt window_clause orderby_opt limit_opt",
 /* 160 */ "values ::= VALUES LP nexprlist RP",
 /* 161 */ "values ::= values COMMA LP exprlist RP",
 /* 162 */ "multiselect_op ::= UNION",
 /* 163 */ "multiselect_op ::= UNION ALL",
 /* 164 */ "multiselect_op ::= EXCEPT",
 /* 165 */ "multiselect_op ::= INTERSECT",
 /* 166 */ "distinct ::= DISTINCT",
 /* 167 */ "distinct ::= ALL",
 /* 168 */ "distinct ::=",
 /* 169 */ "sclp ::= selcollist COMMA",
 /* 170 */ "sclp ::=",
 /* 171 */ "selcollist ::= sclp expr as",
 /* 172 */ "selcollist ::= sclp STAR",
 /* 173 */ "selcollist ::= sclp tnm DOT STAR",
 /* 174 */ "selcollist ::= sclp",
 /* 175 */ "selcollist ::= sclp ID_TAB DOT STAR",
 /* 176 */ "as ::= AS nm",
 /* 177 */ "as ::= ids",
 /* 178 */ "as ::= AS ID_ALIAS",
 /* 179 */ "as ::= ID_ALIAS",
 /* 180 */ "as ::=",
 /* 181 */ "from ::=",
 /* 182 */ "from ::= FROM joinsrc",
 /* 183 */ "joinsrc ::= singlesrc seltablist",
 /* 184 */ "joinsrc ::=",
 /* 185 */ "seltablist ::= seltablist joinop singlesrc joinconstr_opt",
 /* 186 */ "seltablist ::=",
 /* 187 */ "singlesrc ::= nm dbnm as indexed_opt",
 /* 188 */ "singlesrc ::= LP select RP as",
 /* 189 */ "singlesrc ::= LP joinsrc RP as",
 /* 190 */ "singlesrc ::= nm dbnm LP exprlist RP as",
 /* 191 */ "singlesrc ::=",
 /* 192 */ "singlesrc ::= nm DOT",
 /* 193 */ "singlesrc ::= nm DOT ID_TAB",
 /* 194 */ "singlesrc ::= ID_DB|ID_TAB",
 /* 195 */ "singlesrc ::= nm DOT ID_VIEW",
 /* 196 */ "singlesrc ::= ID_DB|ID_VIEW",
 /* 197 */ "joinconstr_opt ::= ON expr",
 /* 198 */ "joinconstr_opt ::= USING LP idlist RP",
 /* 199 */ "joinconstr_opt ::=",
 /* 200 */ "joinconstr_opt ::= ON CTX_JOIN_EXPR",
 /* 201 */ "dbnm ::=",
 /* 202 */ "dbnm ::= DOT nm",
 /* 203 */ "fullname ::= nm",
 /* 204 */ "fullname ::= nm DOT nm",
 /* 205 */ "xfullname ::= nm",
 /* 206 */ "xfullname ::= nm DOT nm",
 /* 207 */ "xfullname ::= nm DOT nm AS nm",
 /* 208 */ "xfullname ::= nm AS nm",
 /* 209 */ "xfullname ::= nm DOT nm AS ID_ALIAS",
 /* 210 */ "xfullname ::= nm AS ID_ALIAS",
 /* 211 */ "joinop ::= COMMA",
 /* 212 */ "joinop ::= JOIN",
 /* 213 */ "joinop ::= JOIN_KW JOIN",
 /* 214 */ "joinop ::= JOIN_KW nm JOIN",
 /* 215 */ "joinop ::= JOIN_KW nm nm JOIN",
 /* 216 */ "joinop ::= ID_JOIN_OPTS",
 /* 217 */ "indexed_opt ::=",
 /* 218 */ "indexed_opt ::= indexed_by",
 /* 219 */ "indexed_by ::= INDEXED BY nm",
 /* 220 */ "indexed_by ::= NOT INDEXED",
 /* 221 */ "indexed_by ::= INDEXED BY ID_IDX",
 /* 222 */ "orderby_opt ::=",
 /* 223 */ "orderby_opt ::= ORDER BY sortlist",
 /* 224 */ "sortlist ::= sortlist COMMA expr sortorder nulls",
 /* 225 */ "sortlist ::= expr sortorder nulls",
 /* 226 */ "sortorder ::= ASC",
 /* 227 */ "sortorder ::= DESC",
 /* 228 */ "sortorder ::=",
 /* 229 */ "nulls ::= NULLS FIRST",
 /* 230 */ "nulls ::= NULLS LAST",
 /* 231 */ "nulls ::=",
 /* 232 */ "groupby_opt ::=",
 /* 233 */ "groupby_opt ::= GROUP BY nexprlist",
 /* 234 */ "groupby_opt ::= GROUP BY",
 /* 235 */ "having_opt ::=",
 /* 236 */ "having_opt ::= HAVING expr",
 /* 237 */ "limit_opt ::=",
 /* 238 */ "limit_opt ::= LIMIT expr",
 /* 239 */ "limit_opt ::= LIMIT expr OFFSET expr",
 /* 240 */ "limit_opt ::= LIMIT expr COMMA expr",
 /* 241 */ "cmd ::= delete_stmt",
 /* 242 */ "delete_stmt ::= with DELETE FROM xfullname indexed_opt where_opt returning orderby_opt limit_opt",
 /* 243 */ "delete_stmt ::= with DELETE FROM",
 /* 244 */ "delete_stmt ::= with DELETE FROM nm DOT",
 /* 245 */ "delete_stmt ::= with DELETE FROM nm DOT ID_TAB",
 /* 246 */ "delete_stmt ::= with DELETE FROM ID_DB|ID_TAB",
 /* 247 */ "where_opt ::=",
 /* 248 */ "where_opt ::= WHERE expr",
 /* 249 */ "where_opt ::= WHERE",
 /* 250 */ "returning ::=",
 /* 251 */ "returning ::= RETURNING selcollist",
 /* 252 */ "cmd ::= update_stmt",
 /* 253 */ "update_stmt ::= with UPDATE orconf xfullname indexed_opt SET setlist from where_opt returning orderby_opt limit_opt",
 /* 254 */ "update_stmt ::= with UPDATE orconf",
 /* 255 */ "update_stmt ::= with UPDATE orconf nm DOT",
 /* 256 */ "update_stmt ::= with UPDATE orconf nm DOT ID_TAB",
 /* 257 */ "update_stmt ::= with UPDATE orconf ID_DB|ID_TAB",
 /* 258 */ "setlist ::= setlist COMMA nm EQ expr",
 /* 259 */ "setlist ::= setlist COMMA LP idlist RP EQ expr",
 /* 260 */ "setlist ::= nm EQ expr",
 /* 261 */ "setlist ::= LP idlist RP EQ expr",
 /* 262 */ "setlist ::=",
 /* 263 */ "setlist ::= setlist COMMA",
 /* 264 */ "setlist ::= setlist COMMA ID_COL",
 /* 265 */ "setlist ::= ID_COL",
 /* 266 */ "idlist_opt ::=",
 /* 267 */ "idlist_opt ::= LP idlist RP",
 /* 268 */ "idlist ::= idlist COMMA nm",
 /* 269 */ "idlist ::= nm",
 /* 270 */ "idlist ::=",
 /* 271 */ "idlist ::= idlist COMMA ID_COL",
 /* 272 */ "idlist ::= ID_COL",
 /* 273 */ "cmd ::= insert_stmt",
 /* 274 */ "insert_stmt ::= with insert_cmd INTO xfullname idlist_opt select upsert returning",
 /* 275 */ "insert_stmt ::= with insert_cmd INTO xfullname idlist_opt DEFAULT VALUES returning",
 /* 276 */ "insert_stmt ::= with insert_cmd INTO xfullname LP idlist rp_opt",
 /* 277 */ "insert_stmt ::= with insert_cmd INTO",
 /* 278 */ "insert_stmt ::= with insert_cmd INTO nm DOT",
 /* 279 */ "insert_stmt ::= with insert_cmd INTO ID_DB|ID_TAB",
 /* 280 */ "insert_stmt ::= with insert_cmd INTO nm DOT ID_TAB",
 /* 281 */ "insert_cmd ::= INSERT orconf",
 /* 282 */ "insert_cmd ::= REPLACE",
 /* 283 */ "upsert ::=",
 /* 284 */ "upsert ::= ON CONFLICT LP sortlist RP where_opt DO UPDATE SET setlist where_opt",
 /* 285 */ "upsert ::= ON CONFLICT LP sortlist RP where_opt DO NOTHING",
 /* 286 */ "upsert ::= ON CONFLICT DO NOTHING",
 /* 287 */ "exprx ::= expr not_opt IN ID_DB",
 /* 288 */ "exprx ::= expr not_opt IN nm DOT ID_TAB",
 /* 289 */ "exprx ::= ID_DB|ID_TAB|ID_COL|ID_FN",
 /* 290 */ "exprx ::= tnm DOT ID_TAB|ID_COL",
 /* 291 */ "exprx ::= tnm DOT nm DOT ID_COL",
 /* 292 */ "exprx ::= expr COLLATE ID_COLLATE",
 /* 293 */ "exprx ::= RAISE LP raisetype COMMA ID_ERR_MSG RP",
 /* 294 */ "exprx ::= CTIME_KW",
 /* 295 */ "exprx ::= LP nexprlist RP",
 /* 296 */ "exprx ::= tnm",
 /* 297 */ "exprx ::= tnm DOT nm",
 /* 298 */ "exprx ::= tnm DOT",
 /* 299 */ "exprx ::= tnm DOT nm DOT nm",
 /* 300 */ "exprx ::= tnm DOT nm DOT",
 /* 301 */ "exprx ::= VARIABLE",
 /* 302 */ "exprx ::= expr COLLATE ids",
 /* 303 */ "exprx ::= CAST LP expr AS typetoken RP",
 /* 304 */ "exprx ::= id LP distinct exprlist RP",
 /* 305 */ "exprx ::= id LP distinct exprlist ORDER BY sortlist RP",
 /* 306 */ "exprx ::= id LP STAR RP",
 /* 307 */ "exprx ::= expr AND expr",
 /* 308 */ "exprx ::= expr OR expr",
 /* 309 */ "exprx ::= expr LT|GT|GE|LE expr",
 /* 310 */ "exprx ::= expr EQ|NE expr",
 /* 311 */ "exprx ::= expr BITAND|BITOR|LSHIFT|RSHIFT expr",
 /* 312 */ "exprx ::= expr PLUS|MINUS expr",
 /* 313 */ "exprx ::= expr STAR|SLASH|REM expr",
 /* 314 */ "exprx ::= expr CONCAT expr",
 /* 315 */ "exprx ::= expr not_opt likeop expr",
 /* 316 */ "exprx ::= expr not_opt likeop expr ESCAPE expr",
 /* 317 */ "exprx ::= expr ISNULL|NOTNULL",
 /* 318 */ "exprx ::= expr NOT NULL",
 /* 319 */ "exprx ::= expr IS not_opt expr",
 /* 320 */ "exprx ::= expr IS NOT DISTINCT FROM expr",
 /* 321 */ "exprx ::= expr IS DISTINCT FROM expr",
 /* 322 */ "exprx ::= NOT expr",
 /* 323 */ "exprx ::= BITNOT expr",
 /* 324 */ "exprx ::= MINUS expr",
 /* 325 */ "exprx ::= PLUS expr",
 /* 326 */ "exprx ::= expr PTR expr",
 /* 327 */ "exprx ::= expr not_opt BETWEEN expr AND expr",
 /* 328 */ "exprx ::= expr not_opt IN LP exprlist RP",
 /* 329 */ "exprx ::= LP select RP",
 /* 330 */ "exprx ::= expr not_opt IN LP select RP",
 /* 331 */ "exprx ::= expr not_opt IN nm dbnm",
 /* 332 */ "exprx ::= EXISTS LP select RP",
 /* 333 */ "exprx ::= CASE case_operand case_exprlist case_else END",
 /* 334 */ "exprx ::= RAISE LP IGNORE RP",
 /* 335 */ "exprx ::= RAISE LP raisetype COMMA expr RP",
 /* 336 */ "exprx ::= id LP distinct exprlist RP filter_over",
 /* 337 */ "exprx ::= id LP distinct exprlist ORDER BY sortlist RP filter_over",
 /* 338 */ "exprx ::= id LP STAR RP filter_over",
 /* 339 */ "exprx ::= id LP distinct exprlist RP WITHIN GROUP LP ORDER BY expr RP",
 /* 340 */ "expr ::=",
 /* 341 */ "expr ::= exprx",
 /* 342 */ "not_opt ::=",
 /* 343 */ "not_opt ::= NOT",
 /* 344 */ "rp_opt ::=",
 /* 345 */ "rp_opt ::= RP",
 /* 346 */ "likeop ::= LIKE_KW|MATCH",
 /* 347 */ "case_exprlist ::= case_exprlist WHEN expr THEN expr",
 /* 348 */ "case_exprlist ::= WHEN expr THEN expr",
 /* 349 */ "case_else ::= ELSE expr",
 /* 350 */ "case_else ::=",
 /* 351 */ "case_operand ::= exprx",
 /* 352 */ "case_operand ::=",
 /* 353 */ "exprlist ::= nexprlist",
 /* 354 */ "exprlist ::=",
 /* 355 */ "nexprlist ::= nexprlist COMMA expr",
 /* 356 */ "nexprlist ::= exprx",
 /* 357 */ "cmd ::= CREATE uniqueflag INDEX ifnotexists nm dbnm ON nm LP sortlist RP where_opt",
 /* 358 */ "cmd ::= CREATE uniqueflag INDEX ifnotexists nm dbnm ON ID_TAB",
 /* 359 */ "cmd ::= CREATE uniqueflag INDEX ifnotexists nm DOT ID_IDX_NEW",
 /* 360 */ "cmd ::= CREATE uniqueflag INDEX ifnotexists ID_DB|ID_IDX_NEW",
 /* 361 */ "uniqueflag ::= UNIQUE",
 /* 362 */ "uniqueflag ::=",
 /* 363 */ "idxlist_opt ::=",
 /* 364 */ "idxlist_opt ::= LP idxlist RP",
 /* 365 */ "idxlist ::= idxlist COMMA idxlist_single",
 /* 366 */ "idxlist ::= idxlist_single",
 /* 367 */ "idxlist_single ::= nm collate sortorder",
 /* 368 */ "idxlist_single ::= ID_COL",
 /* 369 */ "collate ::=",
 /* 370 */ "collate ::= COLLATE ids",
 /* 371 */ "collate ::= COLLATE ID_COLLATE",
 /* 372 */ "cmd ::= DROP INDEX ifexists fullname",
 /* 373 */ "cmd ::= DROP INDEX ifexists nm DOT ID_IDX",
 /* 374 */ "cmd ::= DROP INDEX ifexists ID_DB|ID_IDX",
 /* 375 */ "cmd ::= VACUUM vinto",
 /* 376 */ "cmd ::= VACUUM nm vinto",
 /* 377 */ "vinto ::= INTO expr",
 /* 378 */ "vinto ::=",
 /* 379 */ "cmd ::= PRAGMA nm dbnm",
 /* 380 */ "cmd ::= PRAGMA nm dbnm EQ nmnum",
 /* 381 */ "cmd ::= PRAGMA nm dbnm LP nmnum RP",
 /* 382 */ "cmd ::= PRAGMA nm dbnm EQ minus_num",
 /* 383 */ "cmd ::= PRAGMA nm dbnm LP minus_num RP",
 /* 384 */ "cmd ::= PRAGMA nm DOT ID_PRAGMA",
 /* 385 */ "cmd ::= PRAGMA ID_DB|ID_PRAGMA",
 /* 386 */ "nmnum ::= plus_num",
 /* 387 */ "nmnum ::= nm",
 /* 388 */ "nmnum ::= ON",
 /* 389 */ "nmnum ::= DELETE",
 /* 390 */ "nmnum ::= DEFAULT",
 /* 391 */ "plus_num ::= PLUS number",
 /* 392 */ "plus_num ::= number",
 /* 393 */ "minus_num ::= MINUS number",
 /* 394 */ "number ::= INTEGER",
 /* 395 */ "number ::= FLOAT",
 /* 396 */ "cmd ::= CREATE temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON nm foreach_clause when_clause BEGIN trigger_cmd_list END",
 /* 397 */ "cmd ::= CREATE temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON nm foreach_clause when_clause",
 /* 398 */ "cmd ::= CREATE temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON nm foreach_clause when_clause BEGIN trigger_cmd_list",
 /* 399 */ "cmd ::= CREATE temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON ID_TAB",
 /* 400 */ "cmd ::= CREATE temp TRIGGER ifnotexists nm DOT ID_TRIG_NEW",
 /* 401 */ "cmd ::= CREATE temp TRIGGER ifnotexists ID_DB|ID_TRIG_NEW",
 /* 402 */ "trigger_time ::= BEFORE",
 /* 403 */ "trigger_time ::= AFTER",
 /* 404 */ "trigger_time ::= INSTEAD OF",
 /* 405 */ "trigger_time ::=",
 /* 406 */ "trigger_event ::= DELETE",
 /* 407 */ "trigger_event ::= INSERT",
 /* 408 */ "trigger_event ::= UPDATE",
 /* 409 */ "trigger_event ::= UPDATE OF idlist",
 /* 410 */ "foreach_clause ::=",
 /* 411 */ "foreach_clause ::= FOR EACH ROW",
 /* 412 */ "when_clause ::=",
 /* 413 */ "when_clause ::= WHEN expr",
 /* 414 */ "trigger_cmd_list ::= trigger_cmd_list trigger_cmd SEMI",
 /* 415 */ "trigger_cmd_list ::= trigger_cmd SEMI",
 /* 416 */ "trigger_cmd_list ::= SEMI",
 /* 417 */ "trigger_cmd ::= update_stmt",
 /* 418 */ "trigger_cmd ::= insert_stmt",
 /* 419 */ "trigger_cmd ::= delete_stmt",
 /* 420 */ "trigger_cmd ::= select_stmt",
 /* 421 */ "raisetype ::= ROLLBACK|ABORT|FAIL",
 /* 422 */ "cmd ::= DROP TRIGGER ifexists fullname",
 /* 423 */ "cmd ::= DROP TRIGGER ifexists nm DOT ID_TRIG",
 /* 424 */ "cmd ::= DROP TRIGGER ifexists ID_DB|ID_TRIG",
 /* 425 */ "cmd ::= ATTACH database_kw_opt expr AS expr key_opt",
 /* 426 */ "cmd ::= DETACH database_kw_opt expr",
 /* 427 */ "key_opt ::=",
 /* 428 */ "key_opt ::= KEY expr",
 /* 429 */ "database_kw_opt ::= DATABASE",
 /* 430 */ "database_kw_opt ::=",
 /* 431 */ "cmd ::= REINDEX",
 /* 432 */ "cmd ::= REINDEX nm dbnm",
 /* 433 */ "cmd ::= REINDEX ID_COLLATE",
 /* 434 */ "cmd ::= REINDEX nm DOT ID_TAB|ID_IDX",
 /* 435 */ "cmd ::= REINDEX ID_DB|ID_IDX|ID_TAB",
 /* 436 */ "cmd ::= ANALYZE",
 /* 437 */ "cmd ::= ANALYZE nm dbnm",
 /* 438 */ "cmd ::= ANALYZE nm DOT ID_TAB|ID_IDX",
 /* 439 */ "cmd ::= ANALYZE ID_DB|ID_IDX|ID_TAB",
 /* 440 */ "cmd ::= ALTER TABLE fullname RENAME TO nm",
 /* 441 */ "cmd ::= ALTER TABLE fullname ADD kwcolumn_opt column",
 /* 442 */ "cmd ::= ALTER TABLE fullname DROP kwcolumn_opt nm",
 /* 443 */ "cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt nm TO nm",
 /* 444 */ "cmd ::= ALTER TABLE fullname ALTER kwcolumn_opt nm SET NOT NULL onconf",
 /* 445 */ "cmd ::= ALTER TABLE fullname ALTER kwcolumn_opt nm DROP NOT NULL",
 /* 446 */ "cmd ::= ALTER TABLE fullname ADD CONSTRAINT nm CHECK LP expr RP onconf",
 /* 447 */ "cmd ::= ALTER TABLE fullname ADD CHECK LP expr RP onconf",
 /* 448 */ "cmd ::= ALTER TABLE fullname DROP CONSTRAINT nm",
 /* 449 */ "cmd ::= ALTER TABLE fullname RENAME TO ID_TAB_NEW",
 /* 450 */ "cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt ID_COL",
 /* 451 */ "cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt",
 /* 452 */ "cmd ::= ALTER TABLE fullname ALTER kwcolumn_opt",
 /* 453 */ "cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt nm TO ID_COL_NEW",
 /* 454 */ "cmd ::= ALTER TABLE nm DOT ID_TAB",
 /* 455 */ "cmd ::= ALTER TABLE fullname ALTER kwcolumn_opt ID_COL",
 /* 456 */ "cmd ::= ALTER TABLE ID_DB|ID_TAB",
 /* 457 */ "kwcolumn_opt ::=",
 /* 458 */ "kwcolumn_opt ::= COLUMNKW",
 /* 459 */ "cmd ::= create_vtab",
 /* 460 */ "create_vtab ::= CREATE VIRTUAL TABLE ifnotexists nm dbnm USING nm",
 /* 461 */ "create_vtab ::= CREATE VIRTUAL TABLE ifnotexists nm dbnm USING nm LP vtabarglist RP",
 /* 462 */ "create_vtab ::= CREATE VIRTUAL TABLE ifnotexists nm DOT ID_TAB_NEW",
 /* 463 */ "create_vtab ::= CREATE VIRTUAL TABLE ifnotexists ID_DB|ID_TAB_NEW",
 /* 464 */ "vtabarglist ::= vtabarg",
 /* 465 */ "vtabarglist ::= vtabarglist COMMA vtabarg",
 /* 466 */ "vtabarg ::=",
 /* 467 */ "vtabarg ::= vtabarg vtabargtoken",
 /* 468 */ "vtabargtoken ::= ANY",
 /* 469 */ "vtabargtoken ::= LP anylist RP",
 /* 470 */ "anylist ::=",
 /* 471 */ "anylist ::= anylist LP anylist RP",
 /* 472 */ "anylist ::= anylist ANY",
 /* 473 */ "with ::=",
 /* 474 */ "with ::= WITH wqlist",
 /* 475 */ "with ::= WITH RECURSIVE wqlist",
 /* 476 */ "wqas ::= AS",
 /* 477 */ "wqas ::= AS MATERIALIZED",
 /* 478 */ "wqas ::= AS NOT MATERIALIZED",
 /* 479 */ "wqlist ::= wqcte",
 /* 480 */ "wqlist ::= wqlist COMMA wqcte",
 /* 481 */ "wqlist ::= ID_TAB_NEW",
 /* 482 */ "wqcte ::= nm idxlist_opt wqas LP select RP",
 /* 483 */ "windowdefn_list ::= windowdefn",
 /* 484 */ "windowdefn_list ::= windowdefn_list COMMA windowdefn",
 /* 485 */ "windowdefn ::= nm AS LP window RP",
 /* 486 */ "window ::= PARTITION BY nexprlist orderby_opt frame_opt",
 /* 487 */ "window ::= nm PARTITION BY nexprlist orderby_opt frame_opt",
 /* 488 */ "window ::= ORDER BY sortlist frame_opt",
 /* 489 */ "window ::= nm ORDER BY sortlist frame_opt",
 /* 490 */ "window ::= frame_opt",
 /* 491 */ "window ::= nm frame_opt",
 /* 492 */ "frame_opt ::=",
 /* 493 */ "frame_opt ::= range_or_rows frame_bound_s frame_exclude_opt",
 /* 494 */ "frame_opt ::= range_or_rows BETWEEN frame_bound_s AND frame_bound_e frame_exclude_opt",
 /* 495 */ "range_or_rows ::= RANGE|ROWS|GROUPS",
 /* 496 */ "frame_bound_s ::= frame_bound",
 /* 497 */ "frame_bound_s ::= UNBOUNDED PRECEDING",
 /* 498 */ "frame_bound_e ::= frame_bound",
 /* 499 */ "frame_bound_e ::= UNBOUNDED FOLLOWING",
 /* 500 */ "frame_bound ::= expr PRECEDING|FOLLOWING",
 /* 501 */ "frame_bound ::= CURRENT ROW",
 /* 502 */ "frame_exclude_opt ::=",
 /* 503 */ "frame_exclude_opt ::= EXCLUDE frame_exclude",
 /* 504 */ "frame_exclude ::= NO OTHERS",
 /* 505 */ "frame_exclude ::= CURRENT ROW",
 /* 506 */ "frame_exclude ::= GROUP",
 /* 507 */ "frame_exclude ::= TIES",
 /* 508 */ "window_clause ::= WINDOW windowdefn_list",
 /* 509 */ "filter_over ::= filter_clause over_clause",
 /* 510 */ "filter_over ::= over_clause",
 /* 511 */ "filter_over ::= filter_clause",
 /* 512 */ "over_clause ::= OVER LP window RP",
 /* 513 */ "over_clause ::= OVER nm",
 /* 514 */ "filter_clause ::= FILTER LP WHERE expr RP",
};
#endif /* NDEBUG */


#if YYSTACKDEPTH<=0
/*
** Try to increase the size of the parser stack.
*/
static void yyGrowStack(yyParser *p){
  int newSize;
  yyStackEntry *pNew;

  newSize = p->yystksz*2 + 100;
  pNew = realloc(p->yystack, newSize*sizeof(pNew[0]));
  if( pNew ){
    p->yystack = pNew;
    p->yystksz = newSize;
#ifndef NDEBUG
    if( yyTraceFILE ){
      fprintf(yyTraceFILE,"%sStack grows to %d entries!\n",
              yyTracePrompt, p->yystksz);
    }
#endif
  }
}
#endif

/*
** This function allocates a new parser.
** The only argument is a pointer to a function which works like
** malloc.
**
** Inputs:
** A pointer to the function used to allocate memory.
**
** Outputs:
** A pointer to a parser.  This pointer is used in subsequent calls
** to sqlite3_parse and sqlite3_parseFree.
*/
void *sqlite3_parseAlloc(void *(*mallocProc)(size_t)){
  yyParser *pParser;
  pParser = (yyParser*)(*mallocProc)( (size_t)sizeof(yyParser) );
  if( pParser ){
    pParser->yyidx = -1;
#ifdef YYTRACKMAXSTACKDEPTH
    pParser->yyidxMax = 0;
#endif
#if YYSTACKDEPTH<=0
    pParser->yystack = NULL;
    pParser->yystksz = 0;
    yyGrowStack(pParser);
#endif
  }
  return pParser;
}

/* The following function deletes the value associated with a
** symbol.  The symbol can be either a terminal or nonterminal.
** "yymajor" is the symbol code, and "yypminor" is a pointer to
** the value.
*/
static void yy_destructor(
  yyParser *yypParser,    /* The parser */
  YYCODETYPE yymajor,     /* Type code for object to destroy */
  YYMINORTYPE *yypminor   /* The object to be destroyed */
){
  sqlite3_parseARG_FETCH;
  if (parserContext->executeRules)
  {
      switch( yymajor ){
        /* Here is inserted the actions which take place when a
        ** terminal or non-terminal is destroyed.  This can happen
        ** when the symbol is popped from the stack during a
        ** reduce or during error processing or when a parser is
        ** being destroyed before it is finished parsing.
        **
        ** Note: during a reduce, the only symbols destroyed are those
        ** which appear on the RHS of the rule, but which are not used
        ** inside the C code.
        */
    case 197: /* cmd */
    case 200: /* ecmd */
    case 202: /* cmdx */
    case 250: /* select_stmt */
    case 281: /* delete_stmt */
    case 283: /* update_stmt */
    case 286: /* insert_stmt */
    case 308: /* trigger_cmd */
    case 312: /* create_vtab */
{
parser_safe_delete((yypminor->yy147));
}
      break;
    case 201: /* explain */
{
parser_safe_delete((yypminor->yy643));
}
      break;
    case 203: /* transtype */
    case 204: /* trans_opt */
{
parser_safe_delete((yypminor->yy284));
}
      break;
    case 205: /* nm */
    case 216: /* columnid */
    case 219: /* id */
    case 220: /* id_opt */
    case 221: /* ids */
    case 223: /* typename */
    case 274: /* dbnm */
    case 299: /* collate */
    case 314: /* vtabarg */
    case 315: /* vtabargtoken */
    case 316: /* anylist */
{
parser_safe_delete((yypminor->yy239));
}
      break;
    case 206: /* savepoint_opt */
    case 208: /* ifnotexists */
    case 232: /* autoinc */
    case 236: /* gen_always */
    case 242: /* tconscomma */
    case 249: /* ifexists */
    case 289: /* rp_opt */
    case 291: /* not_opt */
    case 297: /* uniqueflag */
    case 309: /* database_kw_opt */
    case 311: /* kwcolumn_opt */
{
parser_safe_delete((yypminor->yy515));
}
      break;
    case 207: /* temp */
    case 256: /* distinct */
{
parser_safe_delete((yypminor->yy642));
}
      break;
    case 209: /* fullname */
{
parser_safe_delete((yypminor->yy648));
}
      break;
    case 210: /* columnlist */
{
parser_safe_delete((yypminor->yy426));
}
      break;
    case 211: /* conslist_opt */
    case 241: /* conslist */
{
parser_safe_delete((yypminor->yy645));
}
      break;
    case 212: /* table_options */
{
parser_safe_delete((yypminor->yy497));
}
      break;
    case 213: /* select */
    case 252: /* selectnowith */
{
parser_safe_delete((yypminor->yy113));
}
      break;
    case 214: /* table_option */
{
parser_safe_delete((yypminor->yy303));
}
      break;
    case 215: /* column */
{
parser_safe_delete((yypminor->yy123));
}
      break;
    case 217: /* type */
    case 222: /* typetoken */
{
parser_safe_delete((yypminor->yy1));
}
      break;
    case 218: /* carglist */
{
parser_safe_delete((yypminor->yy11));
}
      break;
    case 224: /* signed */
    case 225: /* plus_num */
    case 226: /* minus_num */
    case 228: /* term */
    case 301: /* nmnum */
    case 302: /* number */
{
parser_safe_delete((yypminor->yy13));
}
      break;
    case 227: /* ccons */
{
parser_safe_delete((yypminor->yy552));
}
      break;
    case 229: /* expr */
    case 259: /* where_opt */
    case 261: /* having_opt */
    case 290: /* exprx */
    case 293: /* case_operand */
    case 295: /* case_else */
    case 300: /* vinto */
    case 306: /* when_clause */
    case 310: /* key_opt */
{
parser_safe_delete((yypminor->yy456));
}
      break;
    case 230: /* onconf */
    case 246: /* resolvetype */
    case 247: /* orconf */
{
parser_safe_delete((yypminor->yy194));
}
      break;
    case 231: /* sortorder */
{
parser_safe_delete((yypminor->yy509));
}
      break;
    case 233: /* idxlist_opt */
    case 244: /* idxlist */
{
parser_safe_delete((yypminor->yy240));
}
      break;
    case 234: /* refargs */
{
parser_safe_delete((yypminor->yy184));
}
      break;
    case 235: /* defer_subclause */
    case 245: /* defer_subclause_opt */
{
parser_safe_delete((yypminor->yy265));
}
      break;
    case 237: /* tnm */
{
parser_safe_delete((yypminor->yy14));
}
      break;
    case 238: /* refarg */
{
parser_safe_delete((yypminor->yy475));
}
      break;
    case 239: /* refact */
{
parser_safe_delete((yypminor->yy160));
}
      break;
    case 240: /* init_deferred_pred_opt */
{
parser_safe_delete((yypminor->yy136));
}
      break;
    case 243: /* tcons */
{
parser_safe_delete((yypminor->yy15));
}
      break;
    case 251: /* with */
{
parser_safe_delete((yypminor->yy137));
}
      break;
    case 253: /* oneselect */
{
parser_safe_delete((yypminor->yy182));
}
      break;
    case 254: /* multiselect_op */
{
parser_safe_delete((yypminor->yy646));
}
      break;
    case 255: /* values */
{
parser_safe_delete((yypminor->yy294));
}
      break;
    case 257: /* selcollist */
    case 267: /* sclp */
    case 282: /* returning */
{
parser_safe_delete((yypminor->yy373));
}
      break;
    case 258: /* from */
    case 269: /* joinsrc */
{
parser_safe_delete((yypminor->yy455));
}
      break;
    case 260: /* groupby_opt */
    case 265: /* nexprlist */
    case 266: /* exprlist */
    case 294: /* case_exprlist */
{
parser_safe_delete((yypminor->yy143));
}
      break;
    case 262: /* orderby_opt */
    case 279: /* sortlist */
{
parser_safe_delete((yypminor->yy459));
}
      break;
    case 263: /* limit_opt */
{
parser_safe_delete((yypminor->yy44));
}
      break;
    case 264: /* window_clause */
    case 320: /* windowdefn_list */
{
parser_safe_delete((yypminor->yy339));
}
      break;
    case 268: /* as */
{
parser_safe_delete((yypminor->yy457));
}
      break;
    case 270: /* singlesrc */
{
parser_safe_delete((yypminor->yy257));
}
      break;
    case 271: /* seltablist */
{
parser_safe_delete((yypminor->yy659));
}
      break;
    case 272: /* joinop */
{
parser_safe_delete((yypminor->yy85));
}
      break;
    case 273: /* joinconstr_opt */
{
parser_safe_delete((yypminor->yy191));
}
      break;
    case 275: /* indexed_opt */
    case 278: /* indexed_by */
{
parser_safe_delete((yypminor->yy504));
}
      break;
    case 276: /* idlist */
    case 285: /* idlist_opt */
    case 313: /* vtabarglist */
{
parser_safe_delete((yypminor->yy159));
}
      break;
    case 277: /* xfullname */
{
parser_safe_delete((yypminor->yy148));
}
      break;
    case 280: /* nulls */
{
parser_safe_delete((yypminor->yy331));
}
      break;
    case 284: /* setlist */
{
parser_safe_delete((yypminor->yy177));
}
      break;
    case 287: /* insert_cmd */
{
parser_safe_delete((yypminor->yy368));
}
      break;
    case 288: /* upsert */
{
parser_safe_delete((yypminor->yy16));
}
      break;
    case 292: /* likeop */
{
parser_safe_delete((yypminor->yy232));
}
      break;
    case 296: /* filter_over */
{
parser_safe_delete((yypminor->yy567));
}
      break;
    case 298: /* idxlist_single */
{
parser_safe_delete((yypminor->yy460));
}
      break;
    case 303: /* trigger_time */
{
parser_safe_delete((yypminor->yy500));
}
      break;
    case 304: /* trigger_event */
{
parser_safe_delete((yypminor->yy87));
}
      break;
    case 305: /* foreach_clause */
{
parser_safe_delete((yypminor->yy115));
}
      break;
    case 307: /* trigger_cmd_list */
{
parser_safe_delete((yypminor->yy198));
}
      break;
    case 317: /* wqlist */
{
parser_safe_delete((yypminor->yy512));
}
      break;
    case 318: /* wqas */
{
parser_safe_delete((yypminor->yy407));
}
      break;
    case 319: /* wqcte */
{
parser_safe_delete((yypminor->yy158));
}
      break;
    case 321: /* windowdefn */
{
parser_safe_delete((yypminor->yy522));
}
      break;
    case 322: /* window */
{
parser_safe_delete((yypminor->yy199));
}
      break;
    case 323: /* frame_opt */
{
parser_safe_delete((yypminor->yy89));
}
      break;
    case 324: /* range_or_rows */
{
parser_safe_delete((yypminor->yy282));
}
      break;
    case 325: /* frame_bound_s */
    case 327: /* frame_bound_e */
{
parser_safe_delete((yypminor->yy169));
}
      break;
    case 328: /* frame_bound */
{
parser_safe_delete((yypminor->yy169));parser_safe_delete((yypminor->yy169));parser_safe_delete((yypminor->yy169));
}
      break;
    case 330: /* filter_clause */
{
parser_safe_delete((yypminor->yy341));
}
      break;
    case 331: /* over_clause */
{
parser_safe_delete((yypminor->yy663));
}
      break;
        default:  break;   /* If no destructor action specified: do nothing */
      }
  }
}

/*
** Pop the parser's stack once.
**
** If there is a destructor routine associated with the token which
** is popped from the stack, then call it.
**
** Return the major token number for the symbol popped.
*/
static int yy_pop_parser_stack(yyParser *pParser){
  YYCODETYPE yymajor;
  yyStackEntry *yytos = &pParser->yystack[pParser->yyidx];

  /* There is no mechanism by which the parser stack can be popped below
  ** empty in SQLite.  */
  if( pParser->yyidx<0 ) return 0;
#ifndef NDEBUG
  if( yyTraceFILE && pParser->yyidx>=0 ){
    fprintf(yyTraceFILE,"%sPopping %s\n",
      yyTracePrompt,
      yyTokenName[yytos->major]);
  }
#endif
  yymajor = yytos->major;
  yy_destructor(pParser, yymajor, &yytos->minor);
  delete yytos->tokens;
  yytos->tokens = nullptr;
  pParser->yyidx--;
  return yymajor;
}

/*
** Deallocate and destroy a parser.  Destructors are all called for
** all stack elements before shutting the parser down.
**
** Inputs:
** <ul>
** <li>  A pointer to the parser.  This should be a pointer
**       obtained from sqlite3_parseAlloc.
** <li>  A pointer to a function used to reclaim memory obtained
**       from malloc.
** </ul>
*/
void sqlite3_parseFree(
  void *p,                    /* The parser to be deleted */
  void (*freeProc)(void*)     /* Function used to reclaim memory */
){
  yyParser *pParser = (yyParser*)p;
  /* In SQLite, we never try to destroy a parser that was not successfully
  ** created in the first place. */
  if( pParser==0 ) return;
  while( pParser->yyidx>=0 ) yy_pop_parser_stack(pParser);
#if YYSTACKDEPTH<=0
  free(pParser->yystack);
#endif
  (*freeProc)((void*)pParser);
}

/*
** Return the peak depth of the stack for a parser.
*/
#ifdef YYTRACKMAXSTACKDEPTH
int sqlite3_parseStackPeak(void *p){
  yyParser *pParser = (yyParser*)p;
  return pParser->yyidxMax;
}
#endif

/*
** Find the appropriate action for a parser given the terminal
** look-ahead token iLookAhead.
**
** If the look-ahead token is YYNOCODE, then check to see if the action is
** independent of the look-ahead.  If it is, return the action, otherwise
** return YY_NO_ACTION.
*/
static int yy_find_shift_action(
  yyParser *pParser,        /* The parser */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
  int stateno = pParser->yystack[pParser->yyidx].stateno;
  GET_CONTEXT;

  if( stateno>YY_SHIFT_COUNT
   || (i = yy_shift_ofst[stateno])==YY_SHIFT_USE_DFLT ){
    return yy_default[stateno];
  }
  assert( iLookAhead!=YYNOCODE );
  i += iLookAhead;
  if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
    if( iLookAhead>0 ){
#ifdef YYFALLBACK
      YYCODETYPE iFallback;            /* Fallback token */
      if( iLookAhead<sizeof(yyFallback)/sizeof(yyFallback[0])
             && (iFallback = yyFallback[iLookAhead])!=0
             && parserContext->doFallbacks ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE, "%sFALLBACK %s => %s\n",
             yyTracePrompt, yyTokenName[iLookAhead], yyTokenName[iFallback]);
        }
#endif
        return yy_find_shift_action(pParser, iFallback);
      }
#endif
#ifdef YYWILDCARD
      {
        int j = i - iLookAhead + YYWILDCARD;
        if(
#if YY_SHIFT_MIN+YYWILDCARD<0
          j>=0 &&
#endif
#if YY_SHIFT_MAX+YYWILDCARD>=YY_ACTTAB_COUNT
          j<YY_ACTTAB_COUNT &&
#endif
          yy_lookahead[j]==YYWILDCARD
        ){
#ifndef NDEBUG
          if( yyTraceFILE ){
            fprintf(yyTraceFILE, "%sWILDCARD %s => %s\n",
               yyTracePrompt, yyTokenName[iLookAhead], yyTokenName[YYWILDCARD]);
          }
#endif /* NDEBUG */
          return yy_action[j];
        }
      }
#endif /* YYWILDCARD */
    }
    return yy_default[stateno];
  }else{
    return yy_action[i];
  }
}

/*
** Find the appropriate action for a parser given the non-terminal
** look-ahead token iLookAhead.
**
** If the look-ahead token is YYNOCODE, then check to see if the action is
** independent of the look-ahead.  If it is, return the action, otherwise
** return YY_NO_ACTION.
*/
static int yy_find_reduce_action(
  int stateno,              /* Current state number */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
#ifdef YYERRORSYMBOL
  if( stateno>YY_REDUCE_COUNT ){
    return yy_default[stateno];
  }
#else
  assert( stateno<=YY_REDUCE_COUNT );
#endif
  i = yy_reduce_ofst[stateno];
  assert( i!=YY_REDUCE_USE_DFLT );
  assert( iLookAhead!=YYNOCODE );
  i += iLookAhead;
#ifdef YYERRORSYMBOL
  if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
    return yy_default[stateno];
  }
#else
  assert( i>=0 && i<YY_ACTTAB_COUNT );
  assert( yy_lookahead[i]==iLookAhead );
#endif
  return yy_action[i];
}

/*
** The following routine is called if the stack overflows.
*/
static void yyStackOverflow(yyParser *yypParser, YYMINORTYPE *yypMinor){
   Q_UNUSED(yypMinor);
   sqlite3_parseARG_FETCH;
   yypParser->yyidx--;
#ifndef NDEBUG
   if( yyTraceFILE ){
     fprintf(yyTraceFILE,"%sStack Overflow!\n",yyTracePrompt);
   }
#endif
   while( yypParser->yyidx>=0 ) yy_pop_parser_stack(yypParser);
   /* Here code is inserted which will execute if the parser
   ** stack every overflows */

    parserContext->error(QObject::tr("Parser stack overflow"));
   sqlite3_parseARG_STORE; /* Suppress warning about unused %extra_argument var */
}

/*
** Perform a shift action.
*/
static void yy_shift(
  yyParser *yypParser,          /* The parser to be shifted */
  int yyNewState,               /* The new state to shift in */
  int yyMajor,                  /* The major token to shift in */
  YYMINORTYPE *yypMinor         /* Pointer to the minor token to shift in */
){
  yyStackEntry *yytos;
  yypParser->yyidx++;
#ifdef YYTRACKMAXSTACKDEPTH
  if( yypParser->yyidx>yypParser->yyidxMax ){
    yypParser->yyidxMax = yypParser->yyidx;
  }
#endif
#if YYSTACKDEPTH>0
  if( yypParser->yyidx>=YYSTACKDEPTH ){
    yyStackOverflow(yypParser, yypMinor);
    return;
  }
#else
  if( yypParser->yyidx>=yypParser->yystksz ){
    yyGrowStack(yypParser);
    if( yypParser->yyidx>=yypParser->yystksz ){
      yyStackOverflow(yypParser, yypMinor);
      return;
    }
  }
#endif
  yytos = &yypParser->yystack[yypParser->yyidx];
  yytos->stateno = (YYACTIONTYPE)yyNewState;
  yytos->major = (YYCODETYPE)yyMajor;
  yytos->minor = *yypMinor;
  yytos->tokens = new QList<Token*>();
#ifndef NDEBUG
  if( yyTraceFILE && yypParser->yyidx>0 ){
    int i;
    fprintf(yyTraceFILE,"%sShift %d\n",yyTracePrompt,yyNewState);
    fprintf(yyTraceFILE,"%sStack:",yyTracePrompt);
    for(i=1; i<=yypParser->yyidx; i++)
      fprintf(yyTraceFILE," %s",yyTokenName[yypParser->yystack[i].major]);
    fprintf(yyTraceFILE,"\n");
  }
#endif
}

/* The following table contains information about every rule that
** is used during the reduce.
*/
static const struct {
  YYCODETYPE lhs;         /* Symbol on the left-hand side of the rule */
  unsigned char nrhs;     /* Number of right-hand side symbols in the rule */
} yyRuleInfo[] = {
  { 198, 1 },
  { 199, 2 },
  { 199, 1 },
  { 200, 1 },
  { 200, 3 },
  { 201, 0 },
  { 201, 1 },
  { 201, 3 },
  { 202, 1 },
  { 197, 3 },
  { 204, 0 },
  { 204, 1 },
  { 204, 2 },
  { 204, 2 },
  { 203, 0 },
  { 203, 1 },
  { 203, 1 },
  { 203, 1 },
  { 197, 2 },
  { 197, 2 },
  { 197, 2 },
  { 206, 1 },
  { 206, 0 },
  { 197, 2 },
  { 197, 3 },
  { 197, 5 },
  { 197, 2 },
  { 197, 3 },
  { 197, 5 },
  { 197, 10 },
  { 197, 7 },
  { 197, 7 },
  { 197, 5 },
  { 212, 0 },
  { 212, 1 },
  { 212, 3 },
  { 214, 2 },
  { 214, 1 },
  { 214, 2 },
  { 214, 1 },
  { 208, 0 },
  { 208, 3 },
  { 207, 1 },
  { 207, 0 },
  { 210, 3 },
  { 210, 1 },
  { 215, 3 },
  { 216, 1 },
  { 216, 1 },
  { 219, 1 },
  { 220, 1 },
  { 220, 0 },
  { 221, 1 },
  { 221, 1 },
  { 205, 1 },
  { 205, 1 },
  { 205, 1 },
  { 217, 0 },
  { 217, 1 },
  { 222, 1 },
  { 222, 4 },
  { 222, 6 },
  { 223, 1 },
  { 223, 2 },
  { 223, 1 },
  { 224, 1 },
  { 224, 1 },
  { 218, 2 },
  { 218, 0 },
  { 227, 2 },
  { 227, 2 },
  { 227, 4 },
  { 227, 3 },
  { 227, 3 },
  { 227, 2 },
  { 227, 2 },
  { 227, 2 },
  { 227, 3 },
  { 227, 5 },
  { 227, 2 },
  { 227, 4 },
  { 227, 4 },
  { 227, 1 },
  { 227, 2 },
  { 227, 6 },
  { 227, 2 },
  { 227, 2 },
  { 227, 2 },
  { 227, 3 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 237, 1 },
  { 237, 1 },
  { 236, 2 },
  { 236, 0 },
  { 232, 0 },
  { 232, 1 },
  { 234, 0 },
  { 234, 2 },
  { 238, 2 },
  { 238, 3 },
  { 238, 3 },
  { 238, 3 },
  { 238, 2 },
  { 239, 2 },
  { 239, 2 },
  { 239, 1 },
  { 239, 1 },
  { 239, 2 },
  { 235, 3 },
  { 235, 2 },
  { 240, 0 },
  { 240, 2 },
  { 240, 2 },
  { 211, 0 },
  { 211, 2 },
  { 241, 3 },
  { 241, 1 },
  { 242, 1 },
  { 242, 0 },
  { 243, 2 },
  { 243, 7 },
  { 243, 5 },
  { 243, 5 },
  { 243, 10 },
  { 243, 2 },
  { 243, 7 },
  { 243, 4 },
  { 245, 0 },
  { 245, 1 },
  { 230, 0 },
  { 230, 3 },
  { 247, 0 },
  { 247, 2 },
  { 246, 1 },
  { 246, 1 },
  { 246, 1 },
  { 197, 4 },
  { 197, 6 },
  { 197, 4 },
  { 249, 2 },
  { 249, 0 },
  { 197, 8 },
  { 197, 7 },
  { 197, 5 },
  { 197, 4 },
  { 197, 6 },
  { 197, 4 },
  { 197, 1 },
  { 250, 1 },
  { 213, 2 },
  { 252, 1 },
  { 252, 3 },
  { 252, 3 },
  { 252, 1 },
  { 252, 3 },
  { 253, 9 },
  { 253, 10 },
  { 255, 4 },
  { 255, 5 },
  { 254, 1 },
  { 254, 2 },
  { 254, 1 },
  { 254, 1 },
  { 256, 1 },
  { 256, 1 },
  { 256, 0 },
  { 267, 2 },
  { 267, 0 },
  { 257, 3 },
  { 257, 2 },
  { 257, 4 },
  { 257, 1 },
  { 257, 4 },
  { 268, 2 },
  { 268, 1 },
  { 268, 2 },
  { 268, 1 },
  { 268, 0 },
  { 258, 0 },
  { 258, 2 },
  { 269, 2 },
  { 269, 0 },
  { 271, 4 },
  { 271, 0 },
  { 270, 4 },
  { 270, 4 },
  { 270, 4 },
  { 270, 6 },
  { 270, 0 },
  { 270, 2 },
  { 270, 3 },
  { 270, 1 },
  { 270, 3 },
  { 270, 1 },
  { 273, 2 },
  { 273, 4 },
  { 273, 0 },
  { 273, 2 },
  { 274, 0 },
  { 274, 2 },
  { 209, 1 },
  { 209, 3 },
  { 277, 1 },
  { 277, 3 },
  { 277, 5 },
  { 277, 3 },
  { 277, 5 },
  { 277, 3 },
  { 272, 1 },
  { 272, 1 },
  { 272, 2 },
  { 272, 3 },
  { 272, 4 },
  { 272, 1 },
  { 275, 0 },
  { 275, 1 },
  { 278, 3 },
  { 278, 2 },
  { 278, 3 },
  { 262, 0 },
  { 262, 3 },
  { 279, 5 },
  { 279, 3 },
  { 231, 1 },
  { 231, 1 },
  { 231, 0 },
  { 280, 2 },
  { 280, 2 },
  { 280, 0 },
  { 260, 0 },
  { 260, 3 },
  { 260, 2 },
  { 261, 0 },
  { 261, 2 },
  { 263, 0 },
  { 263, 2 },
  { 263, 4 },
  { 263, 4 },
  { 197, 1 },
  { 281, 9 },
  { 281, 3 },
  { 281, 5 },
  { 281, 6 },
  { 281, 4 },
  { 259, 0 },
  { 259, 2 },
  { 259, 1 },
  { 282, 0 },
  { 282, 2 },
  { 197, 1 },
  { 283, 12 },
  { 283, 3 },
  { 283, 5 },
  { 283, 6 },
  { 283, 4 },
  { 284, 5 },
  { 284, 7 },
  { 284, 3 },
  { 284, 5 },
  { 284, 0 },
  { 284, 2 },
  { 284, 3 },
  { 284, 1 },
  { 285, 0 },
  { 285, 3 },
  { 276, 3 },
  { 276, 1 },
  { 276, 0 },
  { 276, 3 },
  { 276, 1 },
  { 197, 1 },
  { 286, 8 },
  { 286, 8 },
  { 286, 7 },
  { 286, 3 },
  { 286, 5 },
  { 286, 4 },
  { 286, 6 },
  { 287, 2 },
  { 287, 1 },
  { 288, 0 },
  { 288, 11 },
  { 288, 8 },
  { 288, 4 },
  { 290, 4 },
  { 290, 6 },
  { 290, 1 },
  { 290, 3 },
  { 290, 5 },
  { 290, 3 },
  { 290, 6 },
  { 290, 1 },
  { 290, 3 },
  { 290, 1 },
  { 290, 3 },
  { 290, 2 },
  { 290, 5 },
  { 290, 4 },
  { 290, 1 },
  { 290, 3 },
  { 290, 6 },
  { 290, 5 },
  { 290, 8 },
  { 290, 4 },
  { 290, 3 },
  { 290, 3 },
  { 290, 3 },
  { 290, 3 },
  { 290, 3 },
  { 290, 3 },
  { 290, 3 },
  { 290, 3 },
  { 290, 4 },
  { 290, 6 },
  { 290, 2 },
  { 290, 3 },
  { 290, 4 },
  { 290, 6 },
  { 290, 5 },
  { 290, 2 },
  { 290, 2 },
  { 290, 2 },
  { 290, 2 },
  { 290, 3 },
  { 290, 6 },
  { 290, 6 },
  { 290, 3 },
  { 290, 6 },
  { 290, 5 },
  { 290, 4 },
  { 290, 5 },
  { 290, 4 },
  { 290, 6 },
  { 290, 6 },
  { 290, 9 },
  { 290, 5 },
  { 290, 12 },
  { 229, 0 },
  { 229, 1 },
  { 291, 0 },
  { 291, 1 },
  { 289, 0 },
  { 289, 1 },
  { 292, 1 },
  { 294, 5 },
  { 294, 4 },
  { 295, 2 },
  { 295, 0 },
  { 293, 1 },
  { 293, 0 },
  { 266, 1 },
  { 266, 0 },
  { 265, 3 },
  { 265, 1 },
  { 197, 12 },
  { 197, 8 },
  { 197, 7 },
  { 197, 5 },
  { 297, 1 },
  { 297, 0 },
  { 233, 0 },
  { 233, 3 },
  { 244, 3 },
  { 244, 1 },
  { 298, 3 },
  { 298, 1 },
  { 299, 0 },
  { 299, 2 },
  { 299, 2 },
  { 197, 4 },
  { 197, 6 },
  { 197, 4 },
  { 197, 2 },
  { 197, 3 },
  { 300, 2 },
  { 300, 0 },
  { 197, 3 },
  { 197, 5 },
  { 197, 6 },
  { 197, 5 },
  { 197, 6 },
  { 197, 4 },
  { 197, 2 },
  { 301, 1 },
  { 301, 1 },
  { 301, 1 },
  { 301, 1 },
  { 301, 1 },
  { 225, 2 },
  { 225, 1 },
  { 226, 2 },
  { 302, 1 },
  { 302, 1 },
  { 197, 15 },
  { 197, 12 },
  { 197, 14 },
  { 197, 10 },
  { 197, 7 },
  { 197, 5 },
  { 303, 1 },
  { 303, 1 },
  { 303, 2 },
  { 303, 0 },
  { 304, 1 },
  { 304, 1 },
  { 304, 1 },
  { 304, 3 },
  { 305, 0 },
  { 305, 3 },
  { 306, 0 },
  { 306, 2 },
  { 307, 3 },
  { 307, 2 },
  { 307, 1 },
  { 308, 1 },
  { 308, 1 },
  { 308, 1 },
  { 308, 1 },
  { 248, 1 },
  { 197, 4 },
  { 197, 6 },
  { 197, 4 },
  { 197, 6 },
  { 197, 3 },
  { 310, 0 },
  { 310, 2 },
  { 309, 1 },
  { 309, 0 },
  { 197, 1 },
  { 197, 3 },
  { 197, 2 },
  { 197, 4 },
  { 197, 2 },
  { 197, 1 },
  { 197, 3 },
  { 197, 4 },
  { 197, 2 },
  { 197, 6 },
  { 197, 6 },
  { 197, 6 },
  { 197, 8 },
  { 197, 10 },
  { 197, 9 },
  { 197, 11 },
  { 197, 9 },
  { 197, 6 },
  { 197, 6 },
  { 197, 6 },
  { 197, 5 },
  { 197, 5 },
  { 197, 8 },
  { 197, 5 },
  { 197, 6 },
  { 197, 3 },
  { 311, 0 },
  { 311, 1 },
  { 197, 1 },
  { 312, 8 },
  { 312, 11 },
  { 312, 7 },
  { 312, 5 },
  { 313, 1 },
  { 313, 3 },
  { 314, 0 },
  { 314, 2 },
  { 315, 1 },
  { 315, 3 },
  { 316, 0 },
  { 316, 4 },
  { 316, 2 },
  { 251, 0 },
  { 251, 2 },
  { 251, 3 },
  { 318, 1 },
  { 318, 2 },
  { 318, 3 },
  { 317, 1 },
  { 317, 3 },
  { 317, 1 },
  { 319, 6 },
  { 320, 1 },
  { 320, 3 },
  { 321, 5 },
  { 322, 5 },
  { 322, 6 },
  { 322, 4 },
  { 322, 5 },
  { 322, 1 },
  { 322, 2 },
  { 323, 0 },
  { 323, 3 },
  { 323, 6 },
  { 324, 1 },
  { 325, 1 },
  { 325, 2 },
  { 327, 1 },
  { 327, 2 },
  { 328, 2 },
  { 328, 2 },
  { 326, 0 },
  { 326, 2 },
  { 329, 2 },
  { 329, 2 },
  { 329, 1 },
  { 329, 1 },
  { 264, 2 },
  { 296, 2 },
  { 296, 1 },
  { 296, 1 },
  { 331, 4 },
  { 331, 2 },
  { 330, 5 },
};

static void yy_accept(yyParser*);  /* Forward Declaration */

/*
** Perform a reduce action and the shift that must immediately
** follow the reduce.
*/
static void yy_reduce(
  yyParser *yypParser,         /* The parser */
  int yyruleno                 /* Number of the rule by which to reduce */
){
  int yygoto;                     /* The next state */
  int yyact;                      /* The next action */
  YYMINORTYPE yygotominor;        /* The LHS of the rule reduced */
  yyStackEntry *yymsp;            /* The top of the parser's stack */
  int yysize;                     /* Amount to pop the stack */
  sqlite3_parseARG_FETCH;
  SqliteStatement* objectForTokens = 0;
  QStringList noTokenInheritanceFields;
  yymsp = &yypParser->yystack[yypParser->yyidx];
#ifndef NDEBUG
  if( yyTraceFILE && yyruleno>=0
        && yyruleno<(int)(sizeof(yyRuleName)/sizeof(yyRuleName[0])) ){
    fprintf(yyTraceFILE, "%sReduce [%s].\n", yyTracePrompt,
      yyRuleName[yyruleno]);
  }
#endif /* NDEBUG */

  /* Silence complaints from purify about yygotominor being uninitialized
  ** in some cases when it is copied into the stack after the following
  ** switch.  yygotominor is uninitialized when a rule reduces that does
  ** not set the value of its left-hand side nonterminal.  Leaving the
  ** value of the nonterminal uninitialized is utterly harmless as long
  ** as the value is never used.  So really the only thing this code
  ** accomplishes is to quieten purify.
  **
  ** 2007-01-16:  The wireshark project (www.wireshark.org) reports that
  ** without this code, their parser segfaults.  I'm not sure what there
  ** parser is doing to make this happen.  This is the second bug report
  ** from wireshark this week.  Clearly they are stressing Lemon in ways
  ** that it has not been previously stressed...  (SQLite ticket #2172)
  */
  /*memset(&yygotominor, 0, sizeof(yygotominor));*/
  yygotominor = yyzerominor;


  if (parserContext->executeRules)
  {
      switch( yyruleno ){
      /* Beginning here are the reduction cases.  A typical example
      ** follows:
      **   case 0:
      **  #line <lineno> <grammarfile>
      **     { ... }           // User supplied code
      **  #line <lineno> <thisfile>
      **     break;
      */
      case 1: /* cmdlist ::= cmdlist ecmd */
{parserContext->addQuery(yymsp[0].minor.yy147); DONT_INHERIT_TOKENS("cmdlist");}
        break;
      case 2: /* cmdlist ::= ecmd */
{parserContext->addQuery(yymsp[0].minor.yy147);}
        break;
      case 3: /* ecmd ::= SEMI */
{yygotominor.yy147 = new SqliteEmptyQuery();}
        break;
      case 4: /* ecmd ::= explain cmdx SEMI */
{
                                                yygotominor.yy147 = yymsp[-1].minor.yy147;
                                                yygotominor.yy147->explain = yymsp[-2].minor.yy643->explain;
                                                yygotominor.yy147->queryPlan = yymsp[-2].minor.yy643->queryPlan;
                                                delete yymsp[-2].minor.yy643;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 5: /* explain ::= */
{yygotominor.yy643 = new ParserStubExplain(false, false);}
        break;
      case 6: /* explain ::= EXPLAIN */
{yygotominor.yy643 = new ParserStubExplain(true, false);}
        break;
      case 7: /* explain ::= EXPLAIN QUERY PLAN */
{yygotominor.yy643 = new ParserStubExplain(true, true);}
        break;
      case 8: /* cmdx ::= cmd */
      case 417: /* trigger_cmd ::= update_stmt */ yytestcase(yyruleno==417);
      case 418: /* trigger_cmd ::= insert_stmt */ yytestcase(yyruleno==418);
      case 419: /* trigger_cmd ::= delete_stmt */ yytestcase(yyruleno==419);
      case 420: /* trigger_cmd ::= select_stmt */ yytestcase(yyruleno==420);
      case 459: /* cmd ::= create_vtab */ yytestcase(yyruleno==459);
{yygotominor.yy147 = yymsp[0].minor.yy147;}
        break;
      case 9: /* cmd ::= BEGIN transtype trans_opt */
{
                                                yygotominor.yy147 = new SqliteBeginTrans(
                                                        yymsp[-1].minor.yy284->type,
                                                        yymsp[0].minor.yy284->transactionKw,
                                                        yymsp[0].minor.yy284->name
                                                    );
                                                delete yymsp[0].minor.yy284;
                                                delete yymsp[-1].minor.yy284;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 10: /* trans_opt ::= */
      case 14: /* transtype ::= */ yytestcase(yyruleno==14);
{yygotominor.yy284 = new ParserStubTransDetails();}
        break;
      case 11: /* trans_opt ::= TRANSACTION */
{
                                                yygotominor.yy284 = new ParserStubTransDetails();
                                                yygotominor.yy284->transactionKw = true;
                                            }
        break;
      case 12: /* trans_opt ::= TRANSACTION nm */
      case 13: /* trans_opt ::= TRANSACTION ID_TRANS */ yytestcase(yyruleno==13);
{
                                                yygotominor.yy284 = new ParserStubTransDetails();
                                                yygotominor.yy284->transactionKw = true;
                                                yygotominor.yy284->name = *(yymsp[0].minor.yy239);
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 15: /* transtype ::= DEFERRED */
{
                                                yygotominor.yy284 = new ParserStubTransDetails();
                                                yygotominor.yy284->type = SqliteBeginTrans::Type::DEFERRED;
                                            }
        break;
      case 16: /* transtype ::= IMMEDIATE */
{
                                                yygotominor.yy284 = new ParserStubTransDetails();
                                                yygotominor.yy284->type = SqliteBeginTrans::Type::IMMEDIATE;
                                            }
        break;
      case 17: /* transtype ::= EXCLUSIVE */
{
                                                yygotominor.yy284 = new ParserStubTransDetails();
                                                yygotominor.yy284->type = SqliteBeginTrans::Type::EXCLUSIVE;
                                            }
        break;
      case 18: /* cmd ::= COMMIT trans_opt */
{
                                                yygotominor.yy147 = new SqliteCommitTrans(
                                                        yymsp[0].minor.yy284->transactionKw,
                                                        yymsp[0].minor.yy284->name,
                                                        false
                                                    );
                                                delete yymsp[0].minor.yy284;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 19: /* cmd ::= END trans_opt */
{
                                                yygotominor.yy147 = new SqliteCommitTrans(
                                                        yymsp[0].minor.yy284->transactionKw,
                                                        yymsp[0].minor.yy284->name,
                                                        true
                                                    );
                                                delete yymsp[0].minor.yy284;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 20: /* cmd ::= ROLLBACK trans_opt */
{
                                                yygotominor.yy147 = new SqliteRollback(
                                                        yymsp[0].minor.yy284->transactionKw,
                                                        yymsp[0].minor.yy284->name
                                                    );
                                                delete yymsp[0].minor.yy284;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 21: /* savepoint_opt ::= SAVEPOINT */
      case 41: /* ifnotexists ::= IF NOT EXISTS */ yytestcase(yyruleno==41);
      case 95: /* gen_always ::= GENERATED ALWAYS */ yytestcase(yyruleno==95);
      case 98: /* autoinc ::= AUTOINCR */ yytestcase(yyruleno==98);
      case 120: /* tconscomma ::= COMMA */ yytestcase(yyruleno==120);
      case 142: /* ifexists ::= IF EXISTS */ yytestcase(yyruleno==142);
      case 343: /* not_opt ::= NOT */ yytestcase(yyruleno==343);
      case 345: /* rp_opt ::= RP */ yytestcase(yyruleno==345);
      case 361: /* uniqueflag ::= UNIQUE */ yytestcase(yyruleno==361);
      case 429: /* database_kw_opt ::= DATABASE */ yytestcase(yyruleno==429);
      case 458: /* kwcolumn_opt ::= COLUMNKW */ yytestcase(yyruleno==458);
{yygotominor.yy515 = new bool(true);}
        break;
      case 22: /* savepoint_opt ::= */
      case 40: /* ifnotexists ::= */ yytestcase(yyruleno==40);
      case 96: /* gen_always ::= */ yytestcase(yyruleno==96);
      case 97: /* autoinc ::= */ yytestcase(yyruleno==97);
      case 121: /* tconscomma ::= */ yytestcase(yyruleno==121);
      case 143: /* ifexists ::= */ yytestcase(yyruleno==143);
      case 342: /* not_opt ::= */ yytestcase(yyruleno==342);
      case 344: /* rp_opt ::= */ yytestcase(yyruleno==344);
      case 362: /* uniqueflag ::= */ yytestcase(yyruleno==362);
      case 430: /* database_kw_opt ::= */ yytestcase(yyruleno==430);
      case 457: /* kwcolumn_opt ::= */ yytestcase(yyruleno==457);
{yygotominor.yy515 = new bool(false);}
        break;
      case 23: /* cmd ::= SAVEPOINT nm */
{
                                                yygotominor.yy147 = new SqliteSavepoint(*(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 24: /* cmd ::= RELEASE savepoint_opt nm */
{
                                                yygotominor.yy147 = new SqliteRelease(*(yymsp[-1].minor.yy515), *(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 25: /* cmd ::= ROLLBACK trans_opt TO savepoint_opt nm */
      case 26: /* cmd ::= SAVEPOINT ID_TRANS */ yytestcase(yyruleno==26);
{
                                                yygotominor.yy147 = new SqliteRollback(
                                                        yymsp[-3].minor.yy284->transactionKw,
                                                        *(yymsp[-1].minor.yy515),
                                                        *(yymsp[0].minor.yy239)
                                                    );
                                                delete yymsp[-1].minor.yy515;
                                                delete yymsp[-3].minor.yy284;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 27: /* cmd ::= RELEASE savepoint_opt ID_TRANS */
      case 28: /* cmd ::= ROLLBACK trans_opt TO savepoint_opt ID_TRANS */ yytestcase(yyruleno==28);
{  yy_destructor(yypParser,206,&yymsp[-1].minor);
}
        break;
      case 29: /* cmd ::= CREATE temp TABLE ifnotexists fullname LP columnlist conslist_opt RP table_options */
{
                                                yygotominor.yy147 = new SqliteCreateTable(
                                                        *(yymsp[-6].minor.yy515),
                                                        *(yymsp[-8].minor.yy642),
                                                        yymsp[-5].minor.yy648->name1,
                                                        yymsp[-5].minor.yy648->name2,
                                                        *(yymsp[-3].minor.yy426),
                                                        *(yymsp[-2].minor.yy645),
                                                        *(yymsp[0].minor.yy497)
                                                    );
                                                delete yymsp[-6].minor.yy515;
                                                delete yymsp[-8].minor.yy642;
                                                delete yymsp[-3].minor.yy426;
                                                delete yymsp[-2].minor.yy645;
                                                delete yymsp[-5].minor.yy648;
                                                delete yymsp[0].minor.yy497;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 30: /* cmd ::= CREATE temp TABLE ifnotexists fullname AS select */
{
                                                yygotominor.yy147 = new SqliteCreateTable(
                                                        *(yymsp[-3].minor.yy515),
                                                        *(yymsp[-5].minor.yy642),
                                                        yymsp[-2].minor.yy648->name1,
                                                        yymsp[-2].minor.yy648->name2,
                                                        yymsp[0].minor.yy113
                                                    );
                                                delete yymsp[-3].minor.yy515;
                                                delete yymsp[-5].minor.yy642;
                                                delete yymsp[-2].minor.yy648;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 31: /* cmd ::= CREATE temp TABLE ifnotexists nm DOT ID_TAB_NEW */
      case 145: /* cmd ::= CREATE temp VIEW ifnotexists nm DOT ID_VIEW_NEW */ yytestcase(yyruleno==145);
      case 400: /* cmd ::= CREATE temp TRIGGER ifnotexists nm DOT ID_TRIG_NEW */ yytestcase(yyruleno==400);
{  yy_destructor(yypParser,207,&yymsp[-5].minor);
  yy_destructor(yypParser,205,&yymsp[-2].minor);
}
        break;
      case 32: /* cmd ::= CREATE temp TABLE ifnotexists ID_DB|ID_TAB_NEW */
      case 146: /* cmd ::= CREATE temp VIEW ifnotexists ID_DB|ID_VIEW_NEW */ yytestcase(yyruleno==146);
      case 401: /* cmd ::= CREATE temp TRIGGER ifnotexists ID_DB|ID_TRIG_NEW */ yytestcase(yyruleno==401);
{  yy_destructor(yypParser,207,&yymsp[-3].minor);
}
        break;
      case 33: /* table_options ::= */
{yygotominor.yy497 = new ParserCreateTableOptionList();}
        break;
      case 34: /* table_options ::= table_option */
{
                                                yygotominor.yy497 = new ParserCreateTableOptionList();
                                                yygotominor.yy497->append(yymsp[0].minor.yy303);
                                            }
        break;
      case 35: /* table_options ::= table_options COMMA table_option */
{
                                                yymsp[-2].minor.yy497->append(yymsp[0].minor.yy303);
                                                yygotominor.yy497 = yymsp[-2].minor.yy497;
                                                DONT_INHERIT_TOKENS("table_options");
                                            }
        break;
      case 36: /* table_option ::= WITHOUT nm */
{
                                                if (yymsp[0].minor.yy239->toLower() != "rowid")
                                                    parserContext->errorAtToken(QString("Invalid table option: %1").arg(*(yymsp[0].minor.yy239)));

                                                yygotominor.yy303 = new ParserStubCreateTableOption(ParserStubCreateTableOption::WITHOUT_ROWID);
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 37: /* table_option ::= nm */
      case 38: /* table_option ::= WITHOUT CTX_ROWID_KW */ yytestcase(yyruleno==38);
      case 39: /* table_option ::= CTX_STRICT_KW */ yytestcase(yyruleno==39);
{
                                                if (yymsp[0].minor.yy239->toLower() != "strict")
                                                    parserContext->errorAtToken(QString("Invalid table option: %1").arg(*(yymsp[0].minor.yy239)));

                                                yygotominor.yy303 = new ParserStubCreateTableOption(ParserStubCreateTableOption::STRICT);
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 42: /* temp ::= TEMP */
{yygotominor.yy642 = new int( (yymsp[0].minor.yy0->value.length() > 4) ? 2 : 1 );}
        break;
      case 43: /* temp ::= */
      case 168: /* distinct ::= */ yytestcase(yyruleno==168);
{yygotominor.yy642 = new int(0);}
        break;
      case 44: /* columnlist ::= columnlist COMMA column */
{
                                                yymsp[-2].minor.yy426->append(yymsp[0].minor.yy123);
                                                yygotominor.yy426 = yymsp[-2].minor.yy426;
                                                DONT_INHERIT_TOKENS("columnlist");
                                            }
        break;
      case 45: /* columnlist ::= column */
{
                                                yygotominor.yy426 = new ParserCreateTableColumnList();
                                                yygotominor.yy426->append(yymsp[0].minor.yy123);
                                            }
        break;
      case 46: /* column ::= columnid type carglist */
{
                                                yygotominor.yy123 = new SqliteCreateTable::Column(*(yymsp[-2].minor.yy239), yymsp[-1].minor.yy1, *(yymsp[0].minor.yy11));
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[0].minor.yy11;
                                                objectForTokens = yygotominor.yy123;
                                            }
        break;
      case 47: /* columnid ::= nm */
      case 48: /* columnid ::= ID_COL_NEW */ yytestcase(yyruleno==48);
      case 54: /* nm ::= id */ yytestcase(yyruleno==54);
      case 62: /* typename ::= ids */ yytestcase(yyruleno==62);
      case 202: /* dbnm ::= DOT nm */ yytestcase(yyruleno==202);
      case 370: /* collate ::= COLLATE ids */ yytestcase(yyruleno==370);
      case 371: /* collate ::= COLLATE ID_COLLATE */ yytestcase(yyruleno==371);
{yygotominor.yy239 = yymsp[0].minor.yy239;}
        break;
      case 49: /* id ::= ID|INDEXED|GENERATED */
{
                                                yygotominor.yy239 = new QString(
                                                    stripObjName(
                                                        yymsp[0].minor.yy0->value
                                                    )
                                                );
                                            }
        break;
      case 50: /* id_opt ::= id */
{
                                                yygotominor.yy239 = yymsp[0].minor.yy239;
                                            }
        break;
      case 51: /* id_opt ::= */
{
                                                yygotominor.yy239 = new QString();
                                            }
        break;
      case 52: /* ids ::= ID */
      case 56: /* nm ::= JOIN_KW */ yytestcase(yyruleno==56);
{yygotominor.yy239 = new QString(stripObjName(yymsp[0].minor.yy0->value));}
        break;
      case 53: /* ids ::= STRING */
      case 55: /* nm ::= STRING */ yytestcase(yyruleno==55);
{yygotominor.yy239 = new QString(stripString(yymsp[0].minor.yy0->value));}
        break;
      case 57: /* type ::= */
{yygotominor.yy1 = nullptr;}
        break;
      case 58: /* type ::= typetoken */
{yygotominor.yy1 = yymsp[0].minor.yy1;}
        break;
      case 59: /* typetoken ::= typename */
{
                                                yygotominor.yy1 = new SqliteColumnType(*(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy1;
                                            }
        break;
      case 60: /* typetoken ::= typename LP signed RP */
{
                                                yygotominor.yy1 = new SqliteColumnType(*(yymsp[-3].minor.yy239), *(yymsp[-1].minor.yy13));
                                                delete yymsp[-3].minor.yy239;
                                                delete yymsp[-1].minor.yy13;
                                                objectForTokens = yygotominor.yy1;
                                            }
        break;
      case 61: /* typetoken ::= typename LP signed COMMA signed RP */
{
                                                yygotominor.yy1 = new SqliteColumnType(*(yymsp[-5].minor.yy239), *(yymsp[-3].minor.yy13), *(yymsp[-1].minor.yy13));
                                                delete yymsp[-5].minor.yy239;
                                                delete yymsp[-3].minor.yy13;
                                                delete yymsp[-1].minor.yy13;
                                                objectForTokens = yygotominor.yy1;
                                            }
        break;
      case 63: /* typename ::= typename ids */
      case 64: /* typename ::= ID_COL_TYPE */ yytestcase(yyruleno==64);
{
                                                yymsp[-1].minor.yy239->append(" " + *(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                                yygotominor.yy239 = yymsp[-1].minor.yy239;
                                            }
        break;
      case 65: /* signed ::= plus_num */
      case 66: /* signed ::= minus_num */ yytestcase(yyruleno==66);
      case 386: /* nmnum ::= plus_num */ yytestcase(yyruleno==386);
      case 391: /* plus_num ::= PLUS number */ yytestcase(yyruleno==391);
      case 392: /* plus_num ::= number */ yytestcase(yyruleno==392);
{yygotominor.yy13 = yymsp[0].minor.yy13;}
        break;
      case 67: /* carglist ::= carglist ccons */
{
                                                yymsp[-1].minor.yy11->append(yymsp[0].minor.yy552);
                                                yygotominor.yy11 = yymsp[-1].minor.yy11;
                                                DONT_INHERIT_TOKENS("carglist");
                                            }
        break;
      case 68: /* carglist ::= */
{yygotominor.yy11 = new ParserCreateTableColumnConstraintList();}
        break;
      case 69: /* ccons ::= CONSTRAINT nm */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initDefNameOnly(*(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 70: /* ccons ::= DEFAULT term */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initDefTerm(*(yymsp[0].minor.yy13));
                                                delete yymsp[0].minor.yy13;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 71: /* ccons ::= DEFAULT LP expr RP */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initDefExpr(yymsp[-1].minor.yy456);
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 72: /* ccons ::= DEFAULT PLUS term */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initDefTerm(*(yymsp[0].minor.yy13), false);
                                                delete yymsp[0].minor.yy13;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 73: /* ccons ::= DEFAULT MINUS term */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initDefTerm(*(yymsp[0].minor.yy13), true);
                                                delete yymsp[0].minor.yy13;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 74: /* ccons ::= DEFAULT id */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initDefId(*(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 75: /* ccons ::= DEFAULT CTIME_KW */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initDefCTime(yymsp[0].minor.yy0->value);
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 76: /* ccons ::= NULL onconf */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initNull(*(yymsp[0].minor.yy194));
                                                delete yymsp[0].minor.yy194;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 77: /* ccons ::= NOT NULL onconf */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initNotNull(*(yymsp[0].minor.yy194));
                                                delete yymsp[0].minor.yy194;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 78: /* ccons ::= PRIMARY KEY sortorder onconf autoinc */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initPk(*(yymsp[-2].minor.yy509), *(yymsp[-1].minor.yy194), *(yymsp[0].minor.yy515));
                                                delete yymsp[-2].minor.yy509;
                                                delete yymsp[0].minor.yy515;
                                                delete yymsp[-1].minor.yy194;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 79: /* ccons ::= UNIQUE onconf */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initUnique(*(yymsp[0].minor.yy194));
                                                delete yymsp[0].minor.yy194;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 80: /* ccons ::= CHECK LP expr RP */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initCheck(yymsp[-1].minor.yy456);
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 81: /* ccons ::= REFERENCES nm idxlist_opt refargs */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initFk(*(yymsp[-2].minor.yy239), *(yymsp[-1].minor.yy240), *(yymsp[0].minor.yy184));
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[0].minor.yy184;
                                                delete yymsp[-1].minor.yy240;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 82: /* ccons ::= defer_subclause */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initDefer(yymsp[0].minor.yy265->initially, yymsp[0].minor.yy265->deferrable);
                                                delete yymsp[0].minor.yy265;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 83: /* ccons ::= COLLATE ids */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initColl(*(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 84: /* ccons ::= gen_always AS LP expr RP id_opt */
      case 85: /* ccons ::= CONSTRAINT ID_CONSTR */ yytestcase(yyruleno==85);
      case 86: /* ccons ::= COLLATE ID_COLLATE */ yytestcase(yyruleno==86);
      case 87: /* ccons ::= REFERENCES ID_TAB */ yytestcase(yyruleno==87);
{
                                                if (!yymsp[0].minor.yy239->isNull() && yymsp[0].minor.yy239->toLower() != "stored" && yymsp[0].minor.yy239->toLower() != "virtual")
                                                    parserContext->errorAtToken(QString("Invalid generated column type: %1").arg(*(yymsp[0].minor.yy239)));

                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initGeneratedAs(yymsp[-2].minor.yy456, *(yymsp[-5].minor.yy515), *(yymsp[0].minor.yy239));
                                                delete yymsp[-5].minor.yy515;
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy552;
                                            }
        break;
      case 88: /* ccons ::= CHECK LP RP */
{
                                                yygotominor.yy552 = new SqliteCreateTable::Column::Constraint();
                                                yygotominor.yy552->initCheck();
                                                objectForTokens = yygotominor.yy552;
                                                parserContext->minorErrorAfterLastToken("Syntax error");
                                            }
        break;
      case 89: /* term ::= NULL */
{yygotominor.yy13 = new QVariant();}
        break;
      case 90: /* term ::= INTEGER */
      case 394: /* number ::= INTEGER */ yytestcase(yyruleno==394);
{yygotominor.yy13 = parserContext->handleNumberToken(yymsp[0].minor.yy0->value);}
        break;
      case 91: /* term ::= FLOAT */
      case 395: /* number ::= FLOAT */ yytestcase(yyruleno==395);
{yygotominor.yy13 = new QVariant(QVariant(yymsp[0].minor.yy0->value).toDouble());}
        break;
      case 92: /* term ::= STRING|BLOB */
{
                                                if (yymsp[0].minor.yy0->value.length() >= 3 && yymsp[0].minor.yy0->value.startsWith("x'", Qt::CaseInsensitive))
                                                    yygotominor.yy13 = new QVariant(blobFromLiteral(yymsp[0].minor.yy0->value));
                                                else
                                                    yygotominor.yy13 = new QVariant(stripString(yymsp[0].minor.yy0->value));
                                            }
        break;
      case 93: /* tnm ::= term */
{
												yygotominor.yy14 = new ParserTermOrLiteral(*(yymsp[0].minor.yy13));
												delete yymsp[0].minor.yy13;
											}
        break;
      case 94: /* tnm ::= nm */
{
												yygotominor.yy14 = new ParserTermOrLiteral(*(yymsp[0].minor.yy239));
												delete yymsp[0].minor.yy239;
											}
        break;
      case 99: /* refargs ::= */
{yygotominor.yy184 = new ParserFkConditionList();}
        break;
      case 100: /* refargs ::= refargs refarg */
{
                                                yymsp[-1].minor.yy184->append(yymsp[0].minor.yy475);
                                                yygotominor.yy184 = yymsp[-1].minor.yy184;
                                                DONT_INHERIT_TOKENS("refargs");
                                            }
        break;
      case 101: /* refarg ::= MATCH nm */
{
                                                yygotominor.yy475 = new SqliteForeignKey::Condition(*(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 102: /* refarg ::= ON INSERT refact */
{yygotominor.yy475 = new SqliteForeignKey::Condition(SqliteForeignKey::Condition::INSERT, *(yymsp[0].minor.yy160)); delete yymsp[0].minor.yy160;}
        break;
      case 103: /* refarg ::= ON DELETE refact */
{yygotominor.yy475 = new SqliteForeignKey::Condition(SqliteForeignKey::Condition::DELETE, *(yymsp[0].minor.yy160)); delete yymsp[0].minor.yy160;}
        break;
      case 104: /* refarg ::= ON UPDATE refact */
      case 105: /* refarg ::= MATCH ID_FK_MATCH */ yytestcase(yyruleno==105);
{yygotominor.yy475 = new SqliteForeignKey::Condition(SqliteForeignKey::Condition::UPDATE, *(yymsp[0].minor.yy160)); delete yymsp[0].minor.yy160;}
        break;
      case 106: /* refact ::= SET NULL */
{yygotominor.yy160 = new SqliteForeignKey::Condition::Reaction(SqliteForeignKey::Condition::SET_NULL);}
        break;
      case 107: /* refact ::= SET DEFAULT */
{yygotominor.yy160 = new SqliteForeignKey::Condition::Reaction(SqliteForeignKey::Condition::SET_DEFAULT);}
        break;
      case 108: /* refact ::= CASCADE */
{yygotominor.yy160 = new SqliteForeignKey::Condition::Reaction(SqliteForeignKey::Condition::CASCADE);}
        break;
      case 109: /* refact ::= RESTRICT */
{yygotominor.yy160 = new SqliteForeignKey::Condition::Reaction(SqliteForeignKey::Condition::RESTRICT);}
        break;
      case 110: /* refact ::= NO ACTION */
{yygotominor.yy160 = new SqliteForeignKey::Condition::Reaction(SqliteForeignKey::Condition::NO_ACTION);}
        break;
      case 111: /* defer_subclause ::= NOT DEFERRABLE init_deferred_pred_opt */
{
                                                yygotominor.yy265 = new ParserDeferSubClause(SqliteDeferrable::NOT_DEFERRABLE, *(yymsp[0].minor.yy136));
                                                delete yymsp[0].minor.yy136;
                                            }
        break;
      case 112: /* defer_subclause ::= DEFERRABLE init_deferred_pred_opt */
{
                                                yygotominor.yy265 = new ParserDeferSubClause(SqliteDeferrable::DEFERRABLE, *(yymsp[0].minor.yy136));
                                                delete yymsp[0].minor.yy136;
                                            }
        break;
      case 113: /* init_deferred_pred_opt ::= */
{yygotominor.yy136 = new SqliteInitially(SqliteInitially::null);}
        break;
      case 114: /* init_deferred_pred_opt ::= INITIALLY DEFERRED */
{yygotominor.yy136 = new SqliteInitially(SqliteInitially::DEFERRED);}
        break;
      case 115: /* init_deferred_pred_opt ::= INITIALLY IMMEDIATE */
{yygotominor.yy136 = new SqliteInitially(SqliteInitially::IMMEDIATE);}
        break;
      case 116: /* conslist_opt ::= */
{yygotominor.yy645 = new ParserCreateTableConstraintList();}
        break;
      case 117: /* conslist_opt ::= COMMA conslist */
{yygotominor.yy645 = yymsp[0].minor.yy645;}
        break;
      case 118: /* conslist ::= conslist tconscomma tcons */
{
                                                yymsp[0].minor.yy15->afterComma = *(yymsp[-1].minor.yy515);
                                                yymsp[-2].minor.yy645->append(yymsp[0].minor.yy15);
                                                yygotominor.yy645 = yymsp[-2].minor.yy645;
                                                delete yymsp[-1].minor.yy515;
                                                DONT_INHERIT_TOKENS("conslist");
                                            }
        break;
      case 119: /* conslist ::= tcons */
{
                                                yygotominor.yy645 = new ParserCreateTableConstraintList();
                                                yygotominor.yy645->append(yymsp[0].minor.yy15);
                                            }
        break;
      case 122: /* tcons ::= CONSTRAINT nm */
{
                                                yygotominor.yy15 = new SqliteCreateTable::Constraint();
                                                yygotominor.yy15->initNameOnly(*(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy15;
                                            }
        break;
      case 123: /* tcons ::= PRIMARY KEY LP idxlist autoinc RP onconf */
{
                                                yygotominor.yy15 = new SqliteCreateTable::Constraint();
                                                yygotominor.yy15->initPk(*(yymsp[-3].minor.yy240), *(yymsp[-2].minor.yy515), *(yymsp[0].minor.yy194));
                                                delete yymsp[-2].minor.yy515;
                                                delete yymsp[0].minor.yy194;
                                                delete yymsp[-3].minor.yy240;
                                                objectForTokens = yygotominor.yy15;
                                            }
        break;
      case 124: /* tcons ::= UNIQUE LP idxlist RP onconf */
{
                                                yygotominor.yy15 = new SqliteCreateTable::Constraint();
                                                yygotominor.yy15->initUnique(*(yymsp[-2].minor.yy240), *(yymsp[0].minor.yy194));
                                                delete yymsp[0].minor.yy194;
                                                delete yymsp[-2].minor.yy240;
                                                objectForTokens = yygotominor.yy15;
                                            }
        break;
      case 125: /* tcons ::= CHECK LP expr RP onconf */
{
                                                yygotominor.yy15 = new SqliteCreateTable::Constraint();
                                                yygotominor.yy15->initCheck(yymsp[-2].minor.yy456, *(yymsp[0].minor.yy194));
                                                objectForTokens = yygotominor.yy15;
                                            }
        break;
      case 126: /* tcons ::= FOREIGN KEY LP idxlist RP REFERENCES nm idxlist_opt refargs defer_subclause_opt */
      case 127: /* tcons ::= CONSTRAINT ID_CONSTR */ yytestcase(yyruleno==127);
      case 128: /* tcons ::= FOREIGN KEY LP idxlist RP REFERENCES ID_TAB */ yytestcase(yyruleno==128);
{
                                                yygotominor.yy15 = new SqliteCreateTable::Constraint();
                                                yygotominor.yy15->initFk(
                                                    *(yymsp[-6].minor.yy240),
                                                    *(yymsp[-3].minor.yy239),
                                                    *(yymsp[-2].minor.yy240),
                                                    *(yymsp[-1].minor.yy184),
                                                    yymsp[0].minor.yy265->initially,
                                                    yymsp[0].minor.yy265->deferrable
                                                );
                                                delete yymsp[-3].minor.yy239;
                                                delete yymsp[-1].minor.yy184;
                                                delete yymsp[0].minor.yy265;
                                                delete yymsp[-2].minor.yy240;
                                                delete yymsp[-6].minor.yy240;
                                                objectForTokens = yygotominor.yy15;
                                            }
        break;
      case 129: /* tcons ::= CHECK LP RP onconf */
{
                                                yygotominor.yy15 = new SqliteCreateTable::Constraint();
                                                yygotominor.yy15->initCheck();
                                                objectForTokens = yygotominor.yy15;
                                                parserContext->minorErrorAfterLastToken("Syntax error");
                                              yy_destructor(yypParser,230,&yymsp[0].minor);
}
        break;
      case 130: /* defer_subclause_opt ::= */
{yygotominor.yy265 = new ParserDeferSubClause(SqliteDeferrable::null, SqliteInitially::null);}
        break;
      case 131: /* defer_subclause_opt ::= defer_subclause */
{yygotominor.yy265 = yymsp[0].minor.yy265;}
        break;
      case 132: /* onconf ::= */
      case 134: /* orconf ::= */ yytestcase(yyruleno==134);
{yygotominor.yy194 = new SqliteConflictAlgo(SqliteConflictAlgo::null);}
        break;
      case 133: /* onconf ::= ON CONFLICT resolvetype */
      case 135: /* orconf ::= OR resolvetype */ yytestcase(yyruleno==135);
{yygotominor.yy194 = yymsp[0].minor.yy194;}
        break;
      case 136: /* resolvetype ::= raisetype */
      case 137: /* resolvetype ::= IGNORE */ yytestcase(yyruleno==137);
      case 138: /* resolvetype ::= REPLACE */ yytestcase(yyruleno==138);
{yygotominor.yy194 = new SqliteConflictAlgo(sqliteConflictAlgo(yymsp[0].minor.yy0->value));}
        break;
      case 139: /* cmd ::= DROP TABLE ifexists fullname */
{
                                                yygotominor.yy147 = new SqliteDropTable(*(yymsp[-1].minor.yy515), yymsp[0].minor.yy648->name1, yymsp[0].minor.yy648->name2);
                                                delete yymsp[-1].minor.yy515;
                                                delete yymsp[0].minor.yy648;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 140: /* cmd ::= DROP TABLE ifexists nm DOT ID_TAB */
      case 141: /* cmd ::= DROP TABLE ifexists ID_DB|ID_TAB */ yytestcase(yyruleno==141);
      case 148: /* cmd ::= DROP VIEW ifexists nm DOT ID_VIEW */ yytestcase(yyruleno==148);
      case 149: /* cmd ::= DROP VIEW ifexists ID_DB|ID_VIEW */ yytestcase(yyruleno==149);
      case 193: /* singlesrc ::= nm DOT ID_TAB */ yytestcase(yyruleno==193);
      case 194: /* singlesrc ::= ID_DB|ID_TAB */ yytestcase(yyruleno==194);
      case 195: /* singlesrc ::= nm DOT ID_VIEW */ yytestcase(yyruleno==195);
      case 196: /* singlesrc ::= ID_DB|ID_VIEW */ yytestcase(yyruleno==196);
      case 359: /* cmd ::= CREATE uniqueflag INDEX ifnotexists nm DOT ID_IDX_NEW */ yytestcase(yyruleno==359);
      case 360: /* cmd ::= CREATE uniqueflag INDEX ifnotexists ID_DB|ID_IDX_NEW */ yytestcase(yyruleno==360);
      case 373: /* cmd ::= DROP INDEX ifexists nm DOT ID_IDX */ yytestcase(yyruleno==373);
      case 374: /* cmd ::= DROP INDEX ifexists ID_DB|ID_IDX */ yytestcase(yyruleno==374);
      case 384: /* cmd ::= PRAGMA nm DOT ID_PRAGMA */ yytestcase(yyruleno==384);
      case 385: /* cmd ::= PRAGMA ID_DB|ID_PRAGMA */ yytestcase(yyruleno==385);
      case 423: /* cmd ::= DROP TRIGGER ifexists nm DOT ID_TRIG */ yytestcase(yyruleno==423);
      case 424: /* cmd ::= DROP TRIGGER ifexists ID_DB|ID_TRIG */ yytestcase(yyruleno==424);
      case 434: /* cmd ::= REINDEX nm DOT ID_TAB|ID_IDX */ yytestcase(yyruleno==434);
      case 435: /* cmd ::= REINDEX ID_DB|ID_IDX|ID_TAB */ yytestcase(yyruleno==435);
      case 438: /* cmd ::= ANALYZE nm DOT ID_TAB|ID_IDX */ yytestcase(yyruleno==438);
      case 439: /* cmd ::= ANALYZE ID_DB|ID_IDX|ID_TAB */ yytestcase(yyruleno==439);
      case 454: /* cmd ::= ALTER TABLE nm DOT ID_TAB */ yytestcase(yyruleno==454);
      case 462: /* create_vtab ::= CREATE VIRTUAL TABLE ifnotexists nm DOT ID_TAB_NEW */ yytestcase(yyruleno==462);
      case 463: /* create_vtab ::= CREATE VIRTUAL TABLE ifnotexists ID_DB|ID_TAB_NEW */ yytestcase(yyruleno==463);
{  yy_destructor(yypParser,205,&yymsp[-2].minor);
}
        break;
      case 144: /* cmd ::= CREATE temp VIEW ifnotexists fullname idxlist_opt AS select */
{
                                                yygotominor.yy147 = new SqliteCreateView(*(yymsp[-6].minor.yy642), *(yymsp[-4].minor.yy515), yymsp[-3].minor.yy648->name1, yymsp[-3].minor.yy648->name2, yymsp[0].minor.yy113, *(yymsp[-2].minor.yy240));
                                                delete yymsp[-6].minor.yy642;
                                                delete yymsp[-4].minor.yy515;
                                                delete yymsp[-3].minor.yy648;
                                                delete yymsp[-2].minor.yy240;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 147: /* cmd ::= DROP VIEW ifexists fullname */
{
                                                yygotominor.yy147 = new SqliteDropView(*(yymsp[-1].minor.yy515), yymsp[0].minor.yy648->name1, yymsp[0].minor.yy648->name2);
                                                delete yymsp[-1].minor.yy515;
                                                delete yymsp[0].minor.yy648;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 150: /* cmd ::= select_stmt */
      case 241: /* cmd ::= delete_stmt */ yytestcase(yyruleno==241);
      case 252: /* cmd ::= update_stmt */ yytestcase(yyruleno==252);
      case 273: /* cmd ::= insert_stmt */ yytestcase(yyruleno==273);
{
                                                yygotominor.yy147 = yymsp[0].minor.yy147;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 151: /* select_stmt ::= select */
{
                                                yygotominor.yy147 = yymsp[0].minor.yy113;
                                                // since it's used in trigger:
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 152: /* select ::= with selectnowith */
{
                                                yygotominor.yy113 = yymsp[0].minor.yy113;
                                                yymsp[0].minor.yy113->setWith(yymsp[-1].minor.yy137);
                                                objectForTokens = yygotominor.yy113;
                                            }
        break;
      case 153: /* selectnowith ::= oneselect */
{
                                                yygotominor.yy113 = SqliteSelect::append(yymsp[0].minor.yy182);
                                                objectForTokens = yygotominor.yy113;
                                            }
        break;
      case 154: /* selectnowith ::= selectnowith multiselect_op oneselect */
{
                                                yygotominor.yy113 = SqliteSelect::append(yymsp[-2].minor.yy113, *(yymsp[-1].minor.yy646), yymsp[0].minor.yy182);
                                                delete yymsp[-1].minor.yy646;
                                                objectForTokens = yygotominor.yy113;
                                            }
        break;
      case 155: /* selectnowith ::= selectnowith multiselect_op values */
{
                                                yygotominor.yy113 = SqliteSelect::append(yymsp[-2].minor.yy113, *(yymsp[-1].minor.yy646), *(yymsp[0].minor.yy294));
                                                delete yymsp[-1].minor.yy646;
                                                delete yymsp[0].minor.yy294;
                                                objectForTokens = yygotominor.yy113;
                                            }
        break;
      case 156: /* selectnowith ::= values */
{
                                                yygotominor.yy113 = SqliteSelect::append(*(yymsp[0].minor.yy294));
                                                delete yymsp[0].minor.yy294;
                                                objectForTokens = yygotominor.yy113;
                                            }
        break;
      case 157: /* selectnowith ::= selectnowith COMMA values */
{
                                                yygotominor.yy113 = SqliteSelect::append(yymsp[-2].minor.yy113, SqliteSelect::CompoundOperator::COMMA, *(yymsp[0].minor.yy294));
                                                delete yymsp[0].minor.yy294;
                                                objectForTokens = yygotominor.yy113;
                                            }
        break;
      case 158: /* oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt orderby_opt limit_opt */
{
                                                yygotominor.yy182 = new SqliteSelect::Core(
                                                        *(yymsp[-7].minor.yy642),
                                                        *(yymsp[-6].minor.yy373),
                                                        yymsp[-5].minor.yy455,
                                                        yymsp[-4].minor.yy456,
                                                        *(yymsp[-3].minor.yy143),
                                                        yymsp[-2].minor.yy456,
                                                        *(yymsp[-1].minor.yy459),
                                                        yymsp[0].minor.yy44
                                                    );
                                                delete yymsp[-6].minor.yy373;
                                                delete yymsp[-7].minor.yy642;
                                                delete yymsp[-3].minor.yy143;
                                                delete yymsp[-1].minor.yy459;
                                                objectForTokens = yygotominor.yy182;
                                            }
        break;
      case 159: /* oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt window_clause orderby_opt limit_opt */
{
                                                yygotominor.yy182 = new SqliteSelect::Core(
                                                        *(yymsp[-8].minor.yy642),
                                                        *(yymsp[-7].minor.yy373),
                                                        yymsp[-6].minor.yy455,
                                                        yymsp[-5].minor.yy456,
                                                        *(yymsp[-4].minor.yy143),
                                                        yymsp[-3].minor.yy456,
                                                        *(yymsp[-2].minor.yy339),
                                                        *(yymsp[-1].minor.yy459),
                                                        yymsp[0].minor.yy44
                                                    );
                                                delete yymsp[-7].minor.yy373;
                                                delete yymsp[-8].minor.yy642;
                                                delete yymsp[-4].minor.yy143;
                                                delete yymsp[-1].minor.yy459;
                                                delete yymsp[-2].minor.yy339;
                                                objectForTokens = yygotominor.yy182;
                                            }
        break;
      case 160: /* values ::= VALUES LP nexprlist RP */
{
                                                yygotominor.yy294 = new ParserExprNestedList();
                                                yygotominor.yy294->append(*(yymsp[-1].minor.yy143));
                                                delete yymsp[-1].minor.yy143;
                                            }
        break;
      case 161: /* values ::= values COMMA LP exprlist RP */
{
                                                yymsp[-4].minor.yy294->append(*(yymsp[-1].minor.yy143));
                                                yygotominor.yy294 = yymsp[-4].minor.yy294;
                                                delete yymsp[-1].minor.yy143;
                                                DONT_INHERIT_TOKENS("values");
                                            }
        break;
      case 162: /* multiselect_op ::= UNION */
{yygotominor.yy646 = new SqliteSelect::CompoundOperator(SqliteSelect::CompoundOperator::UNION);}
        break;
      case 163: /* multiselect_op ::= UNION ALL */
{yygotominor.yy646 = new SqliteSelect::CompoundOperator(SqliteSelect::CompoundOperator::UNION_ALL);}
        break;
      case 164: /* multiselect_op ::= EXCEPT */
{yygotominor.yy646 = new SqliteSelect::CompoundOperator(SqliteSelect::CompoundOperator::EXCEPT);}
        break;
      case 165: /* multiselect_op ::= INTERSECT */
{yygotominor.yy646 = new SqliteSelect::CompoundOperator(SqliteSelect::CompoundOperator::INTERSECT);}
        break;
      case 166: /* distinct ::= DISTINCT */
{yygotominor.yy642 = new int(1);}
        break;
      case 167: /* distinct ::= ALL */
{yygotominor.yy642 = new int(2);}
        break;
      case 169: /* sclp ::= selcollist COMMA */
{yygotominor.yy373 = yymsp[-1].minor.yy373;}
        break;
      case 170: /* sclp ::= */
      case 250: /* returning ::= */ yytestcase(yyruleno==250);
{yygotominor.yy373 = new ParserResultColumnList();}
        break;
      case 171: /* selcollist ::= sclp expr as */
{
                                                SqliteSelect::Core::ResultColumn* obj =
                                                    new SqliteSelect::Core::ResultColumn(
                                                        yymsp[-1].minor.yy456,
                                                        yymsp[0].minor.yy457 ? yymsp[0].minor.yy457->asKw : false,
                                                        yymsp[0].minor.yy457 ? yymsp[0].minor.yy457->name : QString()
                                                    );

                                                yymsp[-2].minor.yy373->append(obj);
                                                yygotominor.yy373 = yymsp[-2].minor.yy373;
                                                delete yymsp[0].minor.yy457;
                                                objectForTokens = obj;
                                                DONT_INHERIT_TOKENS("sclp");
                                            }
        break;
      case 172: /* selcollist ::= sclp STAR */
{
                                                SqliteSelect::Core::ResultColumn* obj =
                                                    new SqliteSelect::Core::ResultColumn(true);

                                                yymsp[-1].minor.yy373->append(obj);
                                                yygotominor.yy373 = yymsp[-1].minor.yy373;
                                                objectForTokens = obj;
                                                DONT_INHERIT_TOKENS("sclp");
                                            }
        break;
      case 173: /* selcollist ::= sclp tnm DOT STAR */
{
                                                SqliteSelect::Core::ResultColumn* obj =
                                                    new SqliteSelect::Core::ResultColumn(
                                                        true,
                                                        yymsp[-2].minor.yy14->toName()
                                                    );
													
												if (!yymsp[-2].minor.yy14->isName())
													parserContext->errorAtToken("Syntax error <expected name, not literal value>", -3);
													
                                                yymsp[-3].minor.yy373->append(obj);
                                                yygotominor.yy373 = yymsp[-3].minor.yy373;
                                                delete yymsp[-2].minor.yy14;
                                                objectForTokens = obj;
                                                DONT_INHERIT_TOKENS("sclp");
                                            }
        break;
      case 174: /* selcollist ::= sclp */
      case 175: /* selcollist ::= sclp ID_TAB DOT STAR */ yytestcase(yyruleno==175);
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy373 = yymsp[0].minor.yy373;
                                            }
        break;
      case 176: /* as ::= AS nm */
{
                                                yygotominor.yy457 = new ParserStubAlias(*(yymsp[0].minor.yy239), true);
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 177: /* as ::= ids */
      case 178: /* as ::= AS ID_ALIAS */ yytestcase(yyruleno==178);
      case 179: /* as ::= ID_ALIAS */ yytestcase(yyruleno==179);
{
                                                yygotominor.yy457 = new ParserStubAlias(*(yymsp[0].minor.yy239), false);
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 180: /* as ::= */
{yygotominor.yy457 = nullptr;}
        break;
      case 181: /* from ::= */
{yygotominor.yy455 = nullptr;}
        break;
      case 182: /* from ::= FROM joinsrc */
{yygotominor.yy455 = yymsp[0].minor.yy455;}
        break;
      case 183: /* joinsrc ::= singlesrc seltablist */
{
                                                yygotominor.yy455 = new SqliteSelect::Core::JoinSource(
                                                        yymsp[-1].minor.yy257,
                                                        *(yymsp[0].minor.yy659)
                                                    );
                                                delete yymsp[0].minor.yy659;
                                                objectForTokens = yygotominor.yy455;
                                            }
        break;
      case 184: /* joinsrc ::= */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy455 = new SqliteSelect::Core::JoinSource();
                                                objectForTokens = yygotominor.yy455;
                                            }
        break;
      case 185: /* seltablist ::= seltablist joinop singlesrc joinconstr_opt */
{
                                                SqliteSelect::Core::JoinSourceOther* src =
                                                    new SqliteSelect::Core::JoinSourceOther(yymsp[-2].minor.yy85, yymsp[-1].minor.yy257, yymsp[0].minor.yy191);

                                                yymsp[-3].minor.yy659->append(src);
                                                yygotominor.yy659 = yymsp[-3].minor.yy659;
                                                objectForTokens = src;
                                                DONT_INHERIT_TOKENS("seltablist");
                                            }
        break;
      case 186: /* seltablist ::= */
{
                                                yygotominor.yy659 = new ParserOtherSourceList();
                                            }
        break;
      case 187: /* singlesrc ::= nm dbnm as indexed_opt */
{
                                                yygotominor.yy257 = new SqliteSelect::Core::SingleSource(
                                                        *(yymsp[-3].minor.yy239),
                                                        *(yymsp[-2].minor.yy239),
                                                        yymsp[-1].minor.yy457 ? yymsp[-1].minor.yy457->asKw : false,
                                                        yymsp[-1].minor.yy457 ? yymsp[-1].minor.yy457->name : QString(),
                                                        yymsp[0].minor.yy504 ? yymsp[0].minor.yy504->notIndexedKw : false,
                                                        yymsp[0].minor.yy504 ? yymsp[0].minor.yy504->indexedBy : QString()
                                                    );
                                                delete yymsp[-3].minor.yy239;
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[-1].minor.yy457;
                                                if (yymsp[0].minor.yy504)
                                                    delete yymsp[0].minor.yy504;
                                                objectForTokens = yygotominor.yy257;
                                            }
        break;
      case 188: /* singlesrc ::= LP select RP as */
{
                                                yygotominor.yy257 = new SqliteSelect::Core::SingleSource(
                                                        yymsp[-2].minor.yy113,
                                                        yymsp[0].minor.yy457 ? yymsp[0].minor.yy457->asKw : false,
                                                        yymsp[0].minor.yy457 ? yymsp[0].minor.yy457->name : QString()
                                                    );
                                                delete yymsp[0].minor.yy457;
                                                objectForTokens = yygotominor.yy257;
                                            }
        break;
      case 189: /* singlesrc ::= LP joinsrc RP as */
{
                                                yygotominor.yy257 = new SqliteSelect::Core::SingleSource(
                                                        yymsp[-2].minor.yy455,
                                                        yymsp[0].minor.yy457 ? yymsp[0].minor.yy457->asKw : false,
                                                        yymsp[0].minor.yy457 ? yymsp[0].minor.yy457->name : QString()
                                                    );
                                                delete yymsp[0].minor.yy457;
                                                objectForTokens = yygotominor.yy257;
                                            }
        break;
      case 190: /* singlesrc ::= nm dbnm LP exprlist RP as */
{
                                                yygotominor.yy257 = new SqliteSelect::Core::SingleSource(
                                                        *(yymsp[-5].minor.yy239),
                                                        *(yymsp[-4].minor.yy239),
                                                        yymsp[0].minor.yy457 ? yymsp[0].minor.yy457->asKw : false,
                                                        yymsp[0].minor.yy457 ? yymsp[0].minor.yy457->name : QString(),
                                                        *(yymsp[-2].minor.yy143)
                                                    );
                                                delete yymsp[-5].minor.yy239;
                                                delete yymsp[-4].minor.yy239;
                                                delete yymsp[0].minor.yy457;
                                                if (yymsp[-2].minor.yy143)
                                                    delete yymsp[-2].minor.yy143;

                                                objectForTokens = yygotominor.yy257;
                                            }
        break;
      case 191: /* singlesrc ::= */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy257 = new SqliteSelect::Core::SingleSource();
                                                objectForTokens = yygotominor.yy257;
                                            }
        break;
      case 192: /* singlesrc ::= nm DOT */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy257 = new SqliteSelect::Core::SingleSource();
                                                yygotominor.yy257->database = *(yymsp[-1].minor.yy239);
                                                delete yymsp[-1].minor.yy239;
                                                objectForTokens = yygotominor.yy257;
                                            }
        break;
      case 197: /* joinconstr_opt ::= ON expr */
{
                                                yygotominor.yy191 = new SqliteSelect::Core::JoinConstraint(yymsp[0].minor.yy456);
                                                objectForTokens = yygotominor.yy191;
                                            }
        break;
      case 198: /* joinconstr_opt ::= USING LP idlist RP */
{
                                                yygotominor.yy191 = new SqliteSelect::Core::JoinConstraint(*(yymsp[-1].minor.yy159));
                                                delete yymsp[-1].minor.yy159;
                                                objectForTokens = yygotominor.yy191;
                                            }
        break;
      case 199: /* joinconstr_opt ::= */
{yygotominor.yy191 = nullptr;}
        break;
      case 200: /* joinconstr_opt ::= ON CTX_JOIN_EXPR */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy191 = new SqliteSelect::Core::JoinConstraint();
                                                objectForTokens = yygotominor.yy191;
                                            }
        break;
      case 201: /* dbnm ::= */
      case 369: /* collate ::= */ yytestcase(yyruleno==369);
      case 466: /* vtabarg ::= */ yytestcase(yyruleno==466);
      case 470: /* anylist ::= */ yytestcase(yyruleno==470);
{yygotominor.yy239 = new QString();}
        break;
      case 203: /* fullname ::= nm */
{
                                                yygotominor.yy648 = new ParserFullName();
                                                yygotominor.yy648->name1 = *(yymsp[0].minor.yy239);
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 204: /* fullname ::= nm DOT nm */
{
                                                yygotominor.yy648 = new ParserFullName();
                                                yygotominor.yy648->name1 = *(yymsp[-2].minor.yy239);
                                                yygotominor.yy648->name2 = *(yymsp[0].minor.yy239);
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 205: /* xfullname ::= nm */
{
                                                yygotominor.yy148 = new ParserXFullName();
                                                yygotominor.yy148->name1 = *(yymsp[0].minor.yy239);
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 206: /* xfullname ::= nm DOT nm */
{
                                                yygotominor.yy148 = new ParserXFullName();
                                                yygotominor.yy148->name1 = *(yymsp[-2].minor.yy239);
                                                yygotominor.yy148->name2 = *(yymsp[0].minor.yy239);
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 207: /* xfullname ::= nm DOT nm AS nm */
{
                                                yygotominor.yy148 = new ParserXFullName();
                                                yygotominor.yy148->name1 = *(yymsp[-4].minor.yy239);
                                                yygotominor.yy148->name2 = *(yymsp[-2].minor.yy239);
                                                yygotominor.yy148->alias = *(yymsp[0].minor.yy239);
                                                delete yymsp[-4].minor.yy239;
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 208: /* xfullname ::= nm AS nm */
{
                                                yygotominor.yy148 = new ParserXFullName();
                                                yygotominor.yy148->name1 = *(yymsp[-2].minor.yy239);
                                                yygotominor.yy148->alias = *(yymsp[0].minor.yy239);
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 209: /* xfullname ::= nm DOT nm AS ID_ALIAS */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy148 = new ParserXFullName();
                                                yygotominor.yy148->name1 = *(yymsp[-4].minor.yy239);
                                                yygotominor.yy148->name2 = *(yymsp[-2].minor.yy239);
                                                delete yymsp[-4].minor.yy239;
                                                delete yymsp[-2].minor.yy239;
                                            }
        break;
      case 210: /* xfullname ::= nm AS ID_ALIAS */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy148 = new ParserXFullName();
                                                yygotominor.yy148->name1 = *(yymsp[-2].minor.yy239);
                                                delete yymsp[-2].minor.yy239;
                                            }
        break;
      case 211: /* joinop ::= COMMA */
{
                                                yygotominor.yy85 = new SqliteSelect::Core::JoinOp(true);
                                                objectForTokens = yygotominor.yy85;
                                            }
        break;
      case 212: /* joinop ::= JOIN */
{
                                                yygotominor.yy85 = new SqliteSelect::Core::JoinOp(false);
                                                objectForTokens = yygotominor.yy85;
                                            }
        break;
      case 213: /* joinop ::= JOIN_KW JOIN */
{
                                                yygotominor.yy85 = new SqliteSelect::Core::JoinOp(yymsp[-1].minor.yy0->value);
                                                objectForTokens = yygotominor.yy85;
                                            }
        break;
      case 214: /* joinop ::= JOIN_KW nm JOIN */
{
                                                yygotominor.yy85 = new SqliteSelect::Core::JoinOp(yymsp[-2].minor.yy0->value, *(yymsp[-1].minor.yy239));
                                                delete yymsp[-1].minor.yy239;
                                                objectForTokens = yygotominor.yy85;
                                            }
        break;
      case 215: /* joinop ::= JOIN_KW nm nm JOIN */
      case 216: /* joinop ::= ID_JOIN_OPTS */ yytestcase(yyruleno==216);
{
                                                yygotominor.yy85 = new SqliteSelect::Core::JoinOp(yymsp[-3].minor.yy0->value, *(yymsp[-2].minor.yy239), *(yymsp[-1].minor.yy239));
                                                delete yymsp[-2].minor.yy239;
                                                objectForTokens = yygotominor.yy85;
                                            }
        break;
      case 217: /* indexed_opt ::= */
{yygotominor.yy504 = nullptr;}
        break;
      case 218: /* indexed_opt ::= indexed_by */
{yygotominor.yy504 = yymsp[0].minor.yy504;}
        break;
      case 219: /* indexed_by ::= INDEXED BY nm */
{
                                                yygotominor.yy504 = new ParserIndexedBy(*(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 220: /* indexed_by ::= NOT INDEXED */
{yygotominor.yy504 = new ParserIndexedBy(true);}
        break;
      case 221: /* indexed_by ::= INDEXED BY ID_IDX */
{parserContext->minorErrorBeforeNextToken("Syntax error");}
        break;
      case 222: /* orderby_opt ::= */
{yygotominor.yy459 = new ParserOrderByList();}
        break;
      case 223: /* orderby_opt ::= ORDER BY sortlist */
{yygotominor.yy459 = yymsp[0].minor.yy459;}
        break;
      case 224: /* sortlist ::= sortlist COMMA expr sortorder nulls */
{
                                                SqliteOrderBy* obj = new SqliteOrderBy(yymsp[-2].minor.yy456, *(yymsp[-1].minor.yy509), *(yymsp[0].minor.yy331));
                                                yymsp[-4].minor.yy459->append(obj);
                                                yygotominor.yy459 = yymsp[-4].minor.yy459;
                                                delete yymsp[-1].minor.yy509;
												delete yymsp[0].minor.yy331;
                                                objectForTokens = obj;
                                                DONT_INHERIT_TOKENS("sortlist");
                                            }
        break;
      case 225: /* sortlist ::= expr sortorder nulls */
{
                                                SqliteOrderBy* obj = new SqliteOrderBy(yymsp[-2].minor.yy456, *(yymsp[-1].minor.yy509), *(yymsp[0].minor.yy331));
                                                yygotominor.yy459 = new ParserOrderByList();
                                                yygotominor.yy459->append(obj);
                                                delete yymsp[-1].minor.yy509;
												delete yymsp[0].minor.yy331;
                                                objectForTokens = obj;
                                            }
        break;
      case 226: /* sortorder ::= ASC */
{yygotominor.yy509 = new SqliteSortOrder(SqliteSortOrder::ASC);}
        break;
      case 227: /* sortorder ::= DESC */
{yygotominor.yy509 = new SqliteSortOrder(SqliteSortOrder::DESC);}
        break;
      case 228: /* sortorder ::= */
{yygotominor.yy509 = new SqliteSortOrder(SqliteSortOrder::null);}
        break;
      case 229: /* nulls ::= NULLS FIRST */
{yygotominor.yy331 = new SqliteNulls(SqliteNulls::FIRST);}
        break;
      case 230: /* nulls ::= NULLS LAST */
{yygotominor.yy331 = new SqliteNulls(SqliteNulls::LAST);}
        break;
      case 231: /* nulls ::= */
{yygotominor.yy331 = new SqliteNulls(SqliteNulls::null);}
        break;
      case 232: /* groupby_opt ::= */
      case 354: /* exprlist ::= */ yytestcase(yyruleno==354);
{yygotominor.yy143 = new ParserExprList();}
        break;
      case 233: /* groupby_opt ::= GROUP BY nexprlist */
      case 353: /* exprlist ::= nexprlist */ yytestcase(yyruleno==353);
{yygotominor.yy143 = yymsp[0].minor.yy143;}
        break;
      case 234: /* groupby_opt ::= GROUP BY */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy143 = new ParserExprList();
                                            }
        break;
      case 235: /* having_opt ::= */
      case 247: /* where_opt ::= */ yytestcase(yyruleno==247);
      case 350: /* case_else ::= */ yytestcase(yyruleno==350);
      case 352: /* case_operand ::= */ yytestcase(yyruleno==352);
      case 378: /* vinto ::= */ yytestcase(yyruleno==378);
      case 412: /* when_clause ::= */ yytestcase(yyruleno==412);
      case 427: /* key_opt ::= */ yytestcase(yyruleno==427);
{yygotominor.yy456 = nullptr;}
        break;
      case 236: /* having_opt ::= HAVING expr */
      case 248: /* where_opt ::= WHERE expr */ yytestcase(yyruleno==248);
      case 341: /* expr ::= exprx */ yytestcase(yyruleno==341);
      case 349: /* case_else ::= ELSE expr */ yytestcase(yyruleno==349);
      case 351: /* case_operand ::= exprx */ yytestcase(yyruleno==351);
      case 377: /* vinto ::= INTO expr */ yytestcase(yyruleno==377);
      case 413: /* when_clause ::= WHEN expr */ yytestcase(yyruleno==413);
      case 428: /* key_opt ::= KEY expr */ yytestcase(yyruleno==428);
{yygotominor.yy456 = yymsp[0].minor.yy456;}
        break;
      case 237: /* limit_opt ::= */
{yygotominor.yy44 = nullptr;}
        break;
      case 238: /* limit_opt ::= LIMIT expr */
{
                                                yygotominor.yy44 = new SqliteLimit(yymsp[0].minor.yy456);
                                                objectForTokens = yygotominor.yy44;
                                            }
        break;
      case 239: /* limit_opt ::= LIMIT expr OFFSET expr */
{
                                                yygotominor.yy44 = new SqliteLimit(yymsp[-2].minor.yy456, yymsp[0].minor.yy456, true);
                                                objectForTokens = yygotominor.yy44;
                                            }
        break;
      case 240: /* limit_opt ::= LIMIT expr COMMA expr */
{
                                                yygotominor.yy44 = new SqliteLimit(yymsp[-2].minor.yy456, yymsp[0].minor.yy456, false);
                                                objectForTokens = yygotominor.yy44;
                                            }
        break;
      case 242: /* delete_stmt ::= with DELETE FROM xfullname indexed_opt where_opt returning orderby_opt limit_opt */
{
                                                if (yymsp[-4].minor.yy504)
                                                {
                                                    if (!yymsp[-4].minor.yy504->indexedBy.isNull())
                                                    {
                                                        yygotominor.yy147 = new SqliteDelete(
                                                                yymsp[-5].minor.yy148->name1,
                                                                yymsp[-5].minor.yy148->name2,
                                                                yymsp[-5].minor.yy148->alias,
                                                                yymsp[-4].minor.yy504->indexedBy,
                                                                yymsp[-3].minor.yy456,
                                                                yymsp[-8].minor.yy137,
                                                                *(yymsp[-2].minor.yy373),
                                                                *(yymsp[-1].minor.yy459),
                                                                yymsp[0].minor.yy44
                                                            );
                                                    }
                                                    else
                                                    {
                                                        yygotominor.yy147 = new SqliteDelete(
                                                                yymsp[-5].minor.yy148->name1,
                                                                yymsp[-5].minor.yy148->name2,
                                                                yymsp[-5].minor.yy148->alias,
                                                                yymsp[-4].minor.yy504->notIndexedKw,
                                                                yymsp[-3].minor.yy456,
                                                                yymsp[-8].minor.yy137,
                                                                *(yymsp[-2].minor.yy373),
                                                                *(yymsp[-1].minor.yy459),
                                                                yymsp[0].minor.yy44
                                                            );
                                                    }
                                                    delete yymsp[-4].minor.yy504;
                                                }
                                                else
                                                {
                                                    yygotominor.yy147 = new SqliteDelete(
                                                            yymsp[-5].minor.yy148->name1,
                                                            yymsp[-5].minor.yy148->name2,
                                                            yymsp[-5].minor.yy148->alias,
                                                            false,
                                                            yymsp[-3].minor.yy456,
                                                            yymsp[-8].minor.yy137,
                                                            *(yymsp[-2].minor.yy373),
                                                            *(yymsp[-1].minor.yy459),
                                                            yymsp[0].minor.yy44
                                                        );
                                                }
                                                delete yymsp[-5].minor.yy148;
                                                delete yymsp[-2].minor.yy373;
                                                delete yymsp[-1].minor.yy459;
                                                // since it's used in trigger:
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 243: /* delete_stmt ::= with DELETE FROM */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                SqliteDelete* q = new SqliteDelete();
                                                q->with = yymsp[-2].minor.yy137;
                                                yygotominor.yy147 = q;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 244: /* delete_stmt ::= with DELETE FROM nm DOT */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                SqliteDelete* q = new SqliteDelete();
                                                q->with = yymsp[-4].minor.yy137;
                                                q->database = *(yymsp[-1].minor.yy239);
                                                yygotominor.yy147 = q;
                                                objectForTokens = yygotominor.yy147;
                                                delete yymsp[-1].minor.yy239;
                                            }
        break;
      case 245: /* delete_stmt ::= with DELETE FROM nm DOT ID_TAB */
      case 256: /* update_stmt ::= with UPDATE orconf nm DOT ID_TAB */ yytestcase(yyruleno==256);
{  yy_destructor(yypParser,251,&yymsp[-5].minor);
  yy_destructor(yypParser,205,&yymsp[-2].minor);
}
        break;
      case 246: /* delete_stmt ::= with DELETE FROM ID_DB|ID_TAB */
      case 257: /* update_stmt ::= with UPDATE orconf ID_DB|ID_TAB */ yytestcase(yyruleno==257);
{  yy_destructor(yypParser,251,&yymsp[-3].minor);
}
        break;
      case 249: /* where_opt ::= WHERE */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy456 = new SqliteExpr();
                                            }
        break;
      case 251: /* returning ::= RETURNING selcollist */
{yygotominor.yy373 = yymsp[0].minor.yy373;}
        break;
      case 253: /* update_stmt ::= with UPDATE orconf xfullname indexed_opt SET setlist from where_opt returning orderby_opt limit_opt */
{
                                                yygotominor.yy147 = new SqliteUpdate(
                                                        *(yymsp[-9].minor.yy194),
                                                        yymsp[-8].minor.yy148->name1,
                                                        yymsp[-8].minor.yy148->name2,
                                                        yymsp[-8].minor.yy148->alias,
                                                        yymsp[-7].minor.yy504 ? yymsp[-7].minor.yy504->notIndexedKw : false,
                                                        yymsp[-7].minor.yy504 ? yymsp[-7].minor.yy504->indexedBy : QString(),
                                                        *(yymsp[-5].minor.yy177),
                                                        yymsp[-4].minor.yy455,
                                                        yymsp[-3].minor.yy456,
                                                        yymsp[-11].minor.yy137,
                                                        *(yymsp[-2].minor.yy373),
                                                        *(yymsp[-1].minor.yy459),
                                                        yymsp[0].minor.yy44
                                                    );
                                                delete yymsp[-9].minor.yy194;
                                                delete yymsp[-8].minor.yy148;
                                                delete yymsp[-5].minor.yy177;
                                                delete yymsp[-2].minor.yy373;
                                                delete yymsp[-1].minor.yy459;
                                                if (yymsp[-7].minor.yy504)
                                                    delete yymsp[-7].minor.yy504;
                                                // since it's used in trigger:
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 254: /* update_stmt ::= with UPDATE orconf */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                SqliteUpdate* q = new SqliteUpdate();
                                                q->with = yymsp[-2].minor.yy137;
                                                yygotominor.yy147 = q;
                                                objectForTokens = yygotominor.yy147;
                                                delete yymsp[0].minor.yy194;
                                            }
        break;
      case 255: /* update_stmt ::= with UPDATE orconf nm DOT */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                SqliteUpdate* q = new SqliteUpdate();
                                                q->with = yymsp[-4].minor.yy137;
                                                q->database = *(yymsp[-1].minor.yy239);
                                                yygotominor.yy147 = q;
                                                objectForTokens = yygotominor.yy147;
                                                delete yymsp[-2].minor.yy194;
                                                delete yymsp[-1].minor.yy239;
                                            }
        break;
      case 258: /* setlist ::= setlist COMMA nm EQ expr */
{
                                                yymsp[-4].minor.yy177->append(ParserSetValue(*(yymsp[-2].minor.yy239), yymsp[0].minor.yy456));
                                                yygotominor.yy177 = yymsp[-4].minor.yy177;
                                                delete yymsp[-2].minor.yy239;
                                            }
        break;
      case 259: /* setlist ::= setlist COMMA LP idlist RP EQ expr */
{
                                                yymsp[-6].minor.yy177->append(ParserSetValue(*(yymsp[-3].minor.yy159), yymsp[0].minor.yy456));
                                                yygotominor.yy177 = yymsp[-6].minor.yy177;
                                                delete yymsp[-3].minor.yy159;
                                            }
        break;
      case 260: /* setlist ::= nm EQ expr */
{
                                                yygotominor.yy177 = new ParserSetValueList();
                                                yygotominor.yy177->append(ParserSetValue(*(yymsp[-2].minor.yy239), yymsp[0].minor.yy456));
                                                delete yymsp[-2].minor.yy239;
                                            }
        break;
      case 261: /* setlist ::= LP idlist RP EQ expr */
{
                                                yygotominor.yy177 = new ParserSetValueList();
                                                yygotominor.yy177->append(ParserSetValue(*(yymsp[-3].minor.yy159), yymsp[0].minor.yy456));
                                                delete yymsp[-3].minor.yy159;
                                            }
        break;
      case 262: /* setlist ::= */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy177 = new ParserSetValueList();
                                            }
        break;
      case 263: /* setlist ::= setlist COMMA */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy177 = yymsp[-1].minor.yy177;
                                            }
        break;
      case 264: /* setlist ::= setlist COMMA ID_COL */
      case 265: /* setlist ::= ID_COL */ yytestcase(yyruleno==265);
{  yy_destructor(yypParser,284,&yymsp[-2].minor);
}
        break;
      case 266: /* idlist_opt ::= */
{yygotominor.yy159 = new QStringList();}
        break;
      case 267: /* idlist_opt ::= LP idlist RP */
{yygotominor.yy159 = yymsp[-1].minor.yy159;}
        break;
      case 268: /* idlist ::= idlist COMMA nm */
{
                                                yygotominor.yy159 = yymsp[-2].minor.yy159;
                                                *(yygotominor.yy159) << *(yymsp[0].minor.yy239);
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 269: /* idlist ::= nm */
{
                                                yygotominor.yy159 = new QStringList();
                                                *(yygotominor.yy159) << *(yymsp[0].minor.yy239);
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 270: /* idlist ::= */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy159 = new QStringList();
                                            }
        break;
      case 271: /* idlist ::= idlist COMMA ID_COL */
      case 272: /* idlist ::= ID_COL */ yytestcase(yyruleno==272);
{  yy_destructor(yypParser,276,&yymsp[-2].minor);
}
        break;
      case 274: /* insert_stmt ::= with insert_cmd INTO xfullname idlist_opt select upsert returning */
{
                                                yygotominor.yy147 = new SqliteInsert(
                                                        yymsp[-6].minor.yy368->replace,
                                                        yymsp[-6].minor.yy368->orConflict,
                                                        yymsp[-4].minor.yy148->name1,
                                                        yymsp[-4].minor.yy148->name2,
                                                        yymsp[-4].minor.yy148->alias,
                                                        *(yymsp[-3].minor.yy159),
                                                        yymsp[-2].minor.yy113,
                                                        yymsp[-7].minor.yy137,
                                                        yymsp[-1].minor.yy16,
                                                        *(yymsp[0].minor.yy373)
                                                    );
                                                delete yymsp[-4].minor.yy148;
                                                delete yymsp[-6].minor.yy368;
                                                delete yymsp[-3].minor.yy159;
                                                delete yymsp[0].minor.yy373;
                                                // since it's used in trigger:
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 275: /* insert_stmt ::= with insert_cmd INTO xfullname idlist_opt DEFAULT VALUES returning */
{
                                                yygotominor.yy147 = new SqliteInsert(
                                                        yymsp[-6].minor.yy368->replace,
                                                        yymsp[-6].minor.yy368->orConflict,
                                                        yymsp[-4].minor.yy148->name1,
                                                        yymsp[-4].minor.yy148->name2,
                                                        yymsp[-4].minor.yy148->alias,
                                                        *(yymsp[-3].minor.yy159),
                                                        yymsp[-7].minor.yy137,
                                                        *(yymsp[0].minor.yy373)
                                                    );
                                                delete yymsp[-4].minor.yy148;
                                                delete yymsp[-6].minor.yy368;
                                                delete yymsp[-3].minor.yy159;
                                                delete yymsp[0].minor.yy373;
                                                // since it's used in trigger:
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 276: /* insert_stmt ::= with insert_cmd INTO xfullname LP idlist rp_opt */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                yygotominor.yy147 = new SqliteInsert(
                                                        yymsp[-5].minor.yy368->replace,
                                                        yymsp[-5].minor.yy368->orConflict,
                                                        yymsp[-3].minor.yy148->name1,
                                                        yymsp[-3].minor.yy148->name2,
                                                        yymsp[-3].minor.yy148->alias,
                                                        *(yymsp[-1].minor.yy159),
                                                        yymsp[-6].minor.yy137,
                                                        QList<SqliteResultColumn*>()
                                                    );
                                                objectForTokens = yygotominor.yy147;
                                                delete yymsp[-3].minor.yy148;
                                                delete yymsp[-1].minor.yy159;
                                                delete yymsp[-5].minor.yy368;
                                                delete yymsp[0].minor.yy515;
                                            }
        break;
      case 277: /* insert_stmt ::= with insert_cmd INTO */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                SqliteInsert* q = new SqliteInsert();
                                                q->replaceKw = yymsp[-1].minor.yy368->replace;
                                                q->onConflict = yymsp[-1].minor.yy368->orConflict;
                                                q->with = yymsp[-2].minor.yy137;
                                                yygotominor.yy147 = q;
                                                objectForTokens = yygotominor.yy147;
                                                delete yymsp[-1].minor.yy368;
                                            }
        break;
      case 278: /* insert_stmt ::= with insert_cmd INTO nm DOT */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                SqliteInsert* q = new SqliteInsert();
                                                q->replaceKw = yymsp[-3].minor.yy368->replace;
                                                q->onConflict = yymsp[-3].minor.yy368->orConflict;
                                                q->with = yymsp[-4].minor.yy137;
                                                q->database = *(yymsp[-1].minor.yy239);
                                                yygotominor.yy147 = q;
                                                objectForTokens = yygotominor.yy147;
                                                delete yymsp[-3].minor.yy368;
                                                delete yymsp[-1].minor.yy239;
                                            }
        break;
      case 279: /* insert_stmt ::= with insert_cmd INTO ID_DB|ID_TAB */
{  yy_destructor(yypParser,251,&yymsp[-3].minor);
  yy_destructor(yypParser,287,&yymsp[-2].minor);
}
        break;
      case 280: /* insert_stmt ::= with insert_cmd INTO nm DOT ID_TAB */
{  yy_destructor(yypParser,251,&yymsp[-5].minor);
  yy_destructor(yypParser,287,&yymsp[-4].minor);
  yy_destructor(yypParser,205,&yymsp[-2].minor);
}
        break;
      case 281: /* insert_cmd ::= INSERT orconf */
{
                                                yygotominor.yy368 = new ParserStubInsertOrReplace(false, *(yymsp[0].minor.yy194));
                                                delete yymsp[0].minor.yy194;
                                            }
        break;
      case 282: /* insert_cmd ::= REPLACE */
{yygotominor.yy368 = new ParserStubInsertOrReplace(true);}
        break;
      case 283: /* upsert ::= */
{
                                                yygotominor.yy16 = nullptr;
                                            }
        break;
      case 284: /* upsert ::= ON CONFLICT LP sortlist RP where_opt DO UPDATE SET setlist where_opt */
{
                                                yygotominor.yy16 = new SqliteUpsert(*(yymsp[-7].minor.yy459), yymsp[-5].minor.yy456, *(yymsp[-1].minor.yy177), yymsp[0].minor.yy456);
                                                delete yymsp[-7].minor.yy459;
                                                delete yymsp[-1].minor.yy177;
                                                objectForTokens = yygotominor.yy16;
                                            }
        break;
      case 285: /* upsert ::= ON CONFLICT LP sortlist RP where_opt DO NOTHING */
{
                                                yygotominor.yy16 = new SqliteUpsert(*(yymsp[-4].minor.yy459), yymsp[-2].minor.yy456);
                                                delete yymsp[-4].minor.yy459;
                                                objectForTokens = yygotominor.yy16;
                                            }
        break;
      case 286: /* upsert ::= ON CONFLICT DO NOTHING */
{
                                                yygotominor.yy16 = new SqliteUpsert();
                                                objectForTokens = yygotominor.yy16;
                                            }
        break;
      case 287: /* exprx ::= expr not_opt IN ID_DB */
{  yy_destructor(yypParser,229,&yymsp[-3].minor);
}
        break;
      case 288: /* exprx ::= expr not_opt IN nm DOT ID_TAB */
      case 289: /* exprx ::= ID_DB|ID_TAB|ID_COL|ID_FN */ yytestcase(yyruleno==289);
{  yy_destructor(yypParser,229,&yymsp[-5].minor);
  yy_destructor(yypParser,205,&yymsp[-2].minor);
}
        break;
      case 290: /* exprx ::= tnm DOT ID_TAB|ID_COL */
{  yy_destructor(yypParser,237,&yymsp[-2].minor);
}
        break;
      case 291: /* exprx ::= tnm DOT nm DOT ID_COL */
{  yy_destructor(yypParser,237,&yymsp[-4].minor);
  yy_destructor(yypParser,205,&yymsp[-2].minor);
}
        break;
      case 292: /* exprx ::= expr COLLATE ID_COLLATE */
      case 293: /* exprx ::= RAISE LP raisetype COMMA ID_ERR_MSG RP */ yytestcase(yyruleno==293);
{  yy_destructor(yypParser,229,&yymsp[-2].minor);
}
        break;
      case 294: /* exprx ::= CTIME_KW */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initCTime(yymsp[0].minor.yy0->value);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 295: /* exprx ::= LP nexprlist RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initRowValue(*(yymsp[-1].minor.yy143));
                                                delete yymsp[-1].minor.yy143;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 296: /* exprx ::= tnm */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                if (yymsp[0].minor.yy14->isLiteral())
                                                    yygotominor.yy456->initLiteral(yymsp[0].minor.yy14->toLiteral());
                                                else
                                                    yygotominor.yy456->initId(yymsp[0].minor.yy14->toName());
                                                    //parserContext->errorBeforeLastToken("Syntax error <expected literal value>");

                                                delete yymsp[0].minor.yy14;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 297: /* exprx ::= tnm DOT nm */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                if (yymsp[-2].minor.yy14->isName())
                                                    yygotominor.yy456->initId(yymsp[-2].minor.yy14->toName(), *(yymsp[0].minor.yy239));
                                                else
                                                    parserContext->errorAtToken("Syntax error <expected name>", -3);

                                                delete yymsp[-2].minor.yy14;
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 298: /* exprx ::= tnm DOT */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                objectForTokens = yygotominor.yy456;
                                                if (yymsp[-1].minor.yy14->isName())
                                                {
                                                    yygotominor.yy456->initId(yymsp[-1].minor.yy14->toName(), QString());
                                                    parserContext->minorErrorAfterLastToken("Syntax error <expr>");
                                                }
                                                else
                                                    parserContext->errorAtToken("Syntax error <expected name>", -3);

                                                delete yymsp[-1].minor.yy14;
                                            }
        break;
      case 299: /* exprx ::= tnm DOT nm DOT nm */
{
                                                yygotominor.yy456 = new SqliteExpr();
												if (yymsp[-4].minor.yy14->isName())
													yygotominor.yy456->initId(yymsp[-4].minor.yy14->toName(), *(yymsp[-2].minor.yy239), *(yymsp[0].minor.yy239));
												else
													parserContext->errorAtToken("Syntax error <expected name>", -5);

                                                delete yymsp[-4].minor.yy14;
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 300: /* exprx ::= tnm DOT nm DOT */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                objectForTokens = yygotominor.yy456;
                                                if (yymsp[-3].minor.yy14->isName())
                                                {
                                                    yygotominor.yy456->initId(yymsp[-3].minor.yy14->toName(), *(yymsp[-1].minor.yy239), QString());
                                                    parserContext->minorErrorAfterLastToken("Syntax error <expr>");
                                                }
                                                else
                                                    parserContext->errorAtToken("Syntax error <expected name>", -5);

                                                delete yymsp[-3].minor.yy14;
                                                delete yymsp[-1].minor.yy239;
                                            }
        break;
      case 301: /* exprx ::= VARIABLE */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initBindParam(yymsp[0].minor.yy0->value);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 302: /* exprx ::= expr COLLATE ids */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initCollate(yymsp[-2].minor.yy456, *(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 303: /* exprx ::= CAST LP expr AS typetoken RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initCast(yymsp[-3].minor.yy456, yymsp[-1].minor.yy1);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 304: /* exprx ::= id LP distinct exprlist RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initFunction(stripObjName(*(yymsp[-4].minor.yy239)), *(yymsp[-2].minor.yy642), *(yymsp[-1].minor.yy143));
                                                delete yymsp[-4].minor.yy239;
                                                delete yymsp[-2].minor.yy642;
                                                delete yymsp[-1].minor.yy143;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 305: /* exprx ::= id LP distinct exprlist ORDER BY sortlist RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initFunction(stripObjName(*(yymsp[-7].minor.yy239)), *(yymsp[-5].minor.yy642), *(yymsp[-4].minor.yy143), *(yymsp[-1].minor.yy459));
                                                delete yymsp[-7].minor.yy239;
                                                delete yymsp[-5].minor.yy642;
                                                delete yymsp[-4].minor.yy143;
                                                delete yymsp[-1].minor.yy459;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 306: /* exprx ::= id LP STAR RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initFunction(stripObjName(*(yymsp[-3].minor.yy239)), true);
                                                delete yymsp[-3].minor.yy239;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 307: /* exprx ::= expr AND expr */
      case 308: /* exprx ::= expr OR expr */ yytestcase(yyruleno==308);
      case 309: /* exprx ::= expr LT|GT|GE|LE expr */ yytestcase(yyruleno==309);
      case 310: /* exprx ::= expr EQ|NE expr */ yytestcase(yyruleno==310);
      case 311: /* exprx ::= expr BITAND|BITOR|LSHIFT|RSHIFT expr */ yytestcase(yyruleno==311);
      case 312: /* exprx ::= expr PLUS|MINUS expr */ yytestcase(yyruleno==312);
      case 313: /* exprx ::= expr STAR|SLASH|REM expr */ yytestcase(yyruleno==313);
      case 314: /* exprx ::= expr CONCAT expr */ yytestcase(yyruleno==314);
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initBinOp(yymsp[-2].minor.yy456, yymsp[-1].minor.yy0->value, yymsp[0].minor.yy456);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 315: /* exprx ::= expr not_opt likeop expr */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initLike(yymsp[-3].minor.yy456, *(yymsp[-2].minor.yy515), *(yymsp[-1].minor.yy232), yymsp[0].minor.yy456);
                                                delete yymsp[-2].minor.yy515;
                                                delete yymsp[-1].minor.yy232;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 316: /* exprx ::= expr not_opt likeop expr ESCAPE expr */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initLike(yymsp[-5].minor.yy456, *(yymsp[-4].minor.yy515), *(yymsp[-3].minor.yy232), yymsp[-2].minor.yy456, yymsp[0].minor.yy456);
                                                delete yymsp[-4].minor.yy515;
                                                delete yymsp[-3].minor.yy232;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 317: /* exprx ::= expr ISNULL|NOTNULL */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initNull(yymsp[-1].minor.yy456, yymsp[0].minor.yy0->value);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 318: /* exprx ::= expr NOT NULL */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initNull(yymsp[-2].minor.yy456, "NOT NULL");
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 319: /* exprx ::= expr IS not_opt expr */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initIs(yymsp[-3].minor.yy456, *(yymsp[-1].minor.yy515), yymsp[0].minor.yy456);
                                                delete yymsp[-1].minor.yy515;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 320: /* exprx ::= expr IS NOT DISTINCT FROM expr */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initDistinct(yymsp[-5].minor.yy456, true, yymsp[0].minor.yy456);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 321: /* exprx ::= expr IS DISTINCT FROM expr */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initDistinct(yymsp[-4].minor.yy456, false, yymsp[0].minor.yy456);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 322: /* exprx ::= NOT expr */
      case 323: /* exprx ::= BITNOT expr */ yytestcase(yyruleno==323);
      case 325: /* exprx ::= PLUS expr */ yytestcase(yyruleno==325);
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initUnaryOp(yymsp[0].minor.yy456, yymsp[-1].minor.yy0->value);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 324: /* exprx ::= MINUS expr */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                if (yymsp[0].minor.yy456->mode == SqliteExpr::Mode::LITERAL_VALUE &&
                                                    parserContext->isCandidateForMaxNegativeNumber() &&
                                                    yymsp[0].minor.yy456->literalValue == static_cast<qint64>(0L))
                                                {
                                                    yygotominor.yy456->initLiteral(std::numeric_limits<qint64>::min());
                                                    delete yymsp[0].minor.yy456;
                                                }
                                                else
                                                {
                                                    yygotominor.yy456->initUnaryOp(yymsp[0].minor.yy456, yymsp[-1].minor.yy0->value);
                                                }
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 326: /* exprx ::= expr PTR expr */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initPtrOp(yymsp[-2].minor.yy456, yymsp[-1].minor.yy0->value, yymsp[0].minor.yy456);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 327: /* exprx ::= expr not_opt BETWEEN expr AND expr */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initBetween(yymsp[-5].minor.yy456, *(yymsp[-4].minor.yy515), yymsp[-2].minor.yy456, yymsp[0].minor.yy456);
                                                delete yymsp[-4].minor.yy515;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 328: /* exprx ::= expr not_opt IN LP exprlist RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initIn(yymsp[-5].minor.yy456, *(yymsp[-4].minor.yy515), *(yymsp[-1].minor.yy143));
                                                delete yymsp[-4].minor.yy515;
                                                delete yymsp[-1].minor.yy143;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 329: /* exprx ::= LP select RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initSubSelect(yymsp[-1].minor.yy113);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 330: /* exprx ::= expr not_opt IN LP select RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initIn(yymsp[-5].minor.yy456, *(yymsp[-4].minor.yy515), yymsp[-1].minor.yy113);
                                                delete yymsp[-4].minor.yy515;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 331: /* exprx ::= expr not_opt IN nm dbnm */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initIn(yymsp[-4].minor.yy456, *(yymsp[-3].minor.yy515), *(yymsp[-1].minor.yy239), *(yymsp[0].minor.yy239));
                                                delete yymsp[-3].minor.yy515;
                                                delete yymsp[-1].minor.yy239;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 332: /* exprx ::= EXISTS LP select RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initExists(yymsp[-1].minor.yy113);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 333: /* exprx ::= CASE case_operand case_exprlist case_else END */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initCase(yymsp[-3].minor.yy456, *(yymsp[-2].minor.yy143), yymsp[-1].minor.yy456);
                                                delete yymsp[-2].minor.yy143;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 334: /* exprx ::= RAISE LP IGNORE RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initRaise(yymsp[-1].minor.yy0->value);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 335: /* exprx ::= RAISE LP raisetype COMMA expr RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initRaise(yymsp[-3].minor.yy0->value, yymsp[-1].minor.yy456);
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 336: /* exprx ::= id LP distinct exprlist RP filter_over */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initWindowFunction(stripObjName(*(yymsp[-5].minor.yy239)), *(yymsp[-3].minor.yy642), *(yymsp[-2].minor.yy143), yymsp[0].minor.yy567);
                                                delete yymsp[-5].minor.yy239;
                                                delete yymsp[-3].minor.yy642;
                                                delete yymsp[-2].minor.yy143;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 337: /* exprx ::= id LP distinct exprlist ORDER BY sortlist RP filter_over */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initWindowFunction(stripObjName(*(yymsp[-8].minor.yy239)), *(yymsp[-6].minor.yy642), *(yymsp[-5].minor.yy143), *(yymsp[-2].minor.yy459), yymsp[0].minor.yy567);
                                                delete yymsp[-8].minor.yy239;
                                                delete yymsp[-6].minor.yy642;
                                                delete yymsp[-5].minor.yy143;
                                                delete yymsp[-2].minor.yy459;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 338: /* exprx ::= id LP STAR RP filter_over */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initWindowFunction(stripObjName(*(yymsp[-4].minor.yy239)), yymsp[0].minor.yy567);
                                                delete yymsp[-4].minor.yy239;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 339: /* exprx ::= id LP distinct exprlist RP WITHIN GROUP LP ORDER BY expr RP */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                yygotominor.yy456->initOrderedSetAggregate(stripObjName(*(yymsp[-11].minor.yy239)), *(yymsp[-9].minor.yy642), *(yymsp[-8].minor.yy143), yymsp[-1].minor.yy456);
                                                delete yymsp[-11].minor.yy239;
                                                delete yymsp[-9].minor.yy642;
                                                delete yymsp[-8].minor.yy143;
                                                objectForTokens = yygotominor.yy456;
                                            }
        break;
      case 340: /* expr ::= */
{
                                                yygotominor.yy456 = new SqliteExpr();
                                                objectForTokens = yygotominor.yy456;
                                                parserContext->minorErrorAfterLastToken("Syntax error <expr>");
                                            }
        break;
      case 346: /* likeop ::= LIKE_KW|MATCH */
{yygotominor.yy232 = new SqliteExpr::LikeOp(SqliteExpr::likeOp(yymsp[0].minor.yy0->value));}
        break;
      case 347: /* case_exprlist ::= case_exprlist WHEN expr THEN expr */
{
                                                yymsp[-4].minor.yy143->append(yymsp[-2].minor.yy456);
                                                yymsp[-4].minor.yy143->append(yymsp[0].minor.yy456);
                                                yygotominor.yy143 = yymsp[-4].minor.yy143;
                                            }
        break;
      case 348: /* case_exprlist ::= WHEN expr THEN expr */
{
                                                yygotominor.yy143 = new ParserExprList();
                                                yygotominor.yy143->append(yymsp[-2].minor.yy456);
                                                yygotominor.yy143->append(yymsp[0].minor.yy456);
                                            }
        break;
      case 355: /* nexprlist ::= nexprlist COMMA expr */
{
                                                yymsp[-2].minor.yy143->append(yymsp[0].minor.yy456);
                                                yygotominor.yy143 = yymsp[-2].minor.yy143;
                                                DONT_INHERIT_TOKENS("nexprlist");
                                            }
        break;
      case 356: /* nexprlist ::= exprx */
{
                                                yygotominor.yy143 = new ParserExprList();
                                                yygotominor.yy143->append(yymsp[0].minor.yy456);
                                            }
        break;
      case 357: /* cmd ::= CREATE uniqueflag INDEX ifnotexists nm dbnm ON nm LP sortlist RP where_opt */
{
                                                yygotominor.yy147 = new SqliteCreateIndex(
                                                        *(yymsp[-10].minor.yy515),
                                                        *(yymsp[-8].minor.yy515),
                                                        *(yymsp[-7].minor.yy239),
                                                        *(yymsp[-6].minor.yy239),
                                                        *(yymsp[-4].minor.yy239),
                                                        *(yymsp[-2].minor.yy459),
                                                        yymsp[0].minor.yy456
                                                    );
                                                delete yymsp[-8].minor.yy515;
                                                delete yymsp[-10].minor.yy515;
                                                delete yymsp[-7].minor.yy239;
                                                delete yymsp[-6].minor.yy239;
                                                delete yymsp[-4].minor.yy239;
                                                delete yymsp[-2].minor.yy459;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 358: /* cmd ::= CREATE uniqueflag INDEX ifnotexists nm dbnm ON ID_TAB */
{  yy_destructor(yypParser,205,&yymsp[-3].minor);
}
        break;
      case 363: /* idxlist_opt ::= */
{yygotominor.yy240 = new ParserIndexedColumnList();}
        break;
      case 364: /* idxlist_opt ::= LP idxlist RP */
{yygotominor.yy240 = yymsp[-1].minor.yy240;}
        break;
      case 365: /* idxlist ::= idxlist COMMA idxlist_single */
{
                                                yymsp[-2].minor.yy240->append(yymsp[0].minor.yy460);
                                                yygotominor.yy240 = yymsp[-2].minor.yy240;
                                                DONT_INHERIT_TOKENS("idxlist");
                                            }
        break;
      case 366: /* idxlist ::= idxlist_single */
{
                                                yygotominor.yy240 = new ParserIndexedColumnList();
                                                yygotominor.yy240->append(yymsp[0].minor.yy460);
                                            }
        break;
      case 367: /* idxlist_single ::= nm collate sortorder */
      case 368: /* idxlist_single ::= ID_COL */ yytestcase(yyruleno==368);
{
                                                SqliteIndexedColumn* obj =
                                                    new SqliteIndexedColumn(
                                                        *(yymsp[-2].minor.yy239),
                                                        *(yymsp[-1].minor.yy239),
                                                        *(yymsp[0].minor.yy509)
                                                    );
                                                yygotominor.yy460 = obj;
                                                delete yymsp[0].minor.yy509;
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[-1].minor.yy239;
                                                objectForTokens = yygotominor.yy460;
                                            }
        break;
      case 372: /* cmd ::= DROP INDEX ifexists fullname */
{
                                                yygotominor.yy147 = new SqliteDropIndex(*(yymsp[-1].minor.yy515), yymsp[0].minor.yy648->name1, yymsp[0].minor.yy648->name2);
                                                delete yymsp[-1].minor.yy515;
                                                delete yymsp[0].minor.yy648;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 375: /* cmd ::= VACUUM vinto */
{
                                                yygotominor.yy147 = new SqliteVacuum(yymsp[0].minor.yy456);
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 376: /* cmd ::= VACUUM nm vinto */
{
                                                yygotominor.yy147 = new SqliteVacuum(*(yymsp[-1].minor.yy239), yymsp[0].minor.yy456);
                                                delete yymsp[-1].minor.yy239;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 379: /* cmd ::= PRAGMA nm dbnm */
{
                                                yygotominor.yy147 = new SqlitePragma(*(yymsp[-1].minor.yy239), *(yymsp[0].minor.yy239));
                                                delete yymsp[-1].minor.yy239;
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 380: /* cmd ::= PRAGMA nm dbnm EQ nmnum */
      case 382: /* cmd ::= PRAGMA nm dbnm EQ minus_num */ yytestcase(yyruleno==382);
{
                                                yygotominor.yy147 = new SqlitePragma(*(yymsp[-3].minor.yy239), *(yymsp[-2].minor.yy239), *(yymsp[0].minor.yy13), true);
                                                delete yymsp[-3].minor.yy239;
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[0].minor.yy13;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 381: /* cmd ::= PRAGMA nm dbnm LP nmnum RP */
      case 383: /* cmd ::= PRAGMA nm dbnm LP minus_num RP */ yytestcase(yyruleno==383);
{
                                                yygotominor.yy147 = new SqlitePragma(*(yymsp[-4].minor.yy239), *(yymsp[-3].minor.yy239), *(yymsp[-1].minor.yy13), false);
                                                delete yymsp[-4].minor.yy239;
                                                delete yymsp[-3].minor.yy239;
                                                delete yymsp[-1].minor.yy13;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 387: /* nmnum ::= nm */
{
                                                yygotominor.yy13 = new QVariant(*(yymsp[0].minor.yy239));
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 388: /* nmnum ::= ON */
      case 389: /* nmnum ::= DELETE */ yytestcase(yyruleno==389);
      case 390: /* nmnum ::= DEFAULT */ yytestcase(yyruleno==390);
{yygotominor.yy13 = new QVariant(yymsp[0].minor.yy0->value);}
        break;
      case 393: /* minus_num ::= MINUS number */
{
                                                if (yymsp[0].minor.yy13->userType() == QMetaType::Double)
                                                    *(yymsp[0].minor.yy13) = -(yymsp[0].minor.yy13->toDouble());
                                                else if (yymsp[0].minor.yy13->userType() == QMetaType::LongLong)
                                                {
                                                    if (parserContext->isCandidateForMaxNegativeNumber())
                                                        *(yymsp[0].minor.yy13) = std::numeric_limits<qint64>::min();
                                                    else
                                                        *(yymsp[0].minor.yy13) = -(yymsp[0].minor.yy13->toLongLong());
                                                }
                                                else
                                                    Q_ASSERT_X(true, "producing minus number", "QVariant is neither of Double or LongLong.");

                                                yygotominor.yy13 = yymsp[0].minor.yy13;
                                            }
        break;
      case 396: /* cmd ::= CREATE temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON nm foreach_clause when_clause BEGIN trigger_cmd_list END */
{
                                                yygotominor.yy147 = new SqliteCreateTrigger(
                                                        *(yymsp[-13].minor.yy642),
                                                        *(yymsp[-11].minor.yy515),
                                                        *(yymsp[-10].minor.yy239),
                                                        *(yymsp[-9].minor.yy239),
                                                        *(yymsp[-5].minor.yy239),
                                                        *(yymsp[-8].minor.yy500),
                                                        yymsp[-7].minor.yy87,
                                                        *(yymsp[-4].minor.yy115),
                                                        yymsp[-3].minor.yy456,
                                                        *(yymsp[-1].minor.yy198),
                                                        3
                                                    );
                                                delete yymsp[-11].minor.yy515;
                                                delete yymsp[-13].minor.yy642;
                                                delete yymsp[-8].minor.yy500;
                                                delete yymsp[-4].minor.yy115;
                                                delete yymsp[-10].minor.yy239;
                                                delete yymsp[-5].minor.yy239;
                                                delete yymsp[-9].minor.yy239;
                                                delete yymsp[-1].minor.yy198;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 397: /* cmd ::= CREATE temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON nm foreach_clause when_clause */
{
                                                QList<SqliteQuery *> CL;

                                                yygotominor.yy147 = new SqliteCreateTrigger(
                                                        *(yymsp[-10].minor.yy642),
                                                        *(yymsp[-8].minor.yy515),
                                                        *(yymsp[-7].minor.yy239),
                                                        *(yymsp[-6].minor.yy239),
                                                        *(yymsp[-2].minor.yy239),
                                                        *(yymsp[-5].minor.yy500),
                                                        yymsp[-4].minor.yy87,
                                                        *(yymsp[-1].minor.yy115),
                                                        yymsp[0].minor.yy456,
                                                        CL,
                                                        3
                                                    );
                                                delete yymsp[-8].minor.yy515;
                                                delete yymsp[-10].minor.yy642;
                                                delete yymsp[-5].minor.yy500;
                                                delete yymsp[-1].minor.yy115;
                                                delete yymsp[-7].minor.yy239;
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[-6].minor.yy239;
                                                objectForTokens = yygotominor.yy147;
                                                parserContext->minorErrorAfterLastToken("Syntax error");
                                            }
        break;
      case 398: /* cmd ::= CREATE temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON nm foreach_clause when_clause BEGIN trigger_cmd_list */
{
                                                yygotominor.yy147 = new SqliteCreateTrigger(
													*(yymsp[-12].minor.yy642),
													*(yymsp[-10].minor.yy515),
													*(yymsp[-9].minor.yy239),
													*(yymsp[-8].minor.yy239),
													*(yymsp[-4].minor.yy239),
													*(yymsp[-7].minor.yy500),
													yymsp[-6].minor.yy87,
													*(yymsp[-3].minor.yy115),
													yymsp[-2].minor.yy456,
													*(yymsp[0].minor.yy198),
													3
                                                );
                                                delete yymsp[-10].minor.yy515;
                                                delete yymsp[-12].minor.yy642;
                                                delete yymsp[-7].minor.yy500;
                                                delete yymsp[-3].minor.yy115;
                                                delete yymsp[-9].minor.yy239;
                                                delete yymsp[-4].minor.yy239;
                                                delete yymsp[-8].minor.yy239;
                                                delete yymsp[0].minor.yy198;
                                                objectForTokens = yygotominor.yy147;
                                                parserContext->minorErrorAfterLastToken("Syntax error");
                                            }
        break;
      case 399: /* cmd ::= CREATE temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON ID_TAB */
{  yy_destructor(yypParser,207,&yymsp[-8].minor);
  yy_destructor(yypParser,205,&yymsp[-5].minor);
  yy_destructor(yypParser,303,&yymsp[-3].minor);
  yy_destructor(yypParser,304,&yymsp[-2].minor);
}
        break;
      case 402: /* trigger_time ::= BEFORE */
{yygotominor.yy500 = new SqliteCreateTrigger::Time(SqliteCreateTrigger::Time::BEFORE);}
        break;
      case 403: /* trigger_time ::= AFTER */
{yygotominor.yy500 = new SqliteCreateTrigger::Time(SqliteCreateTrigger::Time::AFTER);}
        break;
      case 404: /* trigger_time ::= INSTEAD OF */
{yygotominor.yy500 = new SqliteCreateTrigger::Time(SqliteCreateTrigger::Time::INSTEAD_OF);}
        break;
      case 405: /* trigger_time ::= */
{yygotominor.yy500 = new SqliteCreateTrigger::Time(SqliteCreateTrigger::Time::null);}
        break;
      case 406: /* trigger_event ::= DELETE */
{
                                                yygotominor.yy87 = new SqliteCreateTrigger::Event(SqliteCreateTrigger::Event::DELETE);
                                                objectForTokens = yygotominor.yy87;
                                            }
        break;
      case 407: /* trigger_event ::= INSERT */
{
                                                yygotominor.yy87 = new SqliteCreateTrigger::Event(SqliteCreateTrigger::Event::INSERT);
                                                objectForTokens = yygotominor.yy87;
                                            }
        break;
      case 408: /* trigger_event ::= UPDATE */
{
                                                yygotominor.yy87 = new SqliteCreateTrigger::Event(SqliteCreateTrigger::Event::UPDATE);
                                                objectForTokens = yygotominor.yy87;
                                            }
        break;
      case 409: /* trigger_event ::= UPDATE OF idlist */
{
                                                yygotominor.yy87 = new SqliteCreateTrigger::Event(*(yymsp[0].minor.yy159));
                                                delete yymsp[0].minor.yy159;
                                                objectForTokens = yygotominor.yy87;
                                            }
        break;
      case 410: /* foreach_clause ::= */
{yygotominor.yy115 = new SqliteCreateTrigger::Scope(SqliteCreateTrigger::Scope::null);}
        break;
      case 411: /* foreach_clause ::= FOR EACH ROW */
{yygotominor.yy115 = new SqliteCreateTrigger::Scope(SqliteCreateTrigger::Scope::FOR_EACH_ROW);}
        break;
      case 414: /* trigger_cmd_list ::= trigger_cmd_list trigger_cmd SEMI */
{
                                                yymsp[-2].minor.yy198->append(yymsp[-1].minor.yy147);
                                                yygotominor.yy198 = yymsp[-2].minor.yy198;
                                                DONT_INHERIT_TOKENS("trigger_cmd_list");
                                            }
        break;
      case 415: /* trigger_cmd_list ::= trigger_cmd SEMI */
{
                                                yygotominor.yy198 = new ParserQueryList();
                                                yygotominor.yy198->append(yymsp[-1].minor.yy147);
                                            }
        break;
      case 416: /* trigger_cmd_list ::= SEMI */
{
                                                yygotominor.yy198 = new ParserQueryList();
                                                parserContext->minorErrorAfterLastToken("Syntax error");
                                            }
        break;
      case 421: /* raisetype ::= ROLLBACK|ABORT|FAIL */
{yygotominor.yy0 = yymsp[0].minor.yy0;}
        break;
      case 422: /* cmd ::= DROP TRIGGER ifexists fullname */
{
                                                yygotominor.yy147 = new SqliteDropTrigger(*(yymsp[-1].minor.yy515), yymsp[0].minor.yy648->name1, yymsp[0].minor.yy648->name2);
                                                delete yymsp[-1].minor.yy515;
                                                delete yymsp[0].minor.yy648;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 425: /* cmd ::= ATTACH database_kw_opt expr AS expr key_opt */
{
                                                yygotominor.yy147 = new SqliteAttach(*(yymsp[-4].minor.yy515), yymsp[-3].minor.yy456, yymsp[-1].minor.yy456, yymsp[0].minor.yy456);
                                                delete yymsp[-4].minor.yy515;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 426: /* cmd ::= DETACH database_kw_opt expr */
{
                                                yygotominor.yy147 = new SqliteDetach(*(yymsp[-1].minor.yy515), yymsp[0].minor.yy456);
                                                delete yymsp[-1].minor.yy515;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 431: /* cmd ::= REINDEX */
{yygotominor.yy147 = new SqliteReindex();}
        break;
      case 432: /* cmd ::= REINDEX nm dbnm */
      case 433: /* cmd ::= REINDEX ID_COLLATE */ yytestcase(yyruleno==433);
{
                                                yygotominor.yy147 = new SqliteReindex(*(yymsp[-1].minor.yy239), *(yymsp[0].minor.yy239));
                                                delete yymsp[-1].minor.yy239;
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 436: /* cmd ::= ANALYZE */
{
                                                yygotominor.yy147 = new SqliteAnalyze();
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 437: /* cmd ::= ANALYZE nm dbnm */
{
                                                yygotominor.yy147 = new SqliteAnalyze(*(yymsp[-1].minor.yy239), *(yymsp[0].minor.yy239));
                                                delete yymsp[-1].minor.yy239;
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 440: /* cmd ::= ALTER TABLE fullname RENAME TO nm */
{
                                                auto alter = new SqliteAlterTable();
                                                alter->initRenameTable(
                                                        yymsp[-3].minor.yy648->name1,
                                                        yymsp[-3].minor.yy648->name2,
                                                        *(yymsp[0].minor.yy239)
                                                    );
                                                delete yymsp[0].minor.yy239;
                                                delete yymsp[-3].minor.yy648;
                                                yygotominor.yy147 = alter;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 441: /* cmd ::= ALTER TABLE fullname ADD kwcolumn_opt column */
{
                                                auto alter = new SqliteAlterTable();
                                                alter->initAddColumn(
                                                        yymsp[-3].minor.yy648->name1,
                                                        yymsp[-3].minor.yy648->name2,
                                                        *(yymsp[-1].minor.yy515),
                                                        yymsp[0].minor.yy123
                                                    );
                                                delete yymsp[-1].minor.yy515;
                                                delete yymsp[-3].minor.yy648;
                                                yygotominor.yy147 = alter;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 442: /* cmd ::= ALTER TABLE fullname DROP kwcolumn_opt nm */
{
                                                auto alter = new SqliteAlterTable();
                                                alter->initDropColumn(
                                                        yymsp[-3].minor.yy648->name1,
                                                        yymsp[-3].minor.yy648->name2,
                                                        *(yymsp[-1].minor.yy515),
                                                        *(yymsp[0].minor.yy239)
                                                    );
                                                delete yymsp[-1].minor.yy515;
                                                delete yymsp[-3].minor.yy648;
                                                delete yymsp[0].minor.yy239;
                                                yygotominor.yy147 = alter;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 443: /* cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt nm TO nm */
{
                                               auto alter = new SqliteAlterTable();
                                               alter->initRenameColumn(
                                                       yymsp[-5].minor.yy648->name1,
                                                       yymsp[-5].minor.yy648->name2,
                                                       *(yymsp[-3].minor.yy515),
                                                       *(yymsp[-2].minor.yy239),
                                                       *(yymsp[0].minor.yy239)
                                                   );
                                               delete yymsp[-3].minor.yy515;
                                               delete yymsp[-5].minor.yy648;
                                               delete yymsp[-2].minor.yy239;
                                               delete yymsp[0].minor.yy239;
                                               yygotominor.yy147 = alter;
                                               objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 444: /* cmd ::= ALTER TABLE fullname ALTER kwcolumn_opt nm SET NOT NULL onconf */
{
                                                auto alter = new SqliteAlterTable();
                                                alter->initColumnSetNotNull(
                                                        yymsp[-7].minor.yy648->name1,
                                                        yymsp[-7].minor.yy648->name2,
                                                        *(yymsp[-5].minor.yy515),
                                                        *(yymsp[-4].minor.yy239),
                                                        *(yymsp[0].minor.yy194)
                                                    );
                                                delete yymsp[-5].minor.yy515;
                                                delete yymsp[-4].minor.yy239;
                                                delete yymsp[-7].minor.yy648;
                                                delete yymsp[0].minor.yy194;
                                                yygotominor.yy147 = alter;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 445: /* cmd ::= ALTER TABLE fullname ALTER kwcolumn_opt nm DROP NOT NULL */
{
                                                auto alter = new SqliteAlterTable();
                                                alter->initColumnDropNotNull(
                                                        yymsp[-6].minor.yy648->name1,
                                                        yymsp[-6].minor.yy648->name2,
                                                        *(yymsp[-4].minor.yy515),
                                                        *(yymsp[-3].minor.yy239)
                                                    );
                                                delete yymsp[-4].minor.yy515;
                                                delete yymsp[-3].minor.yy239;
                                                delete yymsp[-6].minor.yy648;
                                                yygotominor.yy147 = alter;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 446: /* cmd ::= ALTER TABLE fullname ADD CONSTRAINT nm CHECK LP expr RP onconf */
{
                                                auto alter = new SqliteAlterTable();
                                                alter->initAddCheck(
                                                        yymsp[-8].minor.yy648->name1,
                                                        yymsp[-8].minor.yy648->name2,
                                                        *(yymsp[-5].minor.yy239),
                                                        yymsp[-2].minor.yy456,
                                                        *(yymsp[0].minor.yy194)
                                                    );
                                                delete yymsp[-8].minor.yy648;
                                                delete yymsp[-5].minor.yy239;
                                                delete yymsp[0].minor.yy194;
                                                yygotominor.yy147 = alter;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 447: /* cmd ::= ALTER TABLE fullname ADD CHECK LP expr RP onconf */
{
                                                auto alter = new SqliteAlterTable();
                                                alter->initAddCheck(
                                                        yymsp[-6].minor.yy648->name1,
                                                        yymsp[-6].minor.yy648->name2,
                                                        yymsp[-2].minor.yy456,
                                                        *(yymsp[0].minor.yy194)
                                                    );
                                                delete yymsp[-6].minor.yy648;
                                                delete yymsp[0].minor.yy194;
                                                yygotominor.yy147 = alter;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 448: /* cmd ::= ALTER TABLE fullname DROP CONSTRAINT nm */
{
                                                auto alter = new SqliteAlterTable();
                                                alter->initDropConstraint(
                                                        yymsp[-3].minor.yy648->name1,
                                                        yymsp[-3].minor.yy648->name2,
                                                        *(yymsp[0].minor.yy239)
                                                    );
                                                delete yymsp[-3].minor.yy648;
                                                delete yymsp[0].minor.yy239;
                                                yygotominor.yy147 = alter;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 449: /* cmd ::= ALTER TABLE fullname RENAME TO ID_TAB_NEW */
      case 450: /* cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt ID_COL */ yytestcase(yyruleno==450);
      case 455: /* cmd ::= ALTER TABLE fullname ALTER kwcolumn_opt ID_COL */ yytestcase(yyruleno==455);
      case 456: /* cmd ::= ALTER TABLE ID_DB|ID_TAB */ yytestcase(yyruleno==456);
{  yy_destructor(yypParser,209,&yymsp[-3].minor);
}
        break;
      case 451: /* cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                auto alter = new SqliteAlterTable();
                                                alter->initRenameColumn(
                                                        yymsp[-2].minor.yy648->name1,
                                                        yymsp[-2].minor.yy648->name2,
                                                        *(yymsp[0].minor.yy515),
                                                        QString(),
                                                        QString()
                                                    );
                                                delete yymsp[0].minor.yy515;
                                                delete yymsp[-2].minor.yy648;
                                                yygotominor.yy147 = alter;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 452: /* cmd ::= ALTER TABLE fullname ALTER kwcolumn_opt */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                                auto alter = new SqliteAlterTable();
                                                alter->initColumnSetNotNull(
                                                        yymsp[-2].minor.yy648->name1,
                                                        yymsp[-2].minor.yy648->name2,
                                                        *(yymsp[0].minor.yy515),
                                                        QString(),
                                                        SqliteConflictAlgo::null
                                                    );
                                                delete yymsp[0].minor.yy515;
                                                delete yymsp[-2].minor.yy648;
                                                yygotominor.yy147 = alter;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 453: /* cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt nm TO ID_COL_NEW */
{  yy_destructor(yypParser,209,&yymsp[-5].minor);
  yy_destructor(yypParser,205,&yymsp[-2].minor);
}
        break;
      case 460: /* create_vtab ::= CREATE VIRTUAL TABLE ifnotexists nm dbnm USING nm */
{
                                                yygotominor.yy147 = new SqliteCreateVirtualTable(
                                                        *(yymsp[-4].minor.yy515),
                                                        *(yymsp[-3].minor.yy239),
                                                        *(yymsp[-2].minor.yy239),
                                                        *(yymsp[0].minor.yy239)
                                                    );
                                                delete yymsp[-4].minor.yy515;
                                                delete yymsp[-3].minor.yy239;
                                                delete yymsp[-2].minor.yy239;
                                                delete yymsp[0].minor.yy239;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 461: /* create_vtab ::= CREATE VIRTUAL TABLE ifnotexists nm dbnm USING nm LP vtabarglist RP */
{
                                                yygotominor.yy147 = new SqliteCreateVirtualTable(
                                                        *(yymsp[-7].minor.yy515),
                                                        *(yymsp[-6].minor.yy239),
                                                        *(yymsp[-5].minor.yy239),
                                                        *(yymsp[-3].minor.yy239),
                                                        *(yymsp[-1].minor.yy159)
                                                    );
                                                delete yymsp[-6].minor.yy239;
                                                delete yymsp[-5].minor.yy239;
                                                delete yymsp[-3].minor.yy239;
                                                delete yymsp[-7].minor.yy515;
                                                delete yymsp[-1].minor.yy159;
                                                objectForTokens = yygotominor.yy147;
                                            }
        break;
      case 464: /* vtabarglist ::= vtabarg */
{
                                                yygotominor.yy159 = new QStringList();
                                                yygotominor.yy159->append((yymsp[0].minor.yy239)->mid(1)); // mid(1) to skip the first whitespace added in vtabarg
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 465: /* vtabarglist ::= vtabarglist COMMA vtabarg */
{
                                                yymsp[-2].minor.yy159->append((yymsp[0].minor.yy239)->mid(1)); // mid(1) to skip the first whitespace added in vtabarg
                                                yygotominor.yy159 = yymsp[-2].minor.yy159;
                                                delete yymsp[0].minor.yy239;
                                                DONT_INHERIT_TOKENS("vtabarglist");
                                            }
        break;
      case 467: /* vtabarg ::= vtabarg vtabargtoken */
{
                                                yymsp[-1].minor.yy239->append(" "+ *(yymsp[0].minor.yy239));
                                                yygotominor.yy239 = yymsp[-1].minor.yy239;
                                                delete yymsp[0].minor.yy239;
                                            }
        break;
      case 468: /* vtabargtoken ::= ANY */
{
                                                yygotominor.yy239 = new QString(yymsp[0].minor.yy0->value);
                                            }
        break;
      case 469: /* vtabargtoken ::= LP anylist RP */
{
                                                yygotominor.yy239 = new QString("(");
                                                yygotominor.yy239->append(*(yymsp[-1].minor.yy239));
                                                yygotominor.yy239->append(")");
                                                delete yymsp[-1].minor.yy239;
                                            }
        break;
      case 471: /* anylist ::= anylist LP anylist RP */
{
                                                yygotominor.yy239 = yymsp[-3].minor.yy239;
                                                yygotominor.yy239->append("(");
                                                yygotominor.yy239->append(*(yymsp[-1].minor.yy239));
                                                yygotominor.yy239->append(")");
                                                delete yymsp[-1].minor.yy239;
                                                DONT_INHERIT_TOKENS("anylist");
                                            }
        break;
      case 472: /* anylist ::= anylist ANY */
{
                                                yygotominor.yy239 = yymsp[-1].minor.yy239;
                                                yygotominor.yy239->append(yymsp[0].minor.yy0->value);
                                                DONT_INHERIT_TOKENS("anylist");
                                            }
        break;
      case 473: /* with ::= */
{yygotominor.yy137 = nullptr;}
        break;
      case 474: /* with ::= WITH wqlist */
{
                                                yygotominor.yy137 = new SqliteWith(*(yymsp[0].minor.yy512));
                                                delete yymsp[0].minor.yy512;
                                                objectForTokens = yygotominor.yy137;
                                            }
        break;
      case 475: /* with ::= WITH RECURSIVE wqlist */
{
                                                yygotominor.yy137 = new SqliteWith(*(yymsp[0].minor.yy512));
                                                yygotominor.yy137->recursive = true;
                                                delete yymsp[0].minor.yy512;
                                                objectForTokens = yygotominor.yy137;
                                            }
        break;
      case 476: /* wqas ::= AS */
{yygotominor.yy407 = new SqliteWith::CommonTableExpression::AsMode(SqliteWith::CommonTableExpression::ANY);}
        break;
      case 477: /* wqas ::= AS MATERIALIZED */
{yygotominor.yy407 = new SqliteWith::CommonTableExpression::AsMode(SqliteWith::CommonTableExpression::MATERIALIZED);}
        break;
      case 478: /* wqas ::= AS NOT MATERIALIZED */
{yygotominor.yy407 = new SqliteWith::CommonTableExpression::AsMode(SqliteWith::CommonTableExpression::NOT_MATERIALIZED);}
        break;
      case 479: /* wqlist ::= wqcte */
{
                                                yygotominor.yy512 = new ParserCteList();
                                                yygotominor.yy512->append(yymsp[0].minor.yy158);
                                            }
        break;
      case 480: /* wqlist ::= wqlist COMMA wqcte */
{
                                                yygotominor.yy512 = yymsp[-2].minor.yy512;
                                                yygotominor.yy512->append(yymsp[0].minor.yy158);
                                                DONT_INHERIT_TOKENS("wqlist");
                                            }
        break;
      case 481: /* wqlist ::= ID_TAB_NEW */
{
                                                parserContext->minorErrorBeforeNextToken("Syntax error");
                                            }
        break;
      case 482: /* wqcte ::= nm idxlist_opt wqas LP select RP */
{
                                                yygotominor.yy158 = new SqliteWith::CommonTableExpression(*(yymsp[-5].minor.yy239), *(yymsp[-4].minor.yy240), yymsp[-1].minor.yy113, *(yymsp[-3].minor.yy407));
                                                delete yymsp[-5].minor.yy239;
                                                delete yymsp[-4].minor.yy240;
                                                delete yymsp[-3].minor.yy407;
                                                objectForTokens = yygotominor.yy158;
                                            }
        break;
      case 483: /* windowdefn_list ::= windowdefn */
{
                                                    yygotominor.yy339 = new ParserWindowDefList();
                                                    yygotominor.yy339->append(yymsp[0].minor.yy522);
                                                }
        break;
      case 484: /* windowdefn_list ::= windowdefn_list COMMA windowdefn */
{
												yymsp[-2].minor.yy339->append(yymsp[0].minor.yy522);
												yygotominor.yy339 = yymsp[-2].minor.yy339;
												DONT_INHERIT_TOKENS("windowdefn_list");
											}
        break;
      case 485: /* windowdefn ::= nm AS LP window RP */
{
												yygotominor.yy522 = new SqliteWindowDefinition(*(yymsp[-4].minor.yy239), yymsp[-1].minor.yy199);
												delete yymsp[-4].minor.yy239;
												objectForTokens = yygotominor.yy522;
											}
        break;
      case 486: /* window ::= PARTITION BY nexprlist orderby_opt frame_opt */
{
												yygotominor.yy199 = new SqliteWindowDefinition::Window();
												yygotominor.yy199->initPartitionBy(QString(), *(yymsp[-2].minor.yy143), *(yymsp[-1].minor.yy459), yymsp[0].minor.yy89);
												delete yymsp[-2].minor.yy143;
												delete yymsp[-1].minor.yy459;
												objectForTokens = yygotominor.yy199;
											}
        break;
      case 487: /* window ::= nm PARTITION BY nexprlist orderby_opt frame_opt */
{
												yygotominor.yy199 = new SqliteWindowDefinition::Window();
												yygotominor.yy199->initPartitionBy(*(yymsp[-5].minor.yy239), *(yymsp[-2].minor.yy143), *(yymsp[-1].minor.yy459), yymsp[0].minor.yy89);
												delete yymsp[-2].minor.yy143;
												delete yymsp[-5].minor.yy239;
												delete yymsp[-1].minor.yy459;
												objectForTokens = yygotominor.yy199;
											}
        break;
      case 488: /* window ::= ORDER BY sortlist frame_opt */
{
												yygotominor.yy199 = new SqliteWindowDefinition::Window();
												yygotominor.yy199->initOrderBy(QString(), *(yymsp[-1].minor.yy459), yymsp[0].minor.yy89);
												delete yymsp[-1].minor.yy459;
												objectForTokens = yygotominor.yy199;
											}
        break;
      case 489: /* window ::= nm ORDER BY sortlist frame_opt */
{
												yygotominor.yy199 = new SqliteWindowDefinition::Window();
												yygotominor.yy199->initOrderBy(*(yymsp[-4].minor.yy239), *(yymsp[-1].minor.yy459), yymsp[0].minor.yy89);
												delete yymsp[-1].minor.yy459;
												delete yymsp[-4].minor.yy239;
												objectForTokens = yygotominor.yy199;
											}
        break;
      case 490: /* window ::= frame_opt */
{
												yygotominor.yy199 = new SqliteWindowDefinition::Window();
												yygotominor.yy199->init(QString(), yymsp[0].minor.yy89);
												objectForTokens = yygotominor.yy199;
											}
        break;
      case 491: /* window ::= nm frame_opt */
{
												yygotominor.yy199 = new SqliteWindowDefinition::Window();
												yygotominor.yy199->init(QString(), yymsp[0].minor.yy89);
												delete yymsp[-1].minor.yy239;
												objectForTokens = yygotominor.yy199;
											}
        break;
      case 492: /* frame_opt ::= */
{yygotominor.yy89 = nullptr;}
        break;
      case 493: /* frame_opt ::= range_or_rows frame_bound_s frame_exclude_opt */
{
												yygotominor.yy89 = new SqliteWindowDefinition::Window::Frame(*(yymsp[-2].minor.yy282), yymsp[-1].minor.yy169, nullptr, *(yymsp[0].minor.yy271));
												delete yymsp[-2].minor.yy282;
												delete yymsp[0].minor.yy271;
												objectForTokens = yygotominor.yy89;
											}
        break;
      case 494: /* frame_opt ::= range_or_rows BETWEEN frame_bound_s AND frame_bound_e frame_exclude_opt */
{
												yygotominor.yy89 = new SqliteWindowDefinition::Window::Frame(*(yymsp[-5].minor.yy282), yymsp[-3].minor.yy169, yymsp[-1].minor.yy169, *(yymsp[0].minor.yy271));
												delete yymsp[-5].minor.yy282;
												delete yymsp[0].minor.yy271;
												objectForTokens = yygotominor.yy89;
											}
        break;
      case 495: /* range_or_rows ::= RANGE|ROWS|GROUPS */
{
												yygotominor.yy282 = new SqliteWindowDefinition::Window::Frame::RangeOrRows(
													SqliteWindowDefinition::Window::Frame::toRangeOrRows(yymsp[0].minor.yy0->value)
													);
											}
        break;
      case 496: /* frame_bound_s ::= frame_bound */
      case 498: /* frame_bound_e ::= frame_bound */ yytestcase(yyruleno==498);
{
												yygotominor.yy169 = yymsp[0].minor.yy169;
												objectForTokens = yygotominor.yy169;
											}
        break;
      case 497: /* frame_bound_s ::= UNBOUNDED PRECEDING */
      case 499: /* frame_bound_e ::= UNBOUNDED FOLLOWING */ yytestcase(yyruleno==499);
      case 501: /* frame_bound ::= CURRENT ROW */ yytestcase(yyruleno==501);
{
												yygotominor.yy169 = new SqliteWindowDefinition::Window::Frame::Bound(nullptr, yymsp[-1].minor.yy0->value + " " + yymsp[0].minor.yy0->value);
												objectForTokens = yygotominor.yy169;
											}
        break;
      case 500: /* frame_bound ::= expr PRECEDING|FOLLOWING */
{
												yygotominor.yy169 = new SqliteWindowDefinition::Window::Frame::Bound(yymsp[-1].minor.yy456, yymsp[0].minor.yy0->value);
												objectForTokens = yygotominor.yy169;
											}
        break;
      case 502: /* frame_exclude_opt ::= */
{
												yygotominor.yy271 = new SqliteWindowDefinition::Window::Frame::Exclude(
													SqliteWindowDefinition::Window::Frame::Exclude::null
													);
											}
        break;
      case 503: /* frame_exclude_opt ::= EXCLUDE frame_exclude */
{
												yygotominor.yy271 = yymsp[0].minor.yy271;
											}
        break;
      case 504: /* frame_exclude ::= NO OTHERS */
{
												yygotominor.yy271 = new SqliteWindowDefinition::Window::Frame::Exclude(
													SqliteWindowDefinition::Window::Frame::Exclude::NO_OTHERS
													);
											}
        break;
      case 505: /* frame_exclude ::= CURRENT ROW */
{
												yygotominor.yy271 = new SqliteWindowDefinition::Window::Frame::Exclude(
													SqliteWindowDefinition::Window::Frame::Exclude::CURRENT_ROW
													);
											}
        break;
      case 506: /* frame_exclude ::= GROUP */
{
												yygotominor.yy271 = new SqliteWindowDefinition::Window::Frame::Exclude(
													SqliteWindowDefinition::Window::Frame::Exclude::GROUP
													);
											}
        break;
      case 507: /* frame_exclude ::= TIES */
{
												yygotominor.yy271 = new SqliteWindowDefinition::Window::Frame::Exclude(
													SqliteWindowDefinition::Window::Frame::Exclude::TIES
													);
											}
        break;
      case 508: /* window_clause ::= WINDOW windowdefn_list */
{
												yygotominor.yy339 = yymsp[0].minor.yy339;
											}
        break;
      case 509: /* filter_over ::= filter_clause over_clause */
{
												yygotominor.yy567 = new SqliteFilterOver(yymsp[-1].minor.yy341, yymsp[0].minor.yy663);
												objectForTokens = yygotominor.yy567;
											}
        break;
      case 510: /* filter_over ::= over_clause */
{
												yygotominor.yy567 = new SqliteFilterOver(nullptr, yymsp[0].minor.yy663);
												objectForTokens = yygotominor.yy567;
											}
        break;
      case 511: /* filter_over ::= filter_clause */
{
												yygotominor.yy567 = new SqliteFilterOver(yymsp[0].minor.yy341, nullptr);
												objectForTokens = yygotominor.yy567;
											}
        break;
      case 512: /* over_clause ::= OVER LP window RP */
{
												yygotominor.yy663 = new SqliteFilterOver::Over(yymsp[-1].minor.yy199);
												objectForTokens = yygotominor.yy663;
											}
        break;
      case 513: /* over_clause ::= OVER nm */
{
												yygotominor.yy663 = new SqliteFilterOver::Over(*(yymsp[0].minor.yy239));
												delete yymsp[0].minor.yy239;
												objectForTokens = yygotominor.yy663;
											}
        break;
      case 514: /* filter_clause ::= FILTER LP WHERE expr RP */
{
												yygotominor.yy341 = new SqliteFilterOver::Filter(yymsp[-1].minor.yy456);
												objectForTokens = yygotominor.yy341;
                                            }
        break;
      default:
      /* (0) input ::= cmdlist */ yytestcase(yyruleno==0);
        break;
      };
  }
  assert( yyruleno>=0 && yyruleno<(int)(sizeof(yyRuleInfo)/sizeof(yyRuleInfo[0])) );
  yygoto = yyRuleInfo[yyruleno].lhs;
  yysize = yyRuleInfo[yyruleno].nrhs;

  // Store tokens for the rule in parser context
  QList<Token*> allTokens;
  QList<Token*> allTokensWithAllInherited;
  QString keyForTokensMap;
  int tokensMapKeyCnt;
  if (parserContext->setupTokens)
  {
      if (objectForTokens)
      {
          // In case this is a list with recurrent references we need
          // to clear tokens before adding the new and extended list.
          objectForTokens->tokens.clear();
      }

      for (int i = yypParser->yyidx - yysize + 1; i <= yypParser->yyidx; i++)
      {
          QList<Token*> tokens;
          const char* fieldName = yyTokenName[yypParser->yystack[i].major];

          // Adding token being subject of this reduction. It's usually not included in the inherited tokens,
          // although if inheriting from simple statements, like "FAIL" or "ROLLBACK", this tends to be redundant with the inherited tokens.
          // That's why we're checking if it's not contained in the inherited tokens and add it only then.
          if (parserContext->isManagedToken(yypParser->yystack[i].minor.yy0) && !yypParser->yystack[i].tokens->contains(yypParser->yystack[i].minor.yy0))
              tokens += yypParser->yystack[i].minor.yy0;

          tokens += *(yypParser->yystack[i].tokens);

          if (!noTokenInheritanceFields.contains(fieldName))
          {
              if (objectForTokens)
              {
                  keyForTokensMap = fieldName;
                  tokensMapKeyCnt = 2;
                  while (objectForTokens->tokensMap.contains(keyForTokensMap))
                      keyForTokensMap = fieldName + QString::number(tokensMapKeyCnt++);

                  objectForTokens->tokensMap[keyForTokensMap] = parserContext->getTokenPtrList(tokens);
              }

              allTokens += tokens;
          }
          else
          {
              // If field is mentioned only once, then only one occurance of it will be ignored.
              // Second one should be inherited. See "anylist" definition for explanation why.
              noTokenInheritanceFields.removeOne(fieldName);
          }
          allTokensWithAllInherited += tokens;
      }

      if (objectForTokens)
      {
          objectForTokens->tokens += parserContext->getTokenPtrList(allTokens);
      }
  }

  // Clear token lists
  for (int i = yypParser->yyidx - yysize + 1; i <= yypParser->yyidx; i++)
  {
      delete yypParser->yystack[i].tokens;
      yypParser->yystack[i].tokens = nullptr;
  }

  yypParser->yyidx -= yysize;
  yyact = yy_find_reduce_action(yymsp[-yysize].stateno,(YYCODETYPE)yygoto);
  if( yyact < YYNSTATE ){
#ifdef NDEBUG
    /* If we are not debugging and the reduce action popped at least
    ** one element off the stack, then we can push the new element back
    ** onto the stack here, and skip the stack overflow test in yy_shift().
    ** That gives a significant speed improvement. */
    if( yysize ){
      yypParser->yyidx++;
      yymsp -= yysize-1;
      yymsp->stateno = (YYACTIONTYPE)yyact;
      yymsp->major = (YYCODETYPE)yygoto;
      yymsp->minor = yygotominor;
      if (parserContext->setupTokens)
          *(yypParser->yystack[yypParser->yyidx].tokens) = allTokens;
    }else
#endif
    {
      yy_shift(yypParser,yyact,yygoto,&yygotominor);
      if (parserContext->setupTokens && yypParser->yyidx >= 0)
      {
          QList<Token*>* tokensPtr = yypParser->yystack[yypParser->yyidx].tokens;
          *tokensPtr = allTokensWithAllInherited + *tokensPtr;
      }
    }
  }else{
    assert( yyact == YYNSTATE + YYNRULE + 1 );
    yy_accept(yypParser);
  }
}

/*
** The following code executes when the parse fails
*/
#ifndef YYNOERRORRECOVERY
static void yy_parse_failed(
  yyParser *yypParser           /* The parser */
){
  sqlite3_parseARG_FETCH;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sFail!\n",yyTracePrompt);
  }
#endif
  while( yypParser->yyidx>=0 ) yy_pop_parser_stack(yypParser);
  /* Here code is inserted which will be executed whenever the
  ** parser fails */
  sqlite3_parseARG_STORE; /* Suppress warning about unused %extra_argument variable */
}
#endif /* YYNOERRORRECOVERY */

/*
** The following code executes when a syntax error first occurs.
*/
static void yy_syntax_error(
  yyParser *yypParser,           /* The parser */
  int yymajor,                   /* The major type of the error token */
  YYMINORTYPE yyminor            /* The minor type of the error token */
){
  sqlite3_parseARG_FETCH;
#define TOKEN (yyminor.yy0)

    UNUSED_PARAMETER(yymajor);
    parserContext->error(TOKEN, QObject::tr("Syntax error"));
    //qDebug() << "near " << TOKEN->toString() << ": syntax error";
  sqlite3_parseARG_STORE; /* Suppress warning about unused %extra_argument variable */
}

/*
** The following is executed when the parser accepts
*/
static void yy_accept(
  yyParser *yypParser           /* The parser */
){
  sqlite3_parseARG_FETCH;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sAccept!\n",yyTracePrompt);
  }
#endif
  while( yypParser->yyidx>=0 ) yy_pop_parser_stack(yypParser);
  /* Here code is inserted which will be executed whenever the
  ** parser accepts */
  sqlite3_parseARG_STORE; /* Suppress warning about unused %extra_argument variable */
}

/* The main parser program.
** The first argument is a pointer to a structure obtained from
** "sqlite3_parseAlloc" which describes the current state of the parser.
** The second argument is the major token number.  The third is
** the minor token.  The fourth optional argument is whatever the
** user wants (and specified in the grammar) and is available for
** use by the action routines.
**
** Inputs:
** <ul>
** <li> A pointer to the parser (an opaque structure.)
** <li> The major token number.
** <li> The minor token number.
** <li> An option argument of a grammar-specified type.
** </ul>
**
** Outputs:
** None.
*/
void sqlite3_parse(
  void *yyp,                   /* The parser */
  int yymajor,                 /* The major token code number */
  sqlite3_parseTOKENTYPE yyminor       /* The value for the token */
  sqlite3_parseARG_PDECL               /* Optional %extra_argument parameter */
){
  YYMINORTYPE yyminorunion;
  int yyact;            /* The parser action. */
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  int yyendofinput;     /* True if we are at the end of input */
#endif
#ifdef YYERRORSYMBOL
  int yyerrorhit = 0;   /* True if yymajor has invoked an error */
#endif
  yyParser *yypParser;  /* The parser */

  /* (re)initialize the parser, if necessary */
  yypParser = (yyParser*)yyp;
  if( yypParser->yyidx<0 ){
#if YYSTACKDEPTH<=0
    if( yypParser->yystksz <=0 ){
      /*memset(&yyminorunion, 0, sizeof(yyminorunion));*/
      yyminorunion = yyzerominor;
      yyStackOverflow(yypParser, &yyminorunion);
      return;
    }
#endif
    yypParser->yyidx = 0;
    yypParser->yyerrcnt = -1;
    yypParser->yystack[0].stateno = 0;
    yypParser->yystack[0].major = 0;
    yypParser->yystack[0].tokens = new QList<Token*>();
  }
  yyminorunion.yy0 = yyminor;
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  yyendofinput = (yymajor==0);
#endif
  sqlite3_parseARG_STORE;

#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sInput %s [%s] (lemon type: %s)\n",
            yyTracePrompt,
            yyminor->value.toLatin1().data(),
            yyminor->typeString().toLatin1().data(),
            yyTokenName[yymajor]);  }
#endif

  do{
    yyact = yy_find_shift_action(yypParser,(YYCODETYPE)yymajor);
    if( yyact<YYNSTATE ){
      yy_shift(yypParser,yyact,yymajor,&yyminorunion);
      yypParser->yyerrcnt--;
      yymajor = YYNOCODE;
    }else if( yyact < YYNSTATE + YYNRULE ){
      yy_reduce(yypParser,yyact-YYNSTATE);
    }else{
      assert( yyact == YY_ERROR_ACTION );
#ifdef YYERRORSYMBOL
      int yymx;
#endif
#ifndef NDEBUG
      if( yyTraceFILE ){
        fprintf(yyTraceFILE,"%sSyntax Error!\n",yyTracePrompt);
      }
#endif
#ifdef YYERRORSYMBOL
      /* A syntax error has occurred.
      ** The response to an error depends upon whether or not the
      ** grammar defines an error token "ERROR".
      **
      ** This is what we do if the grammar does define ERROR:
      **
      **  * Call the %syntax_error function.
      **
      **  * Begin popping the stack until we enter a state where
      **    it is legal to shift the error symbol, then shift
      **    the error symbol.
      **
      **  * Set the error count to three.
      **
      **  * Begin accepting and shifting new tokens.  No new error
      **    processing will occur until three tokens have been
      **    shifted successfully.
      **
      */
      if( yypParser->yyerrcnt<0 ){
        yy_syntax_error(yypParser,yymajor,yyminorunion);
      }
      yymx = yypParser->yystack[yypParser->yyidx].major;
      if( yymx==YYERRORSYMBOL || yyerrorhit ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE,"%sDiscard input token %s\n",
             yyTracePrompt,yyTokenName[yymajor]);
        }
#endif
        yy_destructor(yypParser, (YYCODETYPE)yymajor,&yyminorunion);
        yymajor = YYNOCODE;
      }else{
         while(
          yypParser->yyidx >= 0 &&
          yymx != YYERRORSYMBOL &&
          (yyact = yy_find_reduce_action(
                        yypParser->yystack[yypParser->yyidx].stateno,
                        YYERRORSYMBOL)) >= YYNSTATE
        ){
          yy_pop_parser_stack(yypParser);
        }
        if( yypParser->yyidx < 0 || yymajor==0 ){
          yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
          yy_parse_failed(yypParser);
          yymajor = YYNOCODE;
        }else if( yymx!=YYERRORSYMBOL ){
          YYMINORTYPE u2;
          u2.YYERRSYMDT = 0;
          yy_shift(yypParser,yyact,YYERRORSYMBOL,&u2);
        }
      }
      yypParser->yyerrcnt = 1; // not 3 valid tokens, but 1
      yyerrorhit = 1;
#elif defined(YYNOERRORRECOVERY)
      /* If the YYNOERRORRECOVERY macro is defined, then do not attempt to
      ** do any kind of error recovery.  Instead, simply invoke the syntax
      ** error routine and continue going as if nothing had happened.
      **
      ** Applications can set this macro (for example inside %include) if
      ** they intend to abandon the parse upon the first syntax error seen.
      */
      yy_syntax_error(yypParser,yymajor,yyminorunion);
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      yymajor = YYNOCODE;

#else  /* YYERRORSYMBOL is not defined */
      /* This is what we do if the grammar does not define ERROR:
      **
      **  * Report an error message, and throw away the input token.
      **
      **  * If the input token is $, then fail the parse.
      **
      ** As before, subsequent error messages are suppressed until
      ** three input tokens have been successfully shifted.
      */
      if( yypParser->yyerrcnt<=0 ){
        yy_syntax_error(yypParser,yymajor,yyminorunion);
      }
      yypParser->yyerrcnt = 1; // not 3 valid tokens, but 1
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      if( yyendofinput ){
        yy_parse_failed(yypParser);
      }
      yymajor = YYNOCODE;
#endif
    }
  }while( yymajor!=YYNOCODE && yypParser->yyidx>=0 );
  return;
}

int sqlite3ParserFallback(int iToken) {
    return yyFallback[iToken];
}
